/***********************************************
 *  command.c
 *
 *  Copyright 2001 Eric Kratzer.
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.
 *
 *  This file is part of imp3sh.
 *
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* decoder.c */
extern XA_TimecodeInfo dinfo, tinfo;
extern flag_t stream_open;

/* list.c */
extern ListDef *active_list;
extern flag_t user_stream;

/* queue.c */
extern ListDef *history;
extern flag_t inside_queue;

/* shell.c */
extern DCharBuf *chist;
extern char *c_path, *host;

/* keys.c */
extern TermCtrl t_ctrl[];

/* expl.c */
extern char *last_expl;
extern flag_t spackle;

/* main.c */
extern volatile int tty_width;
extern flag_t show_time_stamp, disable_user_exec, piped_io_flag;
extern flag_t ignore_ints;

/* player.c */
extern flag_t external_player;

/************************************
 *        Global Variables          *
 ************************************/

DCharBuf *args;
char *cmnd;
int comm_len, edit_pos;
flag_t prompt_now, parenthesis, ignore_char_input, running_builtin;

CommandDef com[] = { 
{print_info_comm, "printinfo",
 "Print program info----> printinfo"},

{print_gpl_comm, "printgpl",
 "Print copyright-------> printgpl"},

{print_help_comm, "help",
 "Print help------------> help [<grep_strings> | 'keys']"},

{exit_comm, "exit",
 "Exit------------------> exit"},

{lgc_play_comm, "play",
 "Play------------------> play [<stream_name>]"},

{lgc_stop_comm, "stop", 
 "Stop------------------> stop"},

{lgc_pause_comm, "pause",
 "Pause-----------------> pause"},

{lgc_resume_comm, "resume",
 "Resume----------------> resume"},

{lgc_next_comm, "next",
 "Play next entry-------> next"},

{lgc_last_comm, "last",
 "Play last entry-------> last"},

{dec_seek_comm, "seek",
 "Seek------------------> seek (<offset> <scale>) | [+/-]<seconds>"},

{dec_set_mixer_comm, "volume",
 "Set volume/balance----> volume [+/-]<master_vol>"
 " [+/-]<pcm_vol> [+/-]<balance>"},

{dec_mute_comm, "mute",
 "Mute output-----------> mute"},

{dec_unmute_comm, "unmute",
 "Unmute output---------> unmute"},

{print_status_comm, "status",
 "Print status----------> status [<level> [<sublevel>]]"},

{print_time_status_comm, "timestatus",
 "Print time status-----> timestatus"},

{plist_add_comm, "addlist",
 "Add list--------------> addlist <list_name>"},

{plist_remove_comm, "dellist",
 "Delete list-----------> dellist ('all' | <list_name>)"},

{plist_reset_comm, "resetlist",
 "Reset list------------> resetlist ('all' | <list_name>)"},

{plist_rename_comm, "renamelist",
 "Rename list-----------> renamelist <list_name> <new_name>"},

{plist_use_comm, "uselist",
 "Set active list-------> uselist <list_name>"},

{plist_next_comm, "nextlist",
 "Next active list------> nextlist"},

{plist_prev_comm, "prevlist",
 "Previous active list--> prevlist"},

{plist_edit_comm, "editlist",
 "Set edit list---------> editlist <list_name>"},

{plist_print_comm, "printlist",
 "Print list------------> printlist ['-l'] [<grep_strings>]"},

{plist_status_comm, "liststatus",
 "Print list status-----> liststatus [<list_name>]"},

{plist_copy_comm, "copyentries",
 "Copy list entry(s)----> copyentries <dest_list>"
 " ('all' | <index_spec_list> | <grep_strings>)"},

{plist_add_file_comm, "addfile",
 "Add file/stream-------> addfile <stream_name1> <stream_name2> ..."},

{plist_add_dir_comm, "addir",
 "Add directory---------> addir [<path_name>]"},

{plist_add_dir_comm, "addiR",
 "Add dir (recursive)---> addiR [<path_name>]"},

{toggle_read_tags_comm, "readtags",
 "Toggle read-file-tags-> readtags"},

{plist_play_index_comm, "playindex",
 "Play list index-------> playindex <index>"},

{plist_move_comm, "moveindex",
 "Move list index-------> moveindex <targ_index> <dest_index>"},

{plist_delete_comm, "delindex",
 "Delete list index-----> delindex ('all' | <index_spec> | <grep_strings>)"},

{wght_set_index_comm, "weightindex",
 "Weight list index-----> weightindex <weight>"
 " ('all' | <index_spec> | <grep_strings>)"},    

{plist_sort_comm, "sortlist",
 "Sort list-------------> sortlist"},

{plist_save_comm, "savelist",
 "Save list-------------> savelist <file_name>"},

{plist_load_comm, "loadlist",
 "Load list-------------> loadlist <file_name>"},

{plist_save_all_comm, "saveall",
 "Save all lists--------> saveall <file_name>"},

{plist_load_all_comm, "loadall",
 "Load all lists--------> loadall <file_name>"},

{queue_print_comm, "printq",
 "Print queue-----------> printq [<grep_strings>]"},

{queue_index_comm, "addq",
 "Queue list index------> addq (<index_spec_list> | <grep_strings>)"},

{queue_files_comm, "fileq",
 "Queue files or dirs---> fileq <stream_name1> <stream_name2> ..."},

{queue_delete_comm, "delq",
 "Delete queue index----> delq ('all' | <index_spec> | <grep_strings>)"},

{queue_resize_comm, "sizeq",
 "Change queue size-----> sizeq <maximum_size>"},

{plist_set_mode_comm, "once",
 "Set mode Once---------> once"},

{plist_set_mode_comm, "repeat",
 "Set mode Repeat-------> repeat"},

{plist_set_mode_comm, "random",
 "Set mode TrueRandom---> random"},

{plist_set_mode_comm, "srandom",
 "Set mode SmartRandom--> srandom"},

{plist_set_mode_comm, "learn",
 "Set mode Learn--------> learn"},

{plist_set_weights_comm, "setlearn",
 "Set learning weights--> setlearn <play_weight> <skip_weight>"},

{plist_set_view_comm, "viewTF",
 "Set view Tag, Fname---> viewTF"},

{plist_set_view_comm, "viewTPF",
 "Set view Tag, Pth+Fn--> viewTPF"},

{plist_set_view_comm, "viewPF",
 "Set view Path+Fname---> viewPF"},

{plist_set_view_comm, "viewF",
 "Set view Filename-----> viewF"},

{toggle_auto_rprt_comm, "report",
 "Toggle auto report----> report"},

{dec_toggle_dj_comm, "djmode",
 "Toggle DJ mode--------> djmode"},

{set_time_stamp_comm, "stampOff",
 "Set stamp Off---------> stampOff"},

{set_time_stamp_comm, "stampTP",
 "Set stamp Time Played-> stampTP"},

{set_time_stamp_comm, "stampTL",
 "Set stamp Time Left---> stampTL"},

{tag_show_comm, "printtag",
 "Print file tag--------> printtag (<list_index> | <file_name>)"},

{tag_add_comm, "addtag",
 "Add/modify tag info---> addtag (<list_index> | <file_name>) <options...>"},

{tag_print_styles_comm, "printstyles",
 "Print id3 tag styles--> printstyles [<grep_strings>]"},

{plyr_print_comm, "printplayers",
 "Print players---------> printplayers"},

{plyr_add_comm, "addplayer",
 "Add player------------> addplayer ['-q' | '-b'] ['-k']"
 " <name> <command> <filetypes>"},

{dec_print_eq_comm, "eqprint",
 "Print equalizer bands-> eqprint"},

{dec_set_eq_band_comm, "eqsetband",
 "Set equalizer band(s)-> eqsetband ('all' |"
 " <band_number_list>) <band_value_list>"},

{dec_enable_eq_comm, "eqenable",
 "Enable equalizer------> eqenable"},

{dec_disable_eq_comm, "eqdisable",
 "Disable equalizer-----> eqdisable"},

{event_print_comm, "printevents",
 "Print event list------> printevents"},

{event_add_comm, "addevent",
 "Add event-------------> addevent ['-d'] ['-p'] <type_options...> <command>"},

{event_remove_comm, "delevent",
 "Delete event----------> delevent ('all' | <event_id>)"},

{key_user_show_comm, "printkeys",
 "Print special keys----> printkeys"},

{key_user_mod_comm, "modkey",
 "Modify special key----> modkey ['-n'] <key_name>"
 " (<command(s)> | 'default')"},

{source_comm, "source",
 "Exec cmnds in file----> source <file_name>"},

{chist_show_comm, "history",
 "Print command history-> history [<grep_strings>]"},

{alias_add_comm, "alias",
 "Add alias-------------> alias [<alias_name> <command>]"},

{alias_remove_comm, "unalias",
 "Remove alias----------> unalias <alias_name>"},

{proc_print_comm, "jobs",
 "Print jobs------------> jobs"},

{proc_flush_comm, "fjob",
 "Flush job buffer------> fjob <job_number>"},

{proc_2fg_comm, "fg",
 "Send job foreground---> fg [<job_number>]"},

{proc_2bg_comm, "bg",
 "Send job background---> bg [<job_number>]"},

{env_show_comm, "env",
 "Print environment-----> env [<grep_strings>]"},

{env_set_comm, "setenv",
 "Set environment-------> setenv <name> <value>"},

{mess_comm, "mess",
 "Browse text file------> mess ['-s'] <file_name>"},

{dir_change_comm, "cd",
 "Change directory------> cd ['-' | <path_name>]"},

{toggle_audio_procs_comm, "audioprocs",
 "Toggle driver mode----> audioprocs"},

{toggle_auto_recover_comm, "recover",
 "Toggle auto recover---> recover"},

{dec_open_comm, "decstart",
 "Start xaudio decoder--> decstart"},

{dec_set_output_comm, "setoutput",
 "Set xaudio output-----> setoutput <full_path_name>"},

{dlog_show_comm, "printlog",
 "Print debug log-------> printlog"},

{dlog_clear_comm, "clearlog",
 "Clear debug log-------> clearlog"},

{com_exec, NULL, NULL}
};

/************************************
 *        Local variables           *
 ************************************/

static char prompt_str[INFO_SIZE*2], prompt_str2[INFO_SIZE];
static char comm_al[COMM_SIZE], tmpstr_co[COMM_SIZE], *child_argv[COMM_SIZE];
static int old_edit_pos, prompt_len;
static flag_t end_line_flag;

/************************************
 *         Module routines          * 
 ************************************/

unsigned int com_get_bytec()
{
    unsigned int bytec;

    bytec = 0;
    bytec += dbuf_get_bytec(args);
    bytec += COMM_SIZE;

    return bytec;
}

void com_init()
{
    cmnd = malloc(COMM_SIZE);
    args = dbuf_init(NULL, DCHAR_GROW, 128, 128);

    comm_len = edit_pos = 0;
    prompt_len = old_edit_pos = end_line_flag = 0;
    prompt_now = parenthesis = ignore_char_input = running_builtin = 0;

    memset(prompt_str,  0, INFO_SIZE*2);
    memset(prompt_str2, 0, INFO_SIZE);
    memset(comm_al,     0, COMM_SIZE);
    memset(tmpstr_co,   0, COMM_SIZE);
    memset(cmnd,        0, COMM_SIZE);
    memset(child_argv,  0, sizeof(char *) * COMM_SIZE);

    if (!piped_io_flag)
	show_user("\nType \"%s\" for a list of built-in commmands\n",
		  com[PRINT_HELP].name);
}

void com_clear()
{
    free(cmnd);
    dbuf_free(args);
}

/************************************
 *      User command handling       *
 ************************************/

void com_beep()
{
    if (!piped_io_flag)
	beep();
}

int com_find_match(char *comm)
{
    int i;

    if (comm)
	for (i = 0; i < NUM_COMM; i++)
	    if (0 == strcmp(comm, com[i].name)) {
		prompt_now = TRUE;
		return i;
	    }
    
    return NO_BUILTIN_COMM;
}

int com_test_background(int *mode)
{
    char *str;
    int pos, ln;

    pos = args->len - 1;
    str = dbuf_get(args, pos);

    switch (get_cpos('&', str)) {
    case AMP_END:
	ln = get_strlen(str);
	str[ln - 1] = ZCHAR;
	*mode = J_BGRUN;
	break;
    case AMP_ALL:
	if (pos > 0) {
	    dbuf_remove(args, pos);
	    *mode = J_BGRUN;
	}
	break;
    case AMP_MID:
    case AMP_FRONT:
	show_user(ERROR_MSG " Syntax errror: misplaced \"&\"\n");
	return TRUE;
    }

    return FALSE;
}

void com_setup_child(int bg)
{
    int mypid, prgrp;

    mypid = getpid();
    prgrp = mypid;
    setpgid(mypid, prgrp);

    if (bg == J_FGRUN && !piped_io_flag)
	tcsetpgrp(0, prgrp);

    signal(SIGALRM, SIG_DFL);
    signal(SIGPIPE, SIG_DFL);
    signal(SIGTSTP, SIG_DFL);
    signal(SIGTERM, SIG_DFL);
    signal(SIGINT, SIG_DFL);
    signal(SIGTTIN, SIG_DFL);
    signal(SIGTTOU, SIG_DFL);
    signal(SIGWINCH, SIG_DFL);
    signal(SIGCHLD, SIG_DFL);
}

void /* user (indirect) */ com_exec()
{
    int i, redir, bg, tmpstdout[2], tmpstderr[2];
    pid_t pid;

    if (disable_user_exec) {
	show_user(ERROR_MSG " Command not found: \"%s\"", dbuf_get(args, 0));
	prompt_now = TRUE;
	return;
    }

    redir = FALSE;
    bg = J_FGRUN;

    if (!args->len || com_test_background(&bg)) {
	prompt_now = TRUE;
	return;
    }

    if (bg == J_FGRUN && !have_tty())
	bg = J_BGRUN;

    if (bg == J_BGRUN) {
	redir = TRUE;
	if (pipe(tmpstdout) || pipe(tmpstderr)) {
	    show_user(ERROR_MSG " Error creating pipes : %s", 
		      strerror(errno));
	    return;
	}
    }

    pid = fork();
    if (pid == -1) {

	show_user(ERROR_MSG " Failed to fork \"%s\" : %s", dbuf_get(args, 0),
		  strerror(errno));
	prompt_now = TRUE;

    } else if (pid != 0) {

	proc_add(pid, bg, redir, tmpstdout, tmpstderr, dbuf_get(args, 0));
	if (redir)
	    prompt_now = TRUE;

    } else {

	if (redir) {
	    dup2(tmpstdout[1], 1);
	    dup2(tmpstderr[1], 2);
	}

	for (i = 0; i < args->len; i++)
	    child_argv[i] = dbuf_get(args, i);
	child_argv[i] = NULL;

	com_setup_child(bg);
	usleep(15000);
	execvp(*child_argv, child_argv);
	exit(CHILD_EXEC_FAILED);
    }
}

/* very limited right now  :( */
int com_check_envvars(char *str)
{
    char *tmp;

    if (get_cpos('$', str) == AMP_FRONT) {
	tmp = getenv(&str[1]);
	if (tmp) {
	    dbuf_add_str(args, tmp);
	    return TRUE;
	}
    }

    return FALSE;
}

int com_check_glob(char *str)
{
    int i, ret, status;
    glob_t pglob;

    i = 0; 
    status = FALSE;
    memset(&pglob, 0, sizeof(glob_t));
    ret = glob(str, GLOB_TILDE | GLOB_MARK | GLOB_NOSORT, NULL, &pglob);

    if (ret == 0 && pglob.gl_pathc > 0) {

	while(pglob.gl_pathv[i])
	    dbuf_add_str(args, pglob.gl_pathv[i++]);

	status = TRUE;
    }

    globfree(&pglob);
    return status;
}

int com_add_arg(char *str, char paren)
{
    unescape_string(str);

    if (paren)
	dbuf_add_str(args, str);
    else if (com_check_glob(str))
	;
    else if (com_check_envvars(str))
	;
    else
	dbuf_add_str(args, str);

    return TRUE;
}

void com_make_args(char *command)
{
    char *tmem, *cm, paren;

    dbuf_clear(args);
    tmem = NULL;
    cm = arg_tokenize(command, &tmem, ' ', &paren);

    while (cm) {
	if (!com_add_arg(cm, paren))
	    break;
	cm = arg_tokenize(NULL, &tmem, ' ', &paren);
    }
}

void com_make_args_kparen(char *command)
{
    char *c, *t;

    dbuf_clear(args);
    c = command;
    t = tmpstr_co;
    *t = ZCHAR;

    while (TRUE) {
	if (*c == ' ') {
	    if (t != tmpstr_co) {
		*t = ZCHAR;
		t = tmpstr_co;
		dbuf_add_str(args, tmpstr_co);
	    }
	} else if (*c == ZCHAR) {
	    if (t != tmpstr_co) {
		*t = ZCHAR;
		t = tmpstr_co;
		dbuf_add_str(args, tmpstr_co);
	    }

	    break;
	} else if (*c == '\"') {
	    *t++ = *c++;

	    while (*c != ZCHAR && *c != '\"')
		*t++ = *c++;

	    if (*c == '\"')
		*t++ = *c;

	    *t = ZCHAR;
	    t = tmpstr_co;
	    dbuf_add_str(args, tmpstr_co);
	} else
	    *t++ = *c;

	c++;
    }
}

void com_chistory(int ndx, char *start, char *command)
{
    char *ncm;

    ncm = strchr(start, ' ');

    if (ncm) {
	safe_strcpy(tmpstr_co, dbuf_get(chist, ndx), COMM_SIZE);
	strcat(tmpstr_co, ncm);
	safe_strcpy(command, tmpstr_co, COMM_SIZE);
    } else
	safe_strcpy(command, dbuf_get(chist, ndx), COMM_SIZE);

    if (command == cmnd) {
	comm_len = edit_pos = get_strlen(command);
	show_user("%s", command);
    }
}

int com_check_bang(char *command)
{
    char *tmp;
    int i, hdx, len;

    tmp = skip_space(command);

    if (tmp && *tmp == '!') {
	hdx = atoi(&tmp[1]);
	len = get_strlen(&tmp[1]);
	i = chist->len;

	if (len && i) {
	    if (*(tmp + 1) == '!')
		hdx = chist->len;
	    else if (!hdx)
		for (i--; i >= 0; i--)
		    if (strncmp(&tmp[1], dbuf_get(chist, i), len) == 0) {
			hdx = i + 1;
			break;
		    }
	}

	if (hdx > 0 && hdx <= chist->len)
	    com_chistory(hdx - 1, tmp, command);
	else {
	    show_user(ERROR_MSG " No history match found!");
	    prompt_now = TRUE;
	    return TRUE;
	}
    }

    return FALSE;
}

/* Multiple seperator tokenizing: ' ', ';', and '\0'     */
/* Aliasing only allowed on first argument of each cmnd. */
/* Not very efficient, but solid                         */
void com_expand_aliases(char *command)
{
    char *t, *c, *al;
    int adx, go, alias;

    *comm_al = ZCHAR;
    c = skip_space(command);
    go = alias = TRUE;

    if (!c)
	return;

    while (go) {
	t = tmpstr_co;
	while (*c != ' ' && *c != ZCHAR && *c != ';' && *c != '&')
	    *t++ = *c++;

	*t = ZCHAR;

	if (alias) {
	    al = skip_space(tmpstr_co);
	    if (al) {
		alias = FALSE;
		adx = alias_check(al);
		if (adx != -1)
		    safe_strcpy(tmpstr_co, alias_get(adx), COMM_SIZE);
	    }
	}

	if (*c == ZCHAR)
	    go = FALSE;
	else if (*c == ';')
	    alias = TRUE;

	strcat(comm_al, tmpstr_co);
	tmpstr_co[0] = *c++;
	tmpstr_co[1] = ZCHAR;
	strcat(comm_al, tmpstr_co);
    }

    safe_strcpy(command, comm_al, COMM_SIZE);
}

int com_check_special(char *command)
{
    char *cm;
    int a_len;

    cm = skip_space(command);

    if (!cm)
	return FALSE;

    a_len = get_strlen(com[ADD_ALIAS].name);
    if (strncmp(cm, com[ADD_ALIAS].name, a_len) == 0 &&
	command[a_len] == ' ') return TRUE;

    a_len = get_strlen(com[ADD_EVENT].name);
    if (strncmp(cm, com[ADD_EVENT].name, a_len) == 0 &&
	command[a_len] == ' ') return TRUE;

    a_len = get_strlen(com[MOD_KEY].name);
    if (strncmp(cm, com[MOD_KEY].name, a_len) == 0 &&
	command[a_len] == ' ') return TRUE;

    return FALSE;
}

void com_run(int ndx)
{
    if (ndx < NO_BUILTIN_COMM)
	running_builtin = TRUE;
    (com[ndx].func) ();
    if (ndx < NO_BUILTIN_COMM)
	running_builtin = FALSE;
}

void com_handle(char *command)
{
    char *cmem, *cm;
    int ndx;

    if (com_check_bang(command))
	return;

    cmem = cm = NULL;
    ndx = 0;

    chist_add(command);

    if (com_check_special(command)) {
	prompt_now = FALSE;
	com_make_args_kparen(command);
	ndx = com_find_match(dbuf_get(args, 0));
	com_run(ndx);
    } else {
	com_expand_aliases(command);

	cm = tokenize(command, &cmem, ';');
	while (cm) {
	    prompt_now = FALSE;
	    com_make_args(cm);
	    ndx = com_find_match(dbuf_get(args, 0));
	    com_run(ndx);
	    cm = tokenize(NULL, &cmem, ';');
	}
    }
}

/**** Command line manipulation functions *****/

void com_info_clear()
{
    chist->pos = chist->len;
    last_expl = NULL;
    comm_len = old_edit_pos = edit_pos = 0;
    parenthesis = spackle = end_line_flag = FALSE;
    memset(cmnd, 0, COMM_SIZE);
}

void com_ctrl_str(int len, int n)
{
    int i;

    if (!piped_io_flag && len > 0 && t_ctrl[n].string)
	for (i = 0; i < len; i++)
	    out_str_n(t_ctrl[n].string);
}

void com_backup_cursor(int len)
{
    int y, p;

    if (!tty_width)
	return;

    y = len / tty_width;
    p = len % tty_width;

    if (p)
	y++;
    else if (end_line_flag) {
	end_line_flag = FALSE;
	y++;
    }

    com_ctrl_str(y - 1, TC_UP);
}

void com_restore_cursor(int len, int len2)
{
    int y, p, cy;

    if (!tty_width)
	return;

    y = len / tty_width;
    p = len % tty_width;
    cy = len2 / tty_width;

    if (len2 % tty_width)
	cy++;

    if (p)
	y++;
    else if (edit_pos) {
	end_line_flag = TRUE;
	y++;
    }

    com_ctrl_str(cy - y, TC_UP);
    com_ctrl_str(1, TC_CR);
    com_ctrl_str(p, TC_ND);
}

void prompt(int flush)
{
    char list_name[28];
    int tmp, hr, mn, sc;
    flag_t time_stamp;

    if (!have_tty() || piped_io_flag || ignore_ints)
	return;

    if (show_time_stamp && (external_player || !stream_open))
	time_stamp = TIME_NORM;
    else
	time_stamp = show_time_stamp;

    if (user_stream)
	safe_sprintf(prompt_str, INFO_SIZE, "[%s@%s ", NAME, host);
    else if (queue_in() || inside_queue)
	safe_sprintf(prompt_str, INFO_SIZE, "[%d->queue ", history->dcb->pos);
    else if (active_list) {
	safe_strcpy(list_name, active_list->dcb->name, 25);
	if (get_strlen(active_list->dcb->name) > 24)
	    strncpy(&list_name[24], "...\0", 4);
	safe_sprintf(prompt_str, INFO_SIZE, "[%d->%s ", active_list->dcb->pos, 
		     list_name);
    } else
	safe_sprintf(prompt_str, INFO_SIZE, "[%s@%s ", NAME, host);

    switch (time_stamp) {
    case TIME_NORM:
	if (tinfo.h)
	    safe_sprintf(prompt_str2, INFO_SIZE, "%.2d:%.2d:%.2d", tinfo.h, 
			 tinfo.m, tinfo.s);
	else
	    safe_sprintf(prompt_str2, INFO_SIZE, "%.2d:%.2d ", tinfo.m, 
			 tinfo.s);
	strcat(prompt_str, prompt_str2);
	break;
    case TIME_LEFT:
	tmp = ((dinfo.h * 3600) + (dinfo.m * 60) + dinfo.s) - 
	    ((tinfo.h * 3600) + (tinfo.m * 60) + tinfo.s);
	hr = tmp / 3600; tmp -= 3600*hr;
	mn = tmp / 60; tmp -= 60*mn;
	sc = tmp;
	if (hr)
	    safe_sprintf(prompt_str2, INFO_SIZE, "-%.2d:%.2d:%.2d ", 
			 hr, mn, sc);
	else
	    safe_sprintf(prompt_str2, INFO_SIZE, "-%.2d:%.2d ", mn, sc);
	strcat(prompt_str, prompt_str2);
	break;
    }
    
    safe_sprintf(prompt_str2, INFO_SIZE, "%s]> ", file_from_path(c_path));
    strcat(prompt_str, prompt_str2);
    prompt_str[(INFO_SIZE*2)-1] = 0;

    prompt_len = get_strlen(prompt_str);
    out_str_n(prompt_str);
    if (flush)
	out_flush();
}

void update_prompt(int backup)
{   
    if (!have_tty() || piped_io_flag || ignore_ints)
	return;

    if (backup)
	com_backup_cursor(prompt_len + old_edit_pos);

    com_ctrl_str(1, TC_CR);
    prompt(FALSE);
    out_str_n(cmnd);
    com_ctrl_str(1, TC_CLR_TOBOT);

    if (comm_len != edit_pos)
	com_restore_cursor(prompt_len + edit_pos, prompt_len + comm_len);

    old_edit_pos = edit_pos;
    out_flush();
}

void com_backspace()
{
    if (!comm_len || !edit_pos)
	com_beep();
    else {
	if (cmnd[edit_pos - 1] == '\"')
	    parenthesis = !parenthesis;
	if (edit_pos == comm_len)
	    cmnd[edit_pos - 1] = ZCHAR;
	else
	    memmove(&cmnd[edit_pos - 1],
		    &cmnd[edit_pos], (comm_len - edit_pos + 1));
	comm_len--;
	edit_pos--;
	last_expl = NULL;
	update_prompt(TRUE);
    }
}

void com_delete()
{
    if (!outb_empty())
	;
    else if (!comm_len || edit_pos == comm_len)
	com_beep();
    else {
	if (cmnd[edit_pos] == '\"')
	    parenthesis = !parenthesis;
	memmove(&cmnd[edit_pos],
		&cmnd[edit_pos + 1], (comm_len - edit_pos + 1));
	comm_len--;
	last_expl = NULL;
	update_prompt(TRUE);
    }
}

void com_insert(char in)
{
    memmove(&cmnd[edit_pos + 1], &cmnd[edit_pos], comm_len - edit_pos);
    cmnd[edit_pos] = in;
    comm_len++;
    edit_pos++;
    last_expl = NULL;
    update_prompt(TRUE);
}

void com_append(char in)
{
    cmnd[comm_len++] = in;
    edit_pos = comm_len;
    last_expl = NULL;
    update_prompt(TRUE);
}

void com_clear_scr()
{
    com_ctrl_str(1, TC_CLR);
    update_prompt(FALSE);
}

void com_decode(char chr)
{
    char *str;

    switch (chr) {
    case 1 /* ^A */ :
	edit_pos = 0;
	update_prompt(TRUE);
	break;

    case 2: /* ^B */
	key_handle_sp(LEFT_ARROW);
	break;

    case 3: /* ^C */
	com_info_clear();
	if (!piped_io_flag)
	    out_str_n("\n");
	update_prompt(FALSE);
	break;

    case 4: /* ^D */
	if (comm_len == 0) {
	    if (!piped_io_flag)
		out_str("exit");
	    exit_comm();
	} else
	    com_beep();
	break;

    case 5: /* ^E */
	edit_pos = comm_len;
	update_prompt(TRUE);
	break;

    case 6: /* ^F */
	key_handle_sp(RIGHT_ARROW);
	break;

    case 8: /* backspace */
    case 127: 
	com_backspace();
	break;

    case 9: /* tab */
	if (!piped_io_flag)
	    expl_tab();
	break;

    case 11: /* ^K */
	if (edit_pos != comm_len) {
	    memset(&cmnd[edit_pos], 0, COMM_SIZE - edit_pos);
	    last_expl = NULL;
	    comm_len = edit_pos;
	}
	update_prompt(TRUE);
	break;

    case 12: /* ^L */
	com_clear_scr();
	break;

    case 14: /* ^N */
	key_handle_sp(DOWN_ARROW);
	break;

    case 16: /* ^P */
	key_handle_sp(UP_ARROW);
	break;

    case 21: /* ^U */
	if (edit_pos != 0) {
	    memmove(cmnd, &cmnd[edit_pos], comm_len - edit_pos);
	    last_expl = NULL;
	    comm_len -= edit_pos;
	    edit_pos = 0;
	    memset(&cmnd[comm_len], 0, COMM_SIZE - comm_len);
	} else
	    com_beep();

	update_prompt(TRUE);
	break;

    case END_COMM:		/* '\n' */
	if (!piped_io_flag) {
	    out_str_n("\n");
	    out_flush();
	}

	str = skip_space(cmnd);
	if (comm_len && *str != ZCHAR && *str != '#') {
	    cmnd[comm_len++] = ZCHAR;
	    com_handle(cmnd);
	} else
	    prompt_now = TRUE;

	com_info_clear();
	if (prompt_now) {
	    conb_flush();
	    prompt(TRUE);
	}
	break;

    case '"':
	parenthesis = !parenthesis;
    default:
	if (chr < 32 && chr >= 0)
	    break;

	if (comm_len < (COMM_SIZE - 1)) {
	    if (edit_pos == comm_len)
		com_append(chr);
	    else
		com_insert(chr);
	} else {
	    show_user("\n" ERROR_MSG " Command too big, ignoring...\n");
	    com_info_clear();
	    prompt(TRUE);
	}
	break;
    }
}

void com_read_special()
{
    char spchars[16], ins;
    struct timeval tv;
    fd_set readset;
    int i, sps, len, ret;

    sps = 1;
    *spchars = 27;

    do {
	len = 0;

	do {
	    FD_ZERO(&readset);
	    FD_SET(0, &readset);
	    tv.tv_sec = 0;
	    tv.tv_usec = 750000;
	    ret = select(0 + 1, &readset, NULL, NULL, &tv);
	}
	while (ret == -1 && errno == EINTR);

	if (FD_ISSET(0, &readset)) {
	    do {
		len = read(0, &ins, 1);
	    }
	    while (len == -1 && errno == EINTR);

	    if (len <= 0)
		break;
	    else
		spchars[sps++] = ins;

	    if (sps >= 16)
		break;
	    else if (key_check_sp(spchars))
		return;
	}
    }
    while (len);

    if (sps > 1)
	for (i = 0; i < sps; i++)
	    com_decode(spchars[i]);
}

int com_monitor()
{
    char chr;
    int len;

    len = chr = 0;

    do {
	len = read(0, &chr, 1);
    }
    while (len == -1 && errno == EINTR);

    if (len == 0)
	return FALSE;
    else if (len < 0)
	return TRUE;		/* ignore err */

    if (chr == 27)
	com_read_special();
    else if (!outb_check(chr))
	com_decode(chr);

    return TRUE;
}
