/***********************************************
 *  dbuf.c 
 *                                            
 *  Copyright 2001 Eric Kratzer.               
 *
 *  This file is part of imp3sh.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *   DChar buffer manipulation      * 
 ************************************/

char *dbuf_get(DCharBuf * dcb, int pos)
{				/* get data */
    return (pos >= 0 && pos < dcb->len) ? dcb->buf[pos].ptr : NULL;
}

short dbuf_get_info(DCharBuf * dcb, int pos)
{                               /* get info */
    return (pos >= 0 && pos < dcb->len) ? dcb->buf[pos].info : 0;
}

void dbuf_set_info(DCharBuf * dcb, int pos, short info)
{                               /* set info byte */
    if (pos >= 0 && pos < dcb->len) dcb->buf[pos].info = info;
}

unsigned short dbuf_get_len(DCharBuf *dcb, int pos)
{                               /* get mem length */
    return (pos >= 0 && pos < dcb->len) ? dcb->buf[pos].len : 0;
}

void dbuf_clear(DCharBuf * dcb)
{				/* clear */
    if (dcb) dcb->pos = dcb->len = dcb->var = 0;
}

void dbuf_verify(DCharBuf * dcb)
{                               /* verify pos and len */
    if (!dcb)
	return;

    if (dcb->len > dcb->size)
	dcb->len = dcb->size;
    if (dcb->pos >= dcb->len)
	dcb->pos = (dcb->len - 1);
    if (dcb->pos < 0)
	dcb->pos = 0;
}

unsigned int dbuf_get_bytec(DCharBuf * dcb)
{				/* get byte count */
    unsigned int bytec;

    bytec = 0;

    if (dcb) {
	bytec += dcb->bytec;
	bytec += dcb->size * sizeof(DCBNode);
	bytec += sizeof(DCharBuf);
	if (dcb->name)
	    bytec += get_strlen(dcb->name) + 1;
    }

    return bytec;
}

void dbuf_free(DCharBuf * dcb)
{				/* free */
    int i;

    if (!dcb)
	return;

    for (i = 0; i < dcb->size; i++)
	free(dcb->buf[i].ptr);

    free(dcb->name);
    free(dcb->buf);
    free(dcb);
}

DCharBuf *dbuf_init(char *desc, int mode, int size, int rate)
{				/* init */
    DCharBuf *dcb; 

    if (size <= 0)
	return NULL;

    dcb = malloc(sizeof(DCharBuf));
    memset(dcb, 0, sizeof(DCharBuf));
    dcb->buf = malloc(sizeof(DCBNode) * size);
    memset(dcb->buf, 0, sizeof(DCBNode) * size);

    dcb->mode = mode;
    dcb->size = size;

    if (mode == DCHAR_GROW && rate == 0)
	dcb->rate = 1;
    else
	dcb->rate = rate;

    if (desc)
	dcb->name = strdup(desc);

    return dcb;
}

int dbuf_swap(DCharBuf * dcb, int ind1, int ind2)
{				/* swap */
    DCBNode tmp;

    if (ind1 < 0 || ind1 >= dcb->len || ind2 < 0 || ind2 >= dcb->len)
	return FALSE;

    tmp = dcb->buf[ind1];

    dcb->buf[ind1] = dcb->buf[ind2];
    dcb->buf[ind2] = tmp;

    return TRUE;
}

int dbuf_resize(DCharBuf * dcb, int new_size)
{                              /* resize */
    DCBNode *tmp;
    int i;

    tmp = dcb->buf;

    if (new_size < dcb->size)
	for (i = new_size; i < dcb->size; i++) {
	    dcb->bytec -= tmp[i].len;
	    free(tmp[i].ptr);
	}

    dcb->buf = realloc(tmp, new_size * sizeof(DCBNode));

    if (!dcb->buf) {
	dcb->buf = tmp;
	dcb->mode = DCHAR_NORM;
	return FALSE;
    }

    if (new_size > dcb->size)
	memset(&dcb->buf[dcb->size], 0, 
	       (new_size - dcb->size) * sizeof(DCBNode));

    dcb->size = new_size;
    dbuf_verify(dcb);

    return TRUE;
}

int dbuf_remove_range(DCharBuf *dcb, int low, int high)
{
    int i;                      /* remove range          */
    DCBNode *buf;               /* [low, high] inclusive */

    if (!dcb || (low < 0) || (low > high) || high >= dcb->len)
	return FALSE;

    buf = dcb->buf;
    
    for(i = low; i <= high; i++) {
	dcb->bytec -= buf[i].len;
	free(buf[i].ptr);
    }

    if (high < (dcb->size-1)) {
	memmove(&buf[low], &buf[high+1], 
		sizeof(DCBNode) * (dcb->size - (high+1)));
	memset(&buf[dcb->size-(high-low+1)], 0,
	       sizeof(DCBNode) * (high-low+1));
    }

    if (dcb->pos > high)
	dcb->pos -= (high-low) + 1;
    else if (dcb->pos > low)
	dcb->pos = low;

    dcb->len -= (high-low) + 1;
    dbuf_verify(dcb);

    return TRUE;
}

int dbuf_remove(DCharBuf * dcb, int ndx)
{				/* remove index (slow) */
    DCBNode *buf;          
    int len, siz;

    buf = dcb->buf;
    len = dcb->len;
    siz = dcb->size;

    if (ndx < 0 || ndx > (len - 1))
	return FALSE;

    if (ndx < (len - 1)) {
	dcb->bytec -= buf[ndx].len;
	free(buf[ndx].ptr);
	memmove(&buf[ndx], &buf[ndx + 1], sizeof(DCBNode) * (siz - ndx - 1));
	memset(&buf[siz - 1], 0, sizeof(DCBNode));
    }

    dcb->len--;
    if (dcb->pos >= ndx)
	dcb->pos--;
    dbuf_verify(dcb);

    return TRUE;
}

int dbuf_next_pos(DCharBuf *dcb) 
{                               /* get next avail position */
    DCBNode *buf, tmpbuf;
    int pos;

    pos = -1;
    buf = dcb->buf;

    if (dcb->len < dcb->size)
	pos = dcb->len++;
    else
	switch (dcb->mode) {
	case DCHAR_SHIFT:
	    tmpbuf = *buf;
	    memmove(buf, (buf + 1), sizeof(DCBNode) * (dcb->size - 1));
	    pos = dcb->size - 1;
	    buf[pos] = tmpbuf;
	    break;
	case DCHAR_GROW:
	    if (dbuf_resize(dcb, dcb->size + dcb->rate))
		pos = dcb->len++;
	    break;
	}

    return pos;
}

int dbuf_add_ptr(DCharBuf *dcb, void *data, int len, short info)
{			        /* add, mem already allocated   */
    DCBNode *buf;               /* mem is now _ours_ to manage  */
    int pos;

    pos = dbuf_next_pos(dcb);
    if (pos == -1)
	return FALSE;

    buf = &dcb->buf[pos];
    buf->info = info;

    dcb->bytec += len - buf->len;
    buf->len = len;

    free(buf->ptr);
    buf->ptr = (char *) data;

    return TRUE;
}
    
int dbuf_add(DCharBuf *dcb, void *data, int len, short info)
{			        /* add, we copy data  */
    DCBNode *buf;               /* and manage memory  */
    int pos;

    pos = dbuf_next_pos(dcb);
    if (pos == -1)
	return FALSE;

    buf = &dcb->buf[pos];
    buf->info = info;

    if (buf->len == 0) {
	dcb->bytec += len;
	buf->len = len;
	buf->ptr = malloc(len);
    } else if (len > buf->len) {
	dcb->bytec += len - buf->len;
	buf->len = len;
	buf->ptr = realloc(buf->ptr, len);
    }

    memcpy(buf->ptr, data, len);
    return TRUE;
}

/************************************
 *       dbuf string functions      *
 ************************************/

int dbuf_add_str(DCharBuf * dcb, char *str)
{				/* add string */
    return dbuf_add(dcb, (void *)str, get_strlen(str) + 1, 0);
}

int dbuf_add_stri(DCharBuf * dcb, char *str, short info)
{				/* add string with info */
    return dbuf_add(dcb, (void *)str, get_strlen(str) + 1, info);
}

int dbuf_compare_str(const void *vel1, const void *vel2)
{                               /* compare as strings */
    return strcmp(((DCBNode *) (vel1))->ptr, ((DCBNode *) (vel2))->ptr);
}

void dbuf_sort_str(DCharBuf * dcb)
{                               /* sort string dbuf */
    if (dcb && dcb->len > 1)
	qsort((void *) dcb->buf, dcb->len, sizeof(DCBNode), dbuf_compare_str);
}

int dbuf_check_str(DCharBuf * dcb, char *str)
{                               /* check for exact string */
    DCBNode *buf;
    int i;

    if (!dcb || !dcb->len || !str)
        return -1;

    for(i = 0, buf = dcb->buf; i < dcb->len; i++)
        if (0 == strcmp(str, buf[i].ptr))
	    return i;

    return -1;
}

int dbuf_rem_str_dups(DCharBuf *dcb)
{                               /* remove string duplicates (slow) */
    DCBNode *buf;
    int i, j, len, dup;

    if (!dcb || !dcb->len)
	return 0;

    len = dcb->len;
    buf = dcb->buf;
    dup = 0;

    for(i = 0; i < len; i++)
	for(j = i+1; j < len; j++)
	    if (0 == strcmp(buf[i].ptr, buf[j].ptr)) {
		dbuf_remove(dcb, j);
		j --; len --; dup ++;
	    }

    return dup;
}
