/***********************************************
 *  expl.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.               
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* queue.c */
extern ListDef *history;

/* list.c */
extern ListDef *edit_list;

/* command.c */
extern CommandDef com[];
extern char *cmnd;
extern int comm_len, edit_pos;

/* shell.c */
extern DCharBuf *aliases;
extern char *c_path;
extern flag_t prompt_now, parenthesis;

/* main.c */
extern flag_t disable_user_exec;

/************************************
 *        Global Variables          *
 ************************************/

DCharBuf *lcomp, *ecomp, *qcomp, *expl;
char *last_expl;
flag_t spackle;

/************************************
 *        Local variables           *
 ************************************/

static flag_t expl_complete;
static char tmpstr_ex[PATH_MAX], tmpstr2_ex[PATH_MAX];
static char match_ex[PATH_MAX], match2_ex[PATH_MAX]; 

/************************************
 *         Module routines          * 
 ************************************/

unsigned int expl_get_bytec()
{
    unsigned int bytec;

    bytec = 0;
    bytec += dbuf_get_bytec(lcomp);
    bytec += dbuf_get_bytec(ecomp);
    bytec += dbuf_get_bytec(qcomp);
    bytec += dbuf_get_bytec(expl);

    return bytec;
}

void expl_init()
{
    last_expl = NULL;
    expl_complete = spackle = FALSE;

    expl = dbuf_init(NULL, DCHAR_GROW, 64, 64);

    lcomp = dbuf_init(NULL, DCHAR_GROW, 32, 32);
    dbuf_add_str(lcomp, com[ADD_LIST].name);
    dbuf_add_str(lcomp, com[USE_LIST].name);
    dbuf_add_str(lcomp, com[REMOVE_LIST].name);
    dbuf_add_str(lcomp, com[RESET_LIST].name);
    dbuf_add_str(lcomp, com[RENAME_LIST].name);
    dbuf_add_str(lcomp, com[EDIT_LIST].name);
    dbuf_add_str(lcomp, com[PRINT_LIST_STATUS].name);

    /*
     * Current strategy only allows each command to be 
     * in one completion list-type. Do you want this command
     * to complete on list-names or list-entries?? I say 
     * list entries
     */
    /* 
      dbuf_add_str(lcomp, com[ADD_TO_LIST].name);
    */

    ecomp = dbuf_init(NULL, DCHAR_GROW, 32, 32);
    dbuf_add_str(ecomp, com[PRINT_LIST].name);
    dbuf_add_str(ecomp, com[ADD_TO_LIST].name);
    dbuf_add_str(ecomp, com[DELETE_LIST_INDEX].name);
    dbuf_add_str(ecomp, com[SET_INDEX_WEIGHT].name);
    dbuf_add_str(ecomp, com[QUEUE_INDEX].name);

    qcomp = dbuf_init(NULL, DCHAR_GROW, 32, 32);
    dbuf_add_str(qcomp, com[PRINT_QUEUE].name);
    dbuf_add_str(qcomp, com[DELETE_QUEUE_INDEX].name);

    memset(tmpstr_ex,  0, PATH_MAX);
    memset(tmpstr2_ex, 0, PATH_MAX);
    memset(match_ex,   0, PATH_MAX);
    memset(match2_ex,  0, PATH_MAX);
}

void expl_clear()
{
    dbuf_free(lcomp);
    dbuf_free(ecomp);
    dbuf_free(qcomp);
    dbuf_free(expl);
}

/************************************
 *       Completion Functions       *
 ************************************/

void expl_beep(char *s)
{
    com_beep();
    last_expl = s;
}

void expl_spackle(char *s, char *str)
{
    if (!s || !*s)
	return;

    if (str && !disable_user_exec)
	outb_add("%s (%s)", s, str);
    else
	outb_add("%s", s);
}

DIR *expl_open_dir(char *dir)
{
    DIR *tmp;

    if (!get_strlen(dir))
	tmp = opendir(".");
    else {
	tmp = opendir(dir);
	chdir(dir);
    }

    return tmp;
}

/* returns true if there is a '/' in the cmnd. */
int expl_is_path(char *srch) 
{
    int i, result;

    result = FALSE;

    for (i = edit_pos; i >= 0; i--) {
	if (srch[i] == '/') {
	    result = TRUE;
	    break;
	}
    }

    return result;
}

/* Find the last important space in cmnd.  Return a pointer into cmnd
   that points to the last space.  Spaces preceded by a backslash are
   ignored because, well, they are escaped and they should be ignored.
    
    returns cmnd if there is no space in the command.
 
    global: uses edit_pos, parenthesis
 
   FIXME: this doesn't deal properly if there are TWO backslashes in a
   row, meaning a *real* backslash. 
*/
char *expl_find_last_space(char *cmnd) 
{
    int i, prev;
    char *srch;

    srch = cmnd;

    for (i = edit_pos; i >= 0; i--) {
	prev = (i < 1) ? i : i-1;
	
	if (((cmnd[i] == ' ' && cmnd[prev] != '\\')  && !parenthesis) 
	    || (parenthesis && cmnd[i] == '\"')) {
	    srch = &cmnd[++i];
	    break;
	}
    }
    
    return srch;
}

int expl_find_common_len(char *one, char *two)
{
    int i, len1, len2;

    len1 = get_strlen(one);
    len2 = get_strlen(two);

    if (len2 < len1)
	len1 = len2;

    for (i = 0; i < len1; i++)
	if (one[i] != two[i])
	    return i;

    return len1;
}

int expl_exch_if_longer(char *orig, char *new)
{
    int retv, olen, nlen;

    retv = FALSE;
    olen = get_strlen(orig);
    nlen = get_strlen(new);

    if (nlen > olen) {
	safe_strcpy(orig, new, INFO_SIZE);
	retv = TRUE;
    }

    return retv;
}

int expl_check_file(int ftype, mode_t mode)
{
    switch (ftype) {
    case ANY_FILE:
	return TRUE;
    case DIRS_ONLY:
	return (S_ISDIR(mode)) ? TRUE : FALSE;
    case EXEC_FILE_ONLY:
	return (((S_IXUSR|S_IXGRP|S_IXOTH) & mode) && 
		!S_ISDIR(mode)) ? TRUE : FALSE; 
    case EXEC_AND_DIRS:
	return (((S_IXUSR | S_IXGRP | S_IXOTH) & mode) || 
		S_ISDIR(mode)) ? TRUE : FALSE;
    }

    return FALSE;
}

/* searches COLON seperated path list <dir_s> for best
   match to <file>.                                      */
void expl_find_file_match(int ftype, char *dir_s, char *file, int flen, 
			  int *gdl, mode_t * type, char *match)
{
    struct dirent *dirent;
    struct stat sstat;
    char *dir, *tmem; 
    int gdl2, pcnt;
    DIR *dirstr;

    tmem = NULL;
    *match = ZCHAR;
    *gdl = gdl2 = pcnt = 0;

    dbuf_clear(expl);
    tokenize_to_buf(expl, dir_s, ':');
    dbuf_rem_str_dups(expl);
    dir = dbuf_get(expl, pcnt++);

    do {
	dirstr = expl_open_dir(dir);
	if (dirstr) {
	    dirent = readdir(dirstr);
	    while (dirent) {
		if (strncmp(dirent->d_name, file, flen) == 0 &&
		    strcmp(dirent->d_name, ".") != 0 &&
		    strcmp(dirent->d_name, "..") != 0) {
		    stat(dirent->d_name, &sstat);
		    if (expl_check_file(ftype, sstat.st_mode)) {
			if (spackle)
			    expl_spackle(dirent->d_name, NULL);
			else if (*match) {
			    gdl2 = expl_find_common_len(match, dirent->d_name);
			    if (gdl2 < *gdl) *gdl = gdl2;
			    if (expl_exch_if_longer(match, dirent->d_name))
				*type = sstat.st_mode;
			} else {
			    safe_strcpy(match, dirent->d_name, INFO_SIZE);
			    *type = sstat.st_mode;
			    *gdl = get_strlen(match);
			}
		    }
		}
		dirent = readdir(dirstr);
	    }
	    closedir(dirstr);
	}
	dir = dbuf_get(expl, pcnt++);
    } while (dir);

    chdir(c_path);
}

void expl_file(char *srch, int file_type)
{
    char *file;
    mode_t type;
    int gdl, len, flen, dlen;

    /* setup vars */
    safe_strcpy(tmpstr_ex, srch, COMM_SIZE);
    dlen = get_strlen(tmpstr_ex);
    file = file_from_path(srch);
    flen = get_strlen(file);
    type = 0;
    gdl = 0;

    /* seperate file from path */
    if (flen) {
	tmpstr_ex[dlen - flen] = ZCHAR;
	dlen -= flen;
    }

    /* check for tilde */
    if (dlen && *tmpstr_ex == '~') {
	safe_strcpy(tmpstr2_ex, getenv("HOME"), PATH_MAX);
	if (dlen > 1)
	    strcat(tmpstr2_ex, &tmpstr_ex[1]);
	safe_strcpy(tmpstr_ex, tmpstr2_ex, PATH_MAX);
    }

    memset(match_ex, 0, PATH_MAX);
    expl_find_file_match(file_type, tmpstr_ex, file, flen, &gdl, &type, 
			 match_ex);
    if (!gdl || gdl < flen)
	expl_beep(srch);
    else {
	len = get_strlen(match_ex);

	if (gdl < len)
	    match_ex[gdl] = ZCHAR;	
	strcat(cmnd, &match_ex[flen]);

	if (gdl != len)
	    expl_beep(srch);
	else {
	    expl_complete = TRUE;
	    if (S_ISDIR(type))
		strcat(cmnd, "/");
	    else if (parenthesis) {
		parenthesis = FALSE;
		strcat(cmnd, "\" ");
	    } else
		strcat(cmnd, " ");
	}
    }
}

int expl_is_cd(char *srch, int slen)
{
    return (slen > 2 && strncmp(srch, "cd ", 3) == 0) ? TRUE : FALSE;
}

void expl_comp_list_insert(char *name, char *repl, int arglen)
{
    if (arglen != 3)
	return;

    if (-1 != dbuf_check_str(lcomp, repl))
	dbuf_add_str(lcomp, name);
    else if (-1 != dbuf_check_str(ecomp, repl))
	dbuf_add_str(ecomp, name);
    else if (-1 != dbuf_check_str(qcomp, repl))
	dbuf_add_str(qcomp, name);
}

void expl_comp_list_remove(char *name)
{
    int ndx;

    ndx = dbuf_check_str(lcomp, name);
    if (ndx != -1) {
	dbuf_remove(lcomp, ndx);
	return;
    }

    ndx = dbuf_check_str(ecomp, name);
    if (ndx != -1) {
	dbuf_remove(ecomp, ndx);
	return;
    }

    ndx = dbuf_check_str(qcomp, name);
    if (ndx != -1) {
	dbuf_remove(qcomp, ndx);
	return;
    }
}

int expl_check_comp_list(DCharBuf *dcb, char *srch, int slen)
{
    char *str;
    int i, plen;

    for (i=0; i<dcb->len; i++) {
	str = dbuf_get(dcb, i);
	plen = get_strlen(str);

	if (slen > plen && strncmp(str, srch, plen) == 0 &&
	    srch[plen] == ' ')
	    return TRUE;
    }

    return FALSE;
}

void expl_list(ListDef *lst, char *srch)
{
    int i, len, gdl, gdl2, clen, cnt;

    if (!lst) {
	expl_beep(srch);
	return;
    }

    clen = get_strlen(srch);
    cnt = gdl = gdl2 = 0;
    memset(match_ex, 0, PATH_MAX);

    for(i = 0; i < lst->dcb->len; i++) {

	cnt ++;
	str_get_id(lst, i, tmpstr_ex);

	if (strncmp(tmpstr_ex, srch, clen) == 0) {
	    if (spackle) 
		expl_spackle(tmpstr_ex, NULL);
	    else if (gdl) {
		gdl2 = expl_find_common_len(tmpstr_ex, match_ex);
		if (gdl2 < gdl) gdl = gdl2;
		expl_exch_if_longer(match_ex, tmpstr_ex);
	    } else {
		safe_strcpy(match_ex, tmpstr_ex, INFO_SIZE);
		gdl = get_strlen(match_ex);
	    }
	}
    }

    if (!gdl || gdl < clen || (!clen && cnt > 1))
	expl_beep(srch);
    else {
	len = get_strlen(match_ex);
	
	if (gdl < len)
	    match_ex[gdl] = ZCHAR;	
	strcat(cmnd, &match_ex[clen]);

	if (gdl != len)
	    expl_beep(srch);
	else {
	    expl_complete = TRUE;

	    if (parenthesis) {
		parenthesis = FALSE;
		strcat(cmnd, "\" ");
	    } else
		strcat(cmnd, " ");
	}
    }
}

void expl_list_name(char *srch)
{
    ListDef *lst;
    int len, gdl, gdl2, clen, cnt;

    clen = get_strlen(srch);
    lst = plist_get_base();
    cnt = gdl = gdl2 = 0;
    memset(match_ex, 0, PATH_MAX);
      
    while (lst) {

	if (strncmp(lst->dcb->name, srch, clen) == 0) {
	    if (spackle) 
		expl_spackle(lst->dcb->name, NULL);
	    else if (gdl) {
		gdl2 = expl_find_common_len(lst->dcb->name, match_ex);
		if (gdl2 < gdl) gdl = gdl2;
		expl_exch_if_longer(match_ex, lst->dcb->name);
	    } else {
		safe_strcpy(match_ex, lst->dcb->name, INFO_SIZE);
		gdl = get_strlen(match_ex);
	    }
	}

	cnt ++;
	lst = lst->next;
    }

    if (!gdl || gdl < clen || (!clen && cnt > 1))
	expl_beep(srch);
    else {
	len = get_strlen(match_ex);
	
	if (gdl < len)
	    match_ex[gdl] = ZCHAR;	
	strcat(cmnd, &match_ex[clen]);

	if (gdl != len)
	    expl_beep(srch);
	else {
	    expl_complete = TRUE;

	    if (parenthesis) {
		parenthesis = FALSE;
		strcat(cmnd, "\" ");
	    } else
		strcat(cmnd, " ");
	}
    }
}

void expl_command(char *srch)
{
    char *cm, *sptxt;
    int i, len, gdl, gdl2, cgdl, clen;
    mode_t type;

    clen = get_strlen(srch);
    gdl = gdl2 = cgdl = 0;
    type = 0;

    if (!clen && !spackle) {
	expl_beep(srch);
	return;
    }

    memset(match_ex, 0, PATH_MAX);
    /* scan built-ins and aliases */
    for (i = 0; i < NUM_COMM + aliases->len; i++) {
	if (i < NUM_COMM) {
	    cm = com[i].name;
	    sptxt = "built-in";
	} else {
	    safe_strcpy(tmpstr_ex, dbuf_get(aliases, i - NUM_COMM), PATH_MAX);
	    cm = strchr(tmpstr_ex, ALIAS_SEPERATOR);
	    if (cm) *cm = ZCHAR;
	    cm = tmpstr_ex;
	    sptxt = "alias";
	}

	if (strncmp(cm, srch, clen) == 0) {
	    if (spackle)
		expl_spackle(cm, sptxt);
	    else if (gdl) {
		gdl2 = expl_find_common_len(cm, match_ex);
		if (gdl2 < gdl) gdl = gdl2;
		expl_exch_if_longer(match_ex, cm);
	    } else {
		safe_strcpy(match_ex, cm, INFO_SIZE);
		gdl = get_strlen(match_ex);
	    }
	}
    }

    /* scan $PATH */
    if (!disable_user_exec && getenv("PATH")) {
	safe_strcpy(tmpstr_ex, getenv("PATH"), PATH_MAX);
	memset(match2_ex, 0, PATH_MAX);
	expl_find_file_match(EXEC_FILE_ONLY, tmpstr_ex, srch, clen, &cgdl, 
			     &type, match2_ex);

	/* use best len and match */
	if (cgdl) {
	    if (gdl) {
		if (cgdl < gdl) gdl = cgdl;
		gdl2 = expl_find_common_len(match_ex, match2_ex);
		if (gdl2 < gdl) gdl = gdl2;
		expl_exch_if_longer(match_ex, match2_ex);
	    } else {
		safe_strcpy(match_ex, match2_ex, PATH_MAX);
		gdl = cgdl;
	    }
	}
    }

    if (!gdl || gdl < clen)
	expl_beep(srch);
    else {
	len = get_strlen(match_ex);

	if (gdl < len)
	    match_ex[gdl] = ZCHAR;
	strcat(cmnd, &match_ex[clen]);

	if (gdl != len)
	    expl_beep(srch);
	else {
	    expl_complete = TRUE;
	    strcat(cmnd, " ");
	}
    }
}

void expl_tab()
{
    char *srch, *bcmd;
    int path, blen;
  
    /* get search token */
    path = expl_is_path(cmnd);
    srch = expl_find_last_space(cmnd);
    unescape_string(srch);

    /* see if user wants spackled */
    if (last_expl == srch)
	spackle = TRUE;

    /* set vars for expl type */
    expl_complete = FALSE;
    bcmd = skip_space(cmnd);
    blen = get_strlen(bcmd);

    /* do the explosion */
    if (srch == cmnd) {
	if (path)
	    expl_file(srch, EXEC_AND_DIRS);
	else
	    expl_command(srch);
    }  else if (expl_is_cd(bcmd, blen))
	expl_file(srch, DIRS_ONLY);
    else if (expl_check_comp_list(lcomp, bcmd, blen))
	expl_list_name(srch);
    else if (expl_check_comp_list(ecomp, bcmd, blen)) {
	
	if (edit_list)
	    expl_list(edit_list, srch);
	else
	    expl_list(NULL, srch);

    } else if (expl_check_comp_list(qcomp, bcmd, blen))
	expl_list(history, srch);
    else
	expl_file(srch, ANY_FILE);

    /* re-escape and set len */
    escape_string(srch);
    comm_len = edit_pos = get_strlen(cmnd);

    /* check if done and last char is escaped-space */
    if (expl_complete && comm_len >= 2 && 
	strcmp(&cmnd[comm_len-2], "\\ ") == 0) {
	strncpy(&cmnd[comm_len-2], " \0", 2);
	edit_pos = --comm_len;
    }

    if (spackle) {
	prompt_now = spackle = FALSE;
	last_expl = NULL;
	if (!outb_empty()) {
	    out_str_n("\n");
	    outb_sort();
	    outb_flush();
	}
    } else
	update_prompt(TRUE);
}
