/***********************************************
 *  extdec.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.          
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* command.c */
extern DCharBuf *args;
extern flag_t prompt_now;

/* listlogic.c */
extern char *strinfo;

/* main.c */
extern int thisprgp;

/* decoder.c */
extern flag_t multiple_audio_procs;

/************************************
 *        Global Variables          *
 ************************************/

flag_t external_paused, external_player;

PlayerDef *player_list, *delayed_external_player;

/************************************
 *        Local variables           *
 ************************************/

static char *COUTPUT[] = { 
    "n/a",
    "Normal",
    "Supressed",
    "Buffered"
};
static char *SIGNALS[] = { 
    "n/a",
    "SIGTERM", 
    "SIGKILL" 
};
static PlayerDef player_list_s = { 
    "Built-in decoder", "n/a",
#ifdef OGG_MODULE
    ".MP3,.MP2,.OGG",
#else
    ".MP3,.MP2",
#endif
    OUTP_NONE, SIGNAL_NONE, (PlayerDef *) NULL
};
static int external_signal;
static pid_t player_pid;

/************************************
 *         Module routines          *
 ************************************/

unsigned int plyr_get_bytec()
{
    PlayerDef *curr;
    unsigned int bytec;

    curr = player_list->next;
    bytec = 0;

    while (curr) {
	bytec += sizeof(PlayerDef);
	bytec += get_strlen(curr->name) + 1;
	bytec += get_strlen(curr->command) + 1;
	bytec += get_strlen(curr->sup_types) + 1;
	curr = curr->next;
    }

    return bytec;
}

void plyr_init()
{
    external_paused = external_player = FALSE;
    external_signal = SIGNAL_TERM;
    player_pid = 0;
    player_list = &player_list_s;
    delayed_external_player = NULL;
}

void plyr_clear()
{
    PlayerDef *curr, *next;

    if (external_player)
	ext_stop();

    curr = player_list->next;

    while (curr) {
	next = curr->next;
	free(curr->name);
	free(curr->command);
	free(curr->sup_types);
	free(curr);
	curr = next;
    }
}

/************************************
 *   External player-list manip     *
 ************************************/

void /* user */ plyr_add_comm()
{
    char *opt, *name, *types, *command;
    int output, signl, ndx, i;
    PlayerDef *curr;

    output = OUTP_NORM;
    signl = SIGNAL_TERM;
    i = ndx = 1;

    while ((opt = dbuf_get(args, i++)))
	if (strcmp(opt, "-q") == 0) {
	    output = OUTP_QUIET;
	    ndx++;
	} else if (strcmp(opt, "-b") == 0) {
	    output = OUTP_BUFFERED;
	    ndx++;
	} else if (strcmp(opt, "-k") == 0) {
	    signl = SIGNAL_KILL;
	    ndx++;
	}

    name = dbuf_get(args, ndx++);
    command = dbuf_get(args, ndx++);
    types = dbuf_get(args, ndx);

    if (!name || !command || !types)
	print_cmnd_help(ADD_PLAYER);
    else {
	curr = player_list;

	while (curr->next)
	    curr = curr->next;

	curr->next = malloc(sizeof(PlayerDef));
	curr = curr->next;

	curr->name = strdup(name);
	curr->command = strdup(command);
	curr->sup_types = strdup(types);
	curr->output = output;
	curr->signal = signl;
	curr->next = NULL;
    }
}

void /* user */ plyr_print_comm()
{
    PlayerDef *curr;

    curr = player_list;

    while (curr) {
	outb_add(MISC_MSG " %s", curr->name);
	outb_add("\tShell command       : %s", curr->command);
	outb_add("\tSupported format(s) : %s", curr->sup_types);
	outb_add("\tConsole output      : %s", COUTPUT[curr->output]);
	outb_add("\tTerminate signal    : %s", SIGNALS[curr->signal]);
	curr = curr->next;
    }

    outb_flush();
}

/************************************
 *     External Player control      *
 ************************************/

void ext_pause()
{
    external_paused = TRUE;
    killpg(player_pid, SIGSTOP);
}

void ext_resume()
{
    external_paused = FALSE;
    killpg(player_pid, SIGCONT);
}

void ext_stop()
{
    int i, sig;

    i = 0;
    alarm(0);
    external_player = external_paused = FALSE;

    if (external_signal == SIGNAL_KILL)
	sig = SIGKILL;
    else
	sig = SIGTERM;

    while (-1 == killpg(player_pid, sig) && errno == ESRCH) {
	sleep(1);
	if (i++ > 1) {
	    dlog_add(ERROR_MSG " Unable to stop external player! [pid=%d]",
		     player_pid);
	    return;
	}
    }
   
    sleep(2); /* SIGCHLD will bust */
    
    if (!multiple_audio_procs)
	usleep(80000);		
}

void ext_handle_death(pid_t pid)
{
    if (external_player && pid == player_pid) {
	alarm(0);
	external_player = external_paused = FALSE;
	event_post(ESTREAM_EVENT);
	if (!multiple_audio_procs)
	    usleep(80000);
	lgc_determine_next();
    }
}

void ext_play(PlayerDef * plyr, char *stream)
{
    char *cm, *cmem, tmp[INFO_SIZE], *arg[INFO_SIZE];
    int i, fd, mypid, prgrp, output, pstdo[2], pstdr[2];

    output = plyr->output;
    cm = cmem = NULL;
    i = 0;

    lgc_set_state(EXTERNAL);
    external_player = TRUE;

    if (dec_close_output()) {
	delayed_external_player = plyr;
	return; /* get out, must wait for audio resources to free up */
    }

    safe_sprintf(strinfo, INFO_SIZE, "Using \"%s\"", plyr->name);
    mysetenv(CURR_INFO, strinfo);
    lgc_clear_duration();

    safe_strcpy(tmp, plyr->command, INFO_SIZE);
    cm = tokenize(tmp, &cmem, ' ');
    while (cm) {
	arg[i++] = cm;
	cm = tokenize(NULL, &cmem, ' ');
    }

    arg[i++] = stream;
    arg[i] = NULL;

    if (output == OUTP_BUFFERED && (pipe(pstdo) || pipe(pstdr))) {
	conb_add(ERROR_MSG " Failed to create player pipes : %s",
		 strerror(errno));
	output = OUTP_NORM;
    }

    player_pid = fork();

    if (player_pid == -1) {
	lgc_set_state(IDLE);
	external_player = FALSE;
	conb_add(ERROR_MSG" Failed to fork() external player \"%s\" : %s",
		 plyr->command, strerror(errno));
    } else if (player_pid != 0) {
	switch (output) {
	case OUTP_NORM:
	case OUTP_QUIET:
	    proc_add(player_pid, J_DECEXT, FALSE, NULL, NULL, *arg);
	    break;
	case OUTP_BUFFERED:
	    proc_add(player_pid, J_DECEXT, TRUE, pstdo, pstdr, *arg);
	    break;
	}

	external_signal = plyr->signal;
	lgc_clear_time();
	alarm(1);
    } else if (player_pid == 0) {
	mypid = getpid();
	prgrp = mypid;
	setpgid(mypid, prgrp);

	signal(SIGINT, SIG_DFL);
	signal(SIGTSTP, SIG_IGN);
	signal(SIGTTIN, SIG_IGN);
	signal(SIGTTOU, SIG_IGN);
	signal(SIGTERM, SIG_DFL);
	signal(SIGCHLD, SIG_DFL);
	signal(SIGPIPE, SIG_DFL);
	signal(SIGALRM, SIG_DFL);
	signal(SIGSTOP, SIG_DFL);
	signal(SIGCONT, SIG_DFL);
	signal(SIGWINCH, SIG_IGN);

	switch (output) {
	case OUTP_QUIET: {
	    fd = open("/dev/null", O_WRONLY);
	    
	    if (fd != -1) {
		dup2(fd, 1);
		dup2(fd, 2);
	    }
	    break;
	}
	case OUTP_BUFFERED:
	    dup2(pstdo[1], 1);
	    dup2(pstdr[1], 2);
	    break;
	}

	usleep(50000);
	execvp(*arg, arg);
	exit(DECODER_EXEC_FAILED);
    }
}
