/***********************************************
 *  info.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.              
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* decoder.c */
extern XA_VolumeInfo vinfo;
extern XA_EqualizerInfo equalizer;
extern XA_TimecodeInfo dinfo, tinfo;
extern EQBandDef eqbands[];
extern char *xa_output;
extern flag_t equalizer_enabled, djmode_enabled, output_muted, output_open;
extern flag_t stream_open, decoder_open, multiple_audio_procs, auto_recover;

/* command.c */
extern CommandDef com[];
extern DCharBuf *args;

/* listlogic.c */
extern ProcessState state;
extern char *strinfo, *strprop, *STATES[], *current, *current_info;
extern flag_t auto_report, tried_gettag;

/* player.c */
extern flag_t external_player;

/* tags.c */
extern flag_t read_file_tags;

/* buff.c */
extern DCharBuf *dlog;

/************************************
 *        Global Variables          *
 ************************************/

char *OnOff[] = { 
    "Disabled", 
    "Enabled" 
};

/************************************
 *        Local Variables           *
 ************************************/

/************************************
 *         Module routines          * 
 ************************************/

unsigned int info_get_bytec()
{
    unsigned int bytec;

    bytec = 0;

    return bytec;
}

void info_init()
{
    print_info_comm();
}

void info_clear()
{
}

/************************************
 * Help, info, and status Functions *
 ************************************/

void /* user */ print_gpl_comm()
{
    outb_add("********************************************");
    outb_add("imp3sh is free software; you can distribute");
    outb_add("it and/or modify it under the terms of the");
    outb_add("GNU General Public License as published by");
    outb_add("the Free Software Foundation; either version");
    outb_add("2 of the License, or (at your option) any");
    outb_add("later version.");
    outb_add(" ");
    outb_add("imp3sh is distributed in the hope that");
    outb_add("it will be useful, but WITHOUT ANY WARRANTY;");
    outb_add("without even the implied warranty of");
    outb_add("MERCHANTABILITY or FITNESS FOR A PARTICULAR");
    outb_add("PURPOSE. See the GNU General Public License");
    outb_add("for more details.");
    outb_add(" ");
    outb_add("You should have recieved a copy of the GNU");
    outb_add("General Public License along with imp3sh; if");
    outb_add("not, write to the Free Software Foundation,");
    outb_add("Inc., 59 Temple Place, Suite 330, Boston MA");
    outb_add("02111-1307 USA");
    outb_add("********************************************");

    outb_flush();
}

void /* user */ print_info_comm()
{
    show_user("*****************************************************");
    show_user("%s %s", NAME, VERSION);
#ifdef OGG_MODULE
    show_user("Interactive MP3/Ogg-Vorbis file and URL player with");
    show_user("limited shell functionality\n");
#else
    show_user("Interactive MP3 file/URL player with limited shell");
    show_user("functionality.\n");
#endif
    show_user("Copyrights (C) %s Eric Kratzer.", DATE);
    show_user("Includes Xaudio software Copyright (C) 1996-2000");
    show_user("         Xaudio Corporation All Rights Reserved.\n");
    show_user("This program comes with ABSOLUTELY NO WARRANTY and");
    show_user("is free software---you are welcome to redistribute it");
    show_user("under certain conditions; type \"%s\" for details.",
	      com[PRINT_GPL].name);
    show_user("\nReport bugs/comments to: kman_can@yahoo.com");
    show_user("*****************************************************");
}

void print_key_help()
{
    outb_add("Next song-------------> Page-Up");
    outb_add("Last song-------------> Page-Down");
    outb_add("Fast-forward----------> Home");
    outb_add("Rewind----------------> End");
    outb_add("Play------------------> F1");
    outb_add("Pause-----------------> F2");
    outb_add("Resume----------------> F3");
    outb_add("Stop------------------> F4");
    outb_add("Increase master vol---> F5");
    outb_add("Decrease master vol---> F6");
    outb_add("Increase pcm vol------> F7");
    outb_add("Decrease pcm vol------> F8");
    outb_add("Pan balance left------> F9");
    outb_add("Pan balance right-----> F10");
    outb_add("Print status----------> F11");
    outb_add("Print list status-----> F12");
}

void print_cmnd_help(int index)
{
    if (index >= 0 && index < NUM_COMM)
	show_user(DEBUG_MSG " %s", com[index].help);
}

void /* user */ print_help_comm()
{
    char *name;
    int i, len;

    name = dbuf_get(args, 1);
    len = get_strlen(name);

    if (len && strcmp(name, "keys") == 0) {
	print_key_help();
	outb_flush();
	return;
    }

    for (i = 0; i < NUM_COMM; i++) {
	if (len && grep_against_args(1, com[i].help) == FALSE)
	    continue;

	outb_add("%s", com[i].help);

	/* seperate comm groups for readability */
	if (!len && (i == EXIT_DECODER      || i == REPORT_TIME_STATE ||
		     i == EDIT_LIST         || i == TOGGLE_READ_TAGS  ||
		     i == SET_INDEX_WEIGHT  || i == LOAD_ALL_LISTS    ||
		     i == RESIZE_QUEUE      || i == PRINT_STYLES      ||
		     i == SET_PREF_F        || i == DJ_MODE           ||
		     i == SET_STAMP_TIML    || i == ADD_PLAYER        ||
		     i == DISABLE_EQ        || i == REMOVE_EVENT      ||
		     i == REMOVE_ALIAS      || i == SEND_JOB_2BG      ||
		     i == CHANGE_DIR        || i == SET_LEARN_WEIGHTS ||
		     i == PRINT_LIST_STATUS || i == MOD_KEY))
	    outb_add(" ");
    }

    if (!len) {
	outb_add(" ");
	print_key_help();
    }

    outb_flush();
}

void report_memuse(char *sub, char *pre)
{
    unsigned int bytec;
    float fbytec;
    char unit;

    bytec = main_get_bytec();
    fbytec = format_bytec(bytec, &unit);

    if (sub)
	outb_add("%.2f %cBytes malloc'ed", fbytec, unit);
    else
	outb_add("%sMemory    : %.2f %cBytes malloc'ed", pre, fbytec, unit);
}

void /* user */ print_status_comm()
{
    char tmp[INFO_SIZE], tmp2[8], *name, *sub, *pre;
    int i, nlen, slen, cur, dur;
    double perc;

    name = dbuf_get(args, 1);
    nlen = get_strlen(name);
    sub = dbuf_get(args, 2);
    slen = get_strlen(sub);

    if (name)
	pre = "";
    else
	pre = "\t";

    if (!name || strncasecmp(name, "stream", nlen) == 0) { 
	if (!name)
	    outb_add(MISC_MSG " Stream");
	if (!sub || strncasecmp(sub, "tag", slen) == 0) {
	    if (read_file_tags && !tried_gettag) {
		tag_make(current, current_info);
		tried_gettag = TRUE;
	    }
	    if (*current_info) {
		if (sub)
		    outb_add("%s", skip_space(current_info));
		else
		    outb_add("%sTag       : %s", pre, 
				  skip_space(current_info));
	    } 
	}

	if (!sub || strncasecmp(sub, "name", slen) == 0) {
		if (sub)
		    outb_add("%s", file_from_path(current));
		else
		    outb_add("%sName      : %s", pre, 
				  file_from_path(current));
	}

	if (!sub || strncasecmp(sub, "info", slen) == 0) {
	    if (sub)
		outb_add("%s%s", strprop, strinfo);
	    else
		outb_add("%sInfo      : %s%s", pre, strprop, strinfo);
	}

	if (state == EXTERNAL) {
	    if (!sub || strncasecmp(sub, "position", slen) == 0) {
		safe_sprintf(tmp, INFO_SIZE, "%.2d:%.2d:%.2d", 
			     tinfo.h, tinfo.m, tinfo.s);
		if (sub)
		    outb_add("%s", tmp);
		else
		    outb_add("%sPosition  : %s", pre, tmp);
	    }
	} else if (stream_open) {
	    dur = (dinfo.h*3600 + dinfo.m*60 + dinfo.s)*100 + dinfo.f;
	    cur = (tinfo.h*3600 + tinfo.m*60 + tinfo.s)*100 + tinfo.f;

	    if (dur)
		perc = 100.0 * ((float) cur / (float) dur);
	    else
		perc = 0.0;

	    if (!sub || strncasecmp(sub, "length", slen) == 0) {
		safe_sprintf(tmp, INFO_SIZE, "%.2d:%.2d:%.2d", 
			     dinfo.h, dinfo.m, dinfo.s);
		if (sub)
		    outb_add("%s", tmp);
		else
		    outb_add("%sLength    : %s", pre, tmp);
	    }
	    if (!sub || strncasecmp(sub, "position", slen) == 0) {
		safe_sprintf(tmp, INFO_SIZE, "%.2d:%.2d:%.2d (%.1f%%)",
			     tinfo.h, tinfo.m, tinfo.s, perc);
		if (sub)
		    outb_add("%s", tmp);
		else
		    outb_add("%sPosition  : %s", pre, tmp);
	    }
	}
    }

    if (!name || strncasecmp(name, "player", nlen) == 0) {
	if (!name)
	    outb_add(MISC_MSG " Player");
	if (!sub || strncasecmp(sub, "state", slen) == 0) {
	    if (sub)
		outb_add("%s", STATES[state]);
	    else
		outb_add("%sState     : %s", pre, STATES[state]);
	}
	if (!sub || strncasecmp(sub, "report", slen) == 0) {
	    if (sub)
		outb_add("%s", OnOff[auto_report]);
	    else
		outb_add("%sReport    : %s", pre, OnOff[auto_report]);
	}
	if (!sub || strncasecmp(sub, "recover", slen) == 0) {
	    if (sub)
		outb_add("%s", OnOff[auto_recover]);
	    else
		outb_add("%sRecover   : %s", pre, OnOff[auto_recover]);
	}
	if (!sub || strncasecmp(sub, "read tags", slen) == 0) {
	    if (sub)
		outb_add("%s", OnOff[read_file_tags]);
	    else
		outb_add("%sRead tags : %s", pre, OnOff[read_file_tags]);
	}
	if (!sub || strncasecmp(sub, "debug log", slen) == 0) {
	    if (dlog->len)
		safe_sprintf(tmp, INFO_SIZE, "%d message(s)", dlog->len/2);
	    else
		safe_strcpy(tmp, "Empty", INFO_SIZE);
	    if (sub)
		outb_add("%s", tmp);
	    else
		outb_add("%sDebug log : %s", pre, tmp);
	}
	if (!sub || strncasecmp(sub, "memory", slen) == 0)
	    report_memuse(sub, pre);
    }

    if (!name || strncasecmp(name, "audio", nlen) == 0) {
	if (!name)
	    outb_add(MISC_MSG " Audio");
	if (!sub || strncasecmp(sub, "output", slen) == 0) {
	    if (state == EXTERNAL) {
		if (multiple_audio_procs && decoder_open)
		    safe_strcpy(tmp, "Shared", INFO_SIZE);
		else if (external_player)
		    safe_strcpy(tmp, "External", INFO_SIZE);
		else 
		    safe_strcpy(tmp, "Free", INFO_SIZE);
	    } else if (decoder_open && output_open) {
		if (output_muted)
		    safe_strcpy(tmp, "Muted", INFO_SIZE);
		else if (xa_output)
		    safe_strcpy(tmp, xa_output, INFO_SIZE);
		else
		    safe_strcpy(tmp, "Default", INFO_SIZE);
	    } else
		safe_strcpy(tmp, "Free", INFO_SIZE);
	    if (sub)
		outb_add("%s", tmp);
	    else
		outb_add("%sOutput    : %s", pre, tmp);
	}

	if (!sub || strncasecmp(sub, "dj mode", slen) == 0) {
	    if (state != EXTERNAL && decoder_open)
		safe_strcpy(tmp, OnOff[djmode_enabled], INFO_SIZE);
	    else
		safe_strcpy(tmp, "Unavailable", INFO_SIZE);
	    if (sub)
		outb_add("%s", tmp);
	    else
		outb_add("%sDJ mode   : %s", pre, tmp);
	}

	if (!sub || strncasecmp(sub, "equalizer", slen) == 0) {
	    if (state != EXTERNAL && decoder_open) {
		if (equalizer_enabled) {
		    *tmp = ZCHAR;
		    for (i = 0; i < MAX_EQ_BANDS; i++) {
			safe_sprintf(tmp2, INFO_SIZE, "%d", 
				     equalizer.left[eqbands[i].low]);
			strcat(tmp, tmp2);
			if (i != (MAX_EQ_BANDS - 1))
			    strcat(tmp, ".");
		    }
		} else
		    safe_strcpy(tmp, "Disabled", INFO_SIZE);
	    } else
		safe_strcpy(tmp, "Unavailable", INFO_SIZE);
	    if (sub)
		outb_add("%s", tmp);
	    else
		outb_add("%sEqualizer : %s", pre, tmp);
	}
	
	if (!sub || strncasecmp(sub, "mixer", slen) == 0) {
	    if (decoder_open && output_open && 
		(state != EXTERNAL || multiple_audio_procs)) {
		safe_sprintf(tmp, INFO_SIZE, "Master.%d Pcm.%d Balance.%d",
			     vinfo.master_level, vinfo.pcm_level, 
			     vinfo.balance);
	    } else
		safe_strcpy(tmp, "Unavailable", INFO_SIZE);
	    if (sub)
		outb_add("%s", tmp);
	    else
		outb_add("%sMixer     : %s", pre, tmp);
	}
    }

    outb_flush();
}

/*
 * Print out only the time played, and if available total length.
 */
void /* user */ print_time_status_comm()
{
  char tmp[32], tmp2[32];

  if (state == EXTERNAL) {
    safe_sprintf(tmp, 32, "T: %.2d:%.2d:%.2d", tinfo.h, tinfo.m, tinfo.s);
    outb_add("%s", tmp);
  } else if (stream_open) {
    safe_sprintf(tmp2, 32, "T: %.2d:%.2d:%.2d", dinfo.h, dinfo.m, dinfo.s);
    safe_sprintf(tmp,  32, "P: %.2d:%.2d:%.2d", tinfo.h, tinfo.m, tinfo.s);
    outb_add("%s, %s", tmp, tmp2);
  }
  
  outb_flush();
}
