/***********************************************
 *  intdec.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.            
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* command.c */
extern CommandDef com[];
extern DCharBuf *args;
extern flag_t prompt_now;

/* player.c */
extern PlayerDef *delayed_external_player;
extern flag_t external_player;

/* listlogic.c */
extern ProcessState state;
extern char *strinfo, *strprop, *current;

/* info.c */
extern char *OnOff[];

/* main.c */
extern flag_t exit_flag, show_time_stamp;

/* proc.c */
extern sighandler_t sigchild_handler;

/************************************
 *        Global Variables          *
 ************************************/

char *AUDDRV[] = { 
    "Single process only",
    "Multiple processes allowed"
};

char *xa_output;
flag_t decoder_open, stream_open, output_open, output_muted, djmode_enabled; 
flag_t equalizer_enabled, multiple_audio_procs, auto_recover;

XA_TimecodeInfo dinfo, tinfo;
XA_VolumeInfo vinfo;
XA_EqualizerInfo equalizer;

EQBandDef eqbands[] = {
    /* band 0 */ {0, 0},
    /* band 1 */ {1, 1},
    /* band 2 */ {2, 2},
    /* band 3 */ {3, 3},
    /* band 4 */ {4, 4},
    /* band 5 */ {5, 5},
    /* band 6 */ {6, 7},
    /* band 7 */ {8, 11},
    /* band 8 */ {12, 19},
    /* band 9 */ {20, 31}
};

/************************************
 *          Local Variables         *
 ************************************/

static char *LAYERS[] = {"invalid", "I", "II", "III"};
static char *MODES[] = {
    "stereo", 
    "joint-stereo", 
    "dual-channel", 
    "single-channel"
};

static flag_t first_startup;
static XA_Control *decoder;
static XA_Message message;
static XA_TimecodeInfo minfo;

/************************************
 *         Module routines          * 
 ************************************/

unsigned int dec_get_bytec()
{
    unsigned int bytec;

    bytec = 0;
    if (xa_output)
	bytec += get_strlen(xa_output) + 1;

    return bytec;
}

void dec_init()
{
    decoder_open = stream_open = output_open = output_muted = FALSE;
    djmode_enabled = equalizer_enabled = multiple_audio_procs = FALSE;
    auto_recover = FALSE;
    first_startup = TRUE;
    decoder = NULL;
    xa_output = NULL;

    memset(&minfo,     0, sizeof(XA_TimecodeInfo));
    memset(&dinfo,     0, sizeof(XA_TimecodeInfo));
    memset(&tinfo,     0, sizeof(XA_TimecodeInfo));
    memset(&equalizer, 0, sizeof(XA_EqualizerInfo));

    dec_open_comm();
}

void dec_clear()
{
    free(xa_output);
}

/************************************
 *     Internal Decoder control     *
 ************************************/

void /* user */ toggle_audio_procs_comm()
{
    multiple_audio_procs = !multiple_audio_procs;
    show_user(STATUS_MSG " Audio driver: %s", AUDDRV[multiple_audio_procs]);
}

void /* user */ toggle_auto_recover_comm()
{
    auto_recover = !auto_recover;
    show_user(STATUS_MSG " Auto recover %s", OnOff[auto_recover]);
}

void dec_stop()
{
    if (decoder_open) {
	control_message_send(decoder, XA_MSG_COMMAND_STOP);
	lgc_set_state(IDLE);
    }
}

void dec_pause()
{
    if (decoder_open) {
	control_message_send(decoder, XA_MSG_COMMAND_PAUSE);
	lgc_set_state(PAUSED);
    }
}

void dec_resume()
{
    if (decoder_open) {
	control_message_send(decoder, XA_MSG_COMMAND_PLAY);
	lgc_set_state(PLAY);
    }
}

void dec_seek(int offset, int scale)
{
    if (decoder_open && stream_open)
	control_message_send(decoder, XA_MSG_COMMAND_SEEK, offset, scale);
}

void dec_set_mixer(int bal, int pcm, int mas)
{
    if (decoder_open && output_open) {
	control_message_send(decoder, XA_MSG_SET_OUTPUT_VOLUME, bal, pcm, mas);
	control_message_send(decoder, XA_MSG_GET_OUTPUT_VOLUME);
    }
}

void dec_play(char *stream)
{
    if (!decoder_open)
	dec_open_comm();
    else {
	dec_close_input();

	if (!output_open)
	    control_message_send(decoder, XA_MSG_COMMAND_OUTPUT_OPEN, 
				 xa_output);

	control_message_send(decoder, XA_MSG_COMMAND_INPUT_OPEN, stream);
	control_message_send(decoder, XA_MSG_COMMAND_PLAY);
    }
}

int dec_get_read_pipe()
{
    return (decoder_open && decoder) ? decoder->command_read_pipe : 0;
}

void dec_close()
{
    if (decoder_open) {
	if (exit_flag)
	    prompt_now = FALSE;

	dec_close_input();
	control_message_send(decoder, XA_MSG_COMMAND_EXIT);
    }
}

void /* user */ dec_open_comm()
{
    if (decoder_open) {
	show_user(DEBUG_MSG " Xaudio decoder is already running");
	return;
    }

    decoder_open = TRUE;
    output_open = stream_open = equalizer_enabled = FALSE;

    signal(SIGINT, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    signal(SIGCHLD, SIG_DFL);
    signal(SIGPIPE, SIG_DFL);
    signal(SIGALRM, SIG_DFL);
    signal(SIGWINCH, SIG_IGN);

    if (player_new((XA_Control **) &decoder, NULL) != XA_SUCCESS)
	decoder_open = FALSE;

    signal(SIGTSTP, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    signal(SIGINT, (sighandler_t) int_sig_handler);
    signal(SIGCHLD, (sighandler_t) sigchild_handler);
    signal(SIGALRM, (sighandler_t) alarm_sig_handler);
    signal(SIGWINCH, (sighandler_t) winch_sig_handler);

    if (!decoder_open)
	dlog_add(ERROR_MSG " XAudio: decoder init failed");
    else {
	control_message_send(decoder, XA_MSG_SET_NOTIFICATION_MASK,
			     NOTIFY_MASK);
	control_message_send(decoder, XA_MSG_SET_INPUT_TIMECODE_GRANULARITY,
			     DEFAULT_GRANULARITY);
	control_message_send(decoder, XA_MSG_SET_CODEC_EQUALIZER, NULL);
	control_message_send(decoder, XA_MSG_GET_CODEC_EQUALIZER);

#ifdef OGG_MODULE
	{
	    char tmpstr[INFO_SIZE];

	    safe_sprintf(tmpstr, INFO_SIZE, "function@%x",
			 (unsigned int) (*vorbis_codec_module_register));
	    control_message_send(decoder, XA_MSG_COMMAND_CODEC_MODULE_REGISTER,
				 tmpstr);
	    dlog_add(STATUS_MSG " XAudio: Ogg-Vorbis codec registered");
	}
#endif
    }
}

void dec_close_input()
{
    if (state == PLAY || state == PAUSED)
	control_message_send(decoder, XA_MSG_COMMAND_STOP);

    if (stream_open)
	control_message_send(decoder, XA_MSG_COMMAND_INPUT_CLOSE);
}

int dec_close_output()
{
    if (decoder_open) {
	dec_close_input();

	if (output_open && !multiple_audio_procs) {
	    control_message_send(decoder, XA_MSG_COMMAND_OUTPUT_CLOSE);
	    return TRUE;
	}
    }

    return FALSE;
}

void dec_seek_static()
{
    char *tok1, *tok2;
    double offset, scale;

    tok1 = dbuf_get(args, 1);
    tok2 = dbuf_get(args, 2);

    if (!tok1 || !tok2) {
	print_cmnd_help(SEEK_FILE);
	return;
    }

    offset = atof(tok1);
    scale = atof(tok2);
    
    offset *= 100.0;
    scale *= 100.0;

    if (offset > scale)
	show_user(ERROR_MSG " Scale must be larger than offset");
    else if (external_player)
	show_user(DEBUG_MSG
		  " Seeking unavailable when running external player");
    else
	dec_seek((long int) offset, (long int) scale);
}

void dec_seek_offset()
{
    char *tok;
    double inc;
    long int dur, cur;

    tok = dbuf_get(args, 1);

    if (!tok) {
	print_cmnd_help(SEEK_FILE);
	return;
    }

    if (*tok == '+')
	inc = atof((tok + 1));
    else
	inc = atof(tok);

    if (external_player)
	show_user(DEBUG_MSG
		  " Seeking unavailable when running external player");
    else {
	dur = (dinfo.h*3600 + dinfo.m*60 + dinfo.s)*100 + dinfo.f;
	cur = (tinfo.h*3600 + tinfo.m*60 + tinfo.s)*100 + tinfo.f;
	cur += (long int) (inc*100.0);

	if (cur < 0)
	    cur = 0;
	else if (cur > dur)
	    cur = dur;

	dec_seek(cur, dur);
    }
}

void /* user */ dec_seek_comm()
{
    if (args->len == 2)
	dec_seek_offset();
    else if (args->len == 3)
	dec_seek_static();
    else
	print_cmnd_help(SEEK_FILE);
}

void /* user */ dec_set_output_comm()
{
    char *outp;

    outp = dbuf_get(args, 1);

    if (!outp)
	print_cmnd_help(SET_XAUDIO_OUTPUT);
    else if (decoder_open) {
	if (xa_output && strcmp(outp, xa_output) == 0)
	    return;

	if (xa_output)
	    free(xa_output);
	xa_output = strdup(outp);
       
	if (state == PLAY || state == PAUSED)
	    lgc_stop_comm();

	if (stream_open)
	    control_message_send(decoder, XA_MSG_COMMAND_INPUT_CLOSE);

	if (output_open)
	    control_message_send(decoder, XA_MSG_COMMAND_OUTPUT_CLOSE);
    }
}

void /* user */ dec_mute_comm()
{
    if (!decoder_open || !output_open)
	show_user(DEBUG_MSG " Xaudio output not available");
    else {
	if (!output_muted)
	    control_message_send(decoder, XA_MSG_COMMAND_OUTPUT_MUTE);
	output_muted = TRUE;
    }
}

void /* user */ dec_unmute_comm()
{
    if (!decoder_open || !output_open)
	show_user(DEBUG_MSG " Xaudio output not available");
    else {
	if (output_muted)
	    control_message_send(decoder, XA_MSG_COMMAND_OUTPUT_UNMUTE);
	output_muted = FALSE;
    }
}

void /* user */ dec_disable_eq_comm()
{
    if (decoder_open) {
	control_message_send(decoder, XA_MSG_SET_CODEC_EQUALIZER, NULL);
	equalizer_enabled = FALSE;
    }
}

void /* user */ dec_enable_eq_comm()
{
    if (decoder_open) {
	control_message_send(decoder, XA_MSG_SET_CODEC_EQUALIZER, &equalizer);
	control_message_send(decoder, XA_MSG_GET_CODEC_EQUALIZER);
	equalizer_enabled = TRUE;
    }
}

void dec_set_eq_band(int band, char value)
{
    int i;

    for (i = eqbands[band].low; i <= eqbands[band].high; i++) {
	equalizer.left[i] = equalizer.right[i] = value;
    }

    dec_enable_eq_comm();
}

void /* user */ dec_print_eq_comm()
{
    int i;

    for (i = 0; i < MAX_EQ_BANDS; i++)
	outb_add(MISC_MSG " Band %d == %d", i, equalizer.left[eqbands[i].low]);

    if (!equalizer_enabled)
	outb_add(DEBUG_MSG " Equalizer is Disabled");

    outb_flush();
}

void /* user */ dec_set_eq_band_comm()
{
    char *band, *val, *bmem, *vmem, *bm, *vm;
    int i, bandi, vali;

    band = dbuf_get(args, 1);
    val = dbuf_get(args, 2);
    bandi = vali = 0;

    if (!band || !val) {
	print_cmnd_help(SET_EQ_BAND);
	return;
    }

    if (strcmp(band, "all") == 0) {
	vali = atoi(val);

	if (vali >= EQ_BAND_VAL_LOW && vali <= EQ_BAND_VAL_HIGH) {
	    for (i = 0; i < XA_DECODER_EQUALIZER_NB_BANDS; i++)
		equalizer.left[i] = equalizer.right[i] = vali;

	    dec_enable_eq_comm();
	} else
	    show_user(DEBUG_MSG " Invalid band val: %d. Range is [%d, %d]",
		      vali, EQ_BAND_VAL_LOW, EQ_BAND_VAL_HIGH);
    } else {
	bmem = vmem = NULL;
	bm = tokenize(band, &bmem, ',');
	vm = tokenize(val, &vmem, ',');

	while (bm && vm) {
	    bandi = atoi(bm);
	    vali = atoi(vm);

	    if (bandi >= 0 && bandi < MAX_EQ_BANDS) {
		if (vali >= EQ_BAND_VAL_LOW && vali <= EQ_BAND_VAL_HIGH)
		    dec_set_eq_band(bandi, vali);
		else
		    show_user(DEBUG_MSG
			      " Invalid band val: %d. Range is [%d, %d]", 
			      vali, EQ_BAND_VAL_LOW, EQ_BAND_VAL_HIGH);
	    } else
		show_user(DEBUG_MSG " Invalid band: %d. Range is [0, %d]",
			  bandi, MAX_EQ_BANDS - 1);

	    bm = tokenize(NULL, &bmem, ',');
	    vm = tokenize(NULL, &vmem, ',');
	}
    }
}

int dec_get_mixer_param(char *param, int saved)
{
    int val;

    if (*param == '+')
	val = saved + atoi((param + 1));
    else if (*param == '-')
	val = saved - atoi((param + 1));
    else
	val = atoi(param);

    if (val > 100)
	val = 100;
    else if (val < 0)
	val = 0;

    return val;
}

void /* user */ dec_set_mixer_comm()
{
    char *param;
    int master, pcm, bal;

    master = pcm = bal = 0;

    if (args->len != 4) {
	print_cmnd_help(SET_MIXER);
	return;
    }

    param = dbuf_get(args, 1);
    if (param)
	master = dec_get_mixer_param(param, vinfo.master_level);

    param = dbuf_get(args, 2);
    if (param)
	pcm = dec_get_mixer_param(param, vinfo.pcm_level);

    param = dbuf_get(args, 3);
    if (param)
	bal = dec_get_mixer_param(param, vinfo.balance);

    dec_set_mixer(bal, pcm, master);
}

void /* user */ dec_toggle_dj_comm()
{
    if (!decoder_open) {
	show_user(DEBUG_MSG " Xaudio decoder unavailable");
	return;
    }

    djmode_enabled = !djmode_enabled;

    if (djmode_enabled)
	control_message_send(decoder, XA_MSG_SET_INPUT_TIMECODE_GRANULARITY,
			     DJ_GRANULARITY);
    else
	control_message_send(decoder, XA_MSG_SET_INPUT_TIMECODE_GRANULARITY,
			     DEFAULT_GRANULARITY);

    show_user(STATUS_MSG " DJ mode %s", OnOff[djmode_enabled]);
}

void dec_get_stream_props()
{
    int i, np;
    char *name, *layer, *mode, level[32];
    XA_Property *pr;
    XA_PropertyValue valu;

    *level = 0;
    layer = mode = NULL;
    np = message.data.properties.nb_properties;
    pr = message.data.properties.properties;

    for(i = 0; i<np; i++) {
	name = (char *)(pr+i)->name;
	valu = (pr+i)->value;

	if (name && *name) {
	    if (strcmp(name, "mpeg/mode") == 0 && 
		valu.integer >= 0 &&
		valu.integer < 4)
		mode = MODES[valu.integer];
	    else if (strcmp(name, "mpeg/layer") == 0 &&
		     valu.integer >= 0 &&
		     valu.integer < 4)		
		layer = LAYERS[valu.integer];
	    else if (strcmp(name, "mpeg/level") == 0)
		safe_sprintf(level, 32, "%d", (int) valu.integer);
	}
    }
    
    if (*level && layer && mode) {
	safe_sprintf(strprop, INFO_SIZE, "MPEG %s Layer %s (%s) ", 
		     level, layer, mode);
	mysetenv(CURR_PROP, strprop);
    }
}

void dec_state_change(unsigned int st)
{
    switch (st) {
    case XA_PLAYER_STATE_PLAYING:
	lgc_set_state(PLAY);
	break;
    case XA_PLAYER_STATE_PAUSED:
	lgc_set_state(PAUSED);
	break;
    case XA_PLAYER_STATE_STOPPED:
	lgc_set_state(IDLE);
	break;
    case XA_PLAYER_STATE_EOS:
	event_post(ESTREAM_EVENT);
	lgc_determine_next();
	break;
    default:
	break;
    }
}

void dec_handle_message()
{
    switch (message.code) {
    case XA_MSG_NOTIFY_READY:
	proc_add(decoder->decoder_pid, J_DECINT, FALSE, NULL, NULL, 
		 "xaudio decoder");

	if (!first_startup && !external_player && state != EXTERNAL && 
	    *current != ZCHAR) {
	    control_message_send(decoder, XA_MSG_COMMAND_INPUT_OPEN, current);
	    control_message_send(decoder, XA_MSG_COMMAND_PLAY);
	}

	first_startup = FALSE;
	break;
    
    case XA_MSG_NOTIFY_EXITED:
	player_delete((void *) decoder);
	equalizer_enabled = decoder_open = stream_open = output_open = FALSE;
	exit_flag = TRUE;
	break;

    case XA_MSG_NOTIFY_CODEC_EQUALIZER:
	if (message.data.equalizer)
	    equalizer = *message.data.equalizer;
	break;

    case XA_MSG_NOTIFY_OUTPUT_STATE:
	switch (message.data.state) {
	case XA_OUTPUT_STATE_OPEN:
	    output_open = TRUE;
	    break;
	case XA_OUTPUT_STATE_CLOSED:
	    output_open = FALSE;
	    if (delayed_external_player) {
		usleep(100000);	/* audio resources should be freed by now...
				   but good to be safe                       */
		ext_play(delayed_external_player, current);
		delayed_external_player = NULL;
	    }
	    break;
	}
	break;

    case XA_MSG_NOTIFY_INPUT_STATE:
	switch (message.data.state) {
	case XA_INPUT_STATE_OPEN:
	    stream_open = TRUE;
	    break;
	case XA_OUTPUT_STATE_CLOSED:
	    stream_open = FALSE;
	    break;
	}
	break;

    case XA_MSG_NOTIFY_PLAYER_STATE:
	dec_state_change(message.data.state);
	break;

    case XA_MSG_NOTIFY_INPUT_TIMECODE:
	if (state != IDLE && state != EXTERNAL) {
	    minfo = tinfo;
	    tinfo = message.data.timecode;
    
	    if (show_time_stamp && 
		((tinfo.h != minfo.h) || (tinfo.m != minfo.m) ||
		 (tinfo.s != minfo.s)))
		update_prompt(TRUE);
	}
	break;

    case XA_MSG_NOTIFY_OUTPUT_BALANCE:
	vinfo.balance = message.data.volume.balance;
	break;

    case XA_MSG_NOTIFY_OUTPUT_PCM_LEVEL:
	vinfo.pcm_level = message.data.volume.pcm_level;
	break;

    case XA_MSG_NOTIFY_OUTPUT_MASTER_LEVEL:
	vinfo.master_level = message.data.volume.master_level;
	break;

    case XA_MSG_NOTIFY_STREAM_DURATION:
	dinfo.h = message.data.duration / 3600;
	dinfo.m = (message.data.duration - (dinfo.h * 3600)) / 60;
	dinfo.s = message.data.duration % 60;
	break;

    case XA_MSG_NOTIFY_STREAM_PARAMETERS:
	safe_sprintf(strinfo, INFO_SIZE, "%dkbps %ldHz",
		     message.data.stream_parameters.bitrate,
		     message.data.stream_parameters.frequency);
	mysetenv(CURR_INFO, strinfo);
	break;

    case XA_MSG_NOTIFY_STREAM_PROPERTIES:
	dec_get_stream_props();
	break;

    case XA_MSG_NOTIFY_ERROR:
	if (XA_ERROR_TIMEOUT != message.data.error.code)
	    dlog_add(ERROR_MSG "Error %d (%s)", message.data.error.code,
		     message.data.error.message);
	break;

    case XA_MSG_NOTIFY_DEBUG:
	dlog_add(DEBUG_MSG " XAudio: %s", message.data.debug.message);
	break;

    default:
	break;
    }
}

void dec_monitor()
{
    int status;

    status = control_message_get((void *) decoder, &message);
    while (status == 1) {
	dec_handle_message();
	status = control_message_get((void *) decoder, &message);
    }

    if (status == XA_ERROR_PIPE_CLOSED) {
	decoder_open = output_open = stream_open = equalizer_enabled = FALSE;

	if (auto_recover) {
	    dlog_add(ERROR_MSG " XAudio: decoder went down unexpectedly!"
		     " Restarting ...");
	    dec_open_comm();
	} else
	    dlog_add(ERROR_MSG " XAudio: decoder went down unexpectedly!"
		     " Try \"%s\"", com[START_DECODER].name);

    } else if (status < 0)
	dlog_add(ERROR_MSG " XAudio: error %d in control_message_get()",
		 status);
}
