/***********************************************
 *  keys.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.               
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* decoder.c */
extern XA_VolumeInfo vinfo;
extern XA_TimecodeInfo tinfo, dinfo;
extern flag_t djmode_enabled;

/* command.c */
extern DCharBuf *args;
extern char *cmnd;
extern int comm_len, edit_pos;

/* shell.c */
extern DCharBuf *chist;
extern flag_t ignore_hist_input;

/* expl.c */
extern char *last_expl;

/* listlogic.c */
extern flag_t auto_report;

/* event.c */
extern char *YesNo[];

/************************************
 *        Global Variables          *
 ************************************/

flag_t key_press;

DJKeyDef djkeys[] = {
    /* F1/F2   */ {FALSE, 0, 0, NULL},
    /* F3/F4   */ {FALSE, 0, 0, NULL},
    /* F5/F6   */ {FALSE, 0, 0, NULL},
    /* F7/F8   */ {FALSE, 0, 0, NULL},
    /* F9/F10  */ {FALSE, 0, 0, NULL},
    /* F11/F12 */ {FALSE, 0, 0, NULL}
};

TermCtrl t_ctrl[] = {
    {FALSE, "Clear screen",         "cl", NULL, NULL},
    {FALSE, "Clear to bottom",      "cd", NULL, NULL},		
    {FALSE, "Clear to EOL",         "ce", NULL, NULL},
    {FALSE, "Non-destructive left", "nd", NULL, NULL},
    {FALSE, "Cursor up",            "up", NULL, NULL},
    {FALSE, "Carriage return",      "cr", NULL, NULL},

    {FALSE, "Up arrow",             "ku", NULL, NULL},
    {FALSE, "Down arrow",           "kd", NULL, NULL},
    {FALSE, "Right arrow",          "kr", NULL, NULL},
    {FALSE, "Left arrow",           "kl", NULL, NULL},

    {FALSE, "Home",                 "kh", NULL, NULL},
    {FALSE, "End",                  "@7", NULL, NULL},
    {FALSE, "Insert",               "kI", NULL, NULL},
    {FALSE, "Delete",               "kD", NULL, NULL},
    {FALSE, "Page up",              "kP", NULL, NULL},
    {FALSE, "Page down",            "kN", NULL, NULL},

    {FALSE, "F1",                   "k1", NULL, NULL}, 
    {FALSE, "F2",                   "k2", NULL, NULL},
    {FALSE, "F3",                   "k3", NULL, NULL},
    {FALSE, "F4",                   "k4", NULL, NULL},
    {FALSE, "F5",                   "k5", NULL, NULL},
    {FALSE, "F6",                   "k6", NULL, NULL},
    {FALSE, "F7",                   "k7", NULL, NULL},
    {FALSE, "F8",                   "k8", NULL, NULL},
    {FALSE, "F9",                   "k9", NULL, NULL},
    {FALSE, "F10",                  "k;", NULL, NULL},
    {FALSE, "F11",                  "F1", NULL, NULL},
    {FALSE, "F12",                  "F2", NULL, NULL},  

    /* raw arrow and F1-F4 key handling.
       some terminals don't output the byte sequences reported
       for them in the terminfo database (i.e., Eterm, kterm).
    */
    {FALSE, "Alt. up arrow",        "",   "\x1b[A",   NULL},
    {FALSE, "Alt. down arrow",      "",   "\x1b[B",   NULL},
    {FALSE, "Alt. right arrow",     "",   "\x1b[C",   NULL},
    {FALSE, "Alt. left arrow",      "",   "\x1b[D",   NULL},

    {FALSE, "Alt. F1 key",          "",   "\x1b[11~", NULL},
    {FALSE, "Alt. F2 key",          "",   "\x1b[12~", NULL},
    {FALSE, "Alt. F3 key",          "",   "\x1b[13~", NULL},
    {FALSE, "Alt. F4 key",          "",   "\x1b[14~", NULL}
    
};

/************************************
 *        Local Variables           *
 ************************************/

static char tmpstr_ky[COMM_SIZE];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int key_get_bytec()
{
    unsigned int bytec;
    int i;

    bytec = 0;

    for(i = HOME_KEY; i <= F12_KEY; i++)
	if (t_ctrl[i].user_def)
	    bytec += get_strlen(t_ctrl[i].user_def) + 1;

    return bytec;
}

void key_init_m()
{
    key_press = FALSE;
    memset(tmpstr_ky, 0, COMM_SIZE);
}

void key_clear_m()
{
    int i;

    for(i = HOME_KEY; i <= F12_KEY; i++)
	free(t_ctrl[i].user_def);
}

/************************************
 *          Special keys            *
 ************************************/

void key_up_arrow()
{
    if (!outb_empty())
	return;

    if (chist->pos) {
	safe_strcpy(cmnd, dbuf_get(chist, --chist->pos), COMM_SIZE);
	comm_len = edit_pos = get_strlen(cmnd);
	last_expl = NULL;
	update_prompt(TRUE);
    } else
	com_beep();
}

void key_down_arrow()
{
    if (!outb_empty())
	return;

    if (chist->pos < chist->len - 1) {
	safe_strcpy(cmnd, dbuf_get(chist, ++chist->pos), COMM_SIZE);
	comm_len = edit_pos = get_strlen(cmnd);
	last_expl = NULL;
	update_prompt(TRUE);
    } else if (chist->pos == chist->len - 1) {
	com_info_clear();
	update_prompt(TRUE);
    } else
	com_beep();
}

void key_right_arrow()
{
    if (edit_pos < comm_len && outb_empty()) {
	edit_pos++;
	last_expl = NULL;
	update_prompt(TRUE);
    } else
	com_beep();
}

void key_left_arrow()
{
    if (edit_pos > 0 && outb_empty()) {
	edit_pos--;
	last_expl = NULL;
	update_prompt(TRUE);
    } else
	com_beep();
}

void key_fastf()
{
    long int dur, cur;

    dur = (dinfo.h*3600 + dinfo.m*60 + dinfo.s)*100 + dinfo.f;
    cur = (tinfo.h*3600 + tinfo.m*60 + tinfo.s)*100 + tinfo.f;
    cur += (100 * SEEK_JUMP);

    if (cur == dur)
	return;
    else if (cur > dur)
	cur = dur;

    dec_seek(cur, dur);
}

void key_rewind()
{
    long int dur, cur;

    dur = (dinfo.h*3600 + dinfo.m*60 + dinfo.s)*100 + dinfo.f;
    cur = (tinfo.h*3600 + tinfo.m*60 + tinfo.s)*100 + tinfo.f;
    cur -= (100 * SEEK_JUMP);

    if (cur == 0)
	return;
    else if (cur < 0)
	cur = 0;

    dec_seek(cur, dur);
}

void key_inc_masvol()
{
    int mas;

    mas = vinfo.master_level + VOLUME_JUMP;

    if (vinfo.master_level == 100)
	return;
    else if (mas > 100)
	mas = 100;

    dec_set_mixer(vinfo.balance, vinfo.pcm_level, mas);
}

void key_dec_masvol()
{
    int mas;

    mas = vinfo.master_level - VOLUME_JUMP;

    if (vinfo.master_level == 0)
	return;
    else if (mas < 0)
	mas = 0;

    dec_set_mixer(vinfo.balance, vinfo.pcm_level, mas);
}

void key_inc_pcmvol()
{
    int pcm;

    pcm = vinfo.pcm_level + VOLUME_JUMP;

    if (vinfo.pcm_level == 100)
	return;
    else if (pcm > 100)
	pcm = 100;

    dec_set_mixer(vinfo.balance, pcm, vinfo.master_level);
}

void key_dec_pcmvol()
{
    int pcm;

    pcm = vinfo.pcm_level - VOLUME_JUMP;

    if (vinfo.pcm_level == 0)
	return;
    else if (pcm < 0)
	pcm = 0;

    dec_set_mixer(vinfo.balance, pcm, vinfo.master_level);
}

void key_bal_right()
{
    int bal;

    bal = vinfo.balance + BALANCE_JUMP;

    if (vinfo.balance == 100)
	return;
    else if (bal > 100)
	bal = 100;

    dec_set_mixer(bal, vinfo.pcm_level, vinfo.master_level);
}

void key_bal_left()
{
    int bal;

    bal = vinfo.balance - BALANCE_JUMP;

    if (vinfo.balance == 0)
	return;
    else if (bal < 0)
	bal = 0;

    dec_set_mixer(bal, vinfo.pcm_level, vinfo.master_level);
}

void key_dj_get_pos(int ndx)
{
    if (ndx >= 0 && ndx < MAX_DJ_KEYS) {
	djkeys[ndx].set = TRUE;
	djkeys[ndx].offset = (tinfo.h*3600 + tinfo.m*60 + tinfo.s)*100;

	/* By the time we get the key here, some number of milliseconds have
	   passed since what the user heard was actually being decoded, so we
	   must adjust <tinfo>. The number of mulliseconds to adjust will vary
	   depending on audio driver buffering and system processor speed. The
	   number here is just what I came up with by ear on my system.
	*/
	if (djkeys[ndx].offset > 35)
	    djkeys[ndx].offset += (tinfo.f - 35);
	else
	    djkeys[ndx].offset += tinfo.f;

	djkeys[ndx].range = ((dinfo.h*3600 + dinfo.m*60 + dinfo.s)*100 + 
			     dinfo.f);
    }
}

void key_dj_set_pos(int ndx)
{
    if (ndx >= 0 && ndx < MAX_DJ_KEYS && djkeys[ndx].set)
	dec_seek(djkeys[ndx].offset, djkeys[ndx].range);
}

void key_handle_sp(int spec)
{
    switch (spec) {
    case UP_ARROW:
    case UP_ARROW_2:
	key_up_arrow();
	break;
    case DOWN_ARROW:
    case DOWN_ARROW_2:
	key_down_arrow();
	break;
    case RIGHT_ARROW:
    case RIGHT_ARROW_2:
	key_right_arrow();
	break;
    case LEFT_ARROW:
    case LEFT_ARROW_2:
	key_left_arrow();
	break;
    case DELETE_KEY:
	com_delete();
	break;
    case PAGEUP_KEY:
	lgc_next_comm();
	break;
    case PAGEDOWN_KEY:
	lgc_last_comm();
	break;
    case HOME_KEY:
	key_fastf();
	break;
    case END_KEY:
	key_rewind();
	break;
    case F1_KEY:
    case F1_KEY_2:
	dbuf_clear(args);
	lgc_play_comm();
	break;
    case F2_KEY:
    case F2_KEY_2:
	lgc_pause_comm();
	break;
    case F3_KEY:
    case F3_KEY_2:
	lgc_resume_comm();
	break;
    case F4_KEY:
    case F4_KEY_2:
	lgc_stop_comm();
	break;
    case F5_KEY:
	key_inc_masvol();
	break;
    case F6_KEY:
	key_dec_masvol();
	break;
    case F7_KEY:
	key_inc_pcmvol();
	break;
    case F8_KEY:
	key_dec_pcmvol();
	break;
    case F9_KEY:
	key_bal_left();
	break;
    case F10_KEY:
	key_bal_right();
	break;
    case F11_KEY:
	if (outb_empty()) {
	    out_str_n("\n");
	    dbuf_clear(args);
	    print_status_comm();
	    update_prompt(FALSE);
	}
	break;
    case F12_KEY:
	if (outb_empty()) {
	    out_str_n("\n");
	    dbuf_clear(args);
	    plist_status_comm();
	    update_prompt(FALSE);
	}
	break;
    default:
	break;
    }
}

void key_handle_dj(int ndx)
{
    if (ndx >= F1_KEY && ndx <= F6_KEY)
	key_dj_get_pos(ndx - F1_KEY);
    else if (ndx >= F7_KEY && ndx <= F12_KEY)
	key_dj_set_pos(ndx - F7_KEY);
}

void key_handle_user(int ndx)
{
    if (t_ctrl[ndx].newl && have_tty())
	out_str_n("\n");

    safe_strcpy(tmpstr_ky, t_ctrl[ndx].user_def, COMM_SIZE);
    ignore_hist_input = TRUE;
    com_handle(tmpstr_ky);
    ignore_hist_input = FALSE;

    if (t_ctrl[ndx].newl && have_tty())
	update_prompt(FALSE);
}

int key_check_sp(char *ins)
{
    int i, ln;

    for (i = UP_ARROW; i < NUM_TERM_CTRL; i++) {
	ln = get_strlen(t_ctrl[i].string);

	if (ln && strncmp(ins, t_ctrl[i].string, ln) == 0) {
	    key_press = TRUE;

	    if (djmode_enabled && i >= F1_KEY && i <= F12_KEY)
		key_handle_dj(i);
	    else if (t_ctrl[i].user_def)
		key_handle_user(i);
	    else
		key_handle_sp(i);

	    key_press = FALSE;
	    return TRUE;
	}
    }

    return FALSE;
}

/*** user defined keys ***/

void /* user */ key_user_show_comm()
{
    int i, cnt, tot;

    cnt = 0;
    tot = F12_KEY - HOME_KEY + 1;

    for(i = HOME_KEY; i <= F12_KEY; i++)
	if (t_ctrl[i].user_def) {
	    cnt ++;
	    outb_add(MISC_MSG " \"%s\" key", t_ctrl[i].desc);
	    outb_add("\tCommand : %s", t_ctrl[i].user_def);
	    outb_add("\tNewline : %s", YesNo[t_ctrl[i].newl]);
	}

    outb_add(STATUS_MSG " %d modified key(s), %d using default behavior:",
	     cnt, tot-cnt);

    for(i = HOME_KEY; i <= F12_KEY; i++)
	if (!t_ctrl[i].user_def)
	    outb_add("\t%s", t_ctrl[i].desc);

    outb_flush();
}

void /* user */ key_user_mod_comm()
{
    int i, ndx, klen;
    flag_t newl;
    char *key, *act, *opt, *cm;

    ndx = 1;
    opt = dbuf_get(args, 1);
    newl = FALSE;

    if (opt && strcmp(opt, "-n") == 0) {
	newl = TRUE;
	ndx ++;
    }
	
    key = dbuf_get(args, ndx++);
    act = dbuf_get(args, ndx++);

    if (!key || !act) {
	print_cmnd_help(MOD_KEY);
	return;
    }

    if (*key == '\"') {
	key ++;
	opt = strchr(key, '\"');
	if (opt) 
	    *opt = '\0';
    }

    klen = get_strlen(key);

    for(i = HOME_KEY; i <= F12_KEY; i++) 
	if (strncasecmp(key, t_ctrl[i].desc, klen) == 0)
	    break;

    if (i > F12_KEY) {
	show_user(DEBUG_MSG " No such key found: \"%s\"", key);
	return;
    }

    *tmpstr_ky = 0;
    cm = act;
    while (cm) {
	strcat(tmpstr_ky, cm);
	cm = dbuf_get(args, ndx++);
	if (cm)
	    strcat(tmpstr_ky, " ");
    }

    free(t_ctrl[i].user_def);
    t_ctrl[i].newl = newl;
    if (strcasecmp(tmpstr_ky, "default") == 0)
	t_ctrl[i].user_def = NULL;
    else
	t_ctrl[i].user_def = strdup(tmpstr_ky);
}
