/***********************************************
 *  list.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.          
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* queue.c */
extern ListDef *history;
extern flag_t inside_queue;

/* command.c */
extern DCharBuf *args;

/* listlogic.c */
extern ProcessState state;
extern char *PLAYMODES[];
extern flag_t user_stream;

/* player.c */
extern flag_t external_player;

/* keys.c */
extern flag_t key_press;

/************************************
 *        Global Variables          *
 ************************************/

char *SEARCHPRF[] = { 
    "Tag, then Filename",
    "Tag, then Path/Filename",
    "Path/Filename",
    "Filename"
};

flag_t ignore_stream_weight;
ListDef *active_list, *edit_list;

/************************************
 *         Local Variables          *
 ************************************/

static ListDef *list_list;
static char tmpstr_pl[COMM_SIZE];

/************************************
 *         Module routines          *
 ************************************/

unsigned int plist_get_bytec()
{
    ListDef *lst;
    unsigned int bytec;

    lst = list_list;
    bytec = 0;

    while (lst) {
	bytec += dbuf_get_bytec(lst->dcb);
	bytec += sizeof(ListDef);
	lst = lst->next;
    }

    return bytec;
}

void plist_init()
{
    ignore_stream_weight = FALSE;
    list_list = NULL;
    plist_add("default");
    active_list = edit_list = plist_get("default");
    plist_update_env();

    memset(tmpstr_pl,  0, COMM_SIZE);
}

void plist_clear()
{
    ListDef *lst, *tmp;

    lst = list_list;
    tmp = NULL;

    while (lst) {
	tmp = lst;
	lst = lst->next;

	dbuf_free(tmp->dcb);
	free(tmp);
    }

    list_list = edit_list = active_list = NULL;
}

/************************************
 *     List Manip (high-level)      *
 ************************************/

ListDef *plist_get_base()
{
    return list_list;
}

ListDef *plist_get(char *name)
{				/* get list */
    ListDef *lst;

    lst = list_list;

    if (name) {
	while (lst)
	    if (strcmp(lst->dcb->name, name) == 0)
		return lst;
	    else
		lst = lst->next;
    }

    return NULL;
}

int plist_add(char *name)
{				/* add list */
    ListDef *curr, *prev;

    curr = list_list;
    prev = NULL;

    while (curr) {
	if (strcmp(curr->dcb->name, name) == 0)
	    return FALSE;
	prev = curr;
	curr = curr->next;
    }

    if (!prev) {
	list_list = malloc(sizeof(ListDef));
	prev = list_list;
    } else {
	prev->next = malloc(sizeof(ListDef));
	prev = prev->next;
    }

    prev->dcb = dbuf_init(name, DCHAR_GROW, LIST_SIZE, LIST_SIZE);
    prev->next = NULL;
    prev->playmode = ONCE_MODE;
    prev->listview = TAG_FILE_PREF;
    prev->global_weight = 0;
    prev->play_weight = 1;
    prev->skip_weight = 1;

    return TRUE;
}

int plist_remove(char *name)
{				/* remove */
    ListDef *curr, *prev;

    curr = list_list;
    prev = NULL;

    while (curr) {
	if (strcmp(curr->dcb->name, name) == 0)
	    break;
	prev = curr;
	curr = curr->next;
    }

    if (curr) {
	if (prev)
	    prev->next = curr->next;
	else
	    list_list = curr->next;

	dbuf_free(curr->dcb);
	free(curr);
	return TRUE;
    }

    return FALSE;
}

void plist_verify(ListDef *lst)
{
    int i;

    if (lst) {
	i = lst->play_weight;
	if (!i)
	    lst->play_weight = 1;
	else if (i < WEIGHT_MIN)
	    lst->play_weight = WEIGHT_MIN;
	else if (i > WEIGHT_MAX)
	    lst->play_weight = WEIGHT_MAX;

	i = lst->skip_weight;
	if (!i)
	    lst->skip_weight = 1;
	else if (i < WEIGHT_MIN)
	    lst->skip_weight = WEIGHT_MIN;
	else if (i > WEIGHT_MAX)
	    lst->skip_weight = WEIGHT_MAX;

	dbuf_verify(lst->dcb);

	lst->global_weight = 0;
	for (i = 0; i < lst->dcb->len; i++)
	    lst->global_weight += lst->dcb->buf[i].info;
    }
}

void plist_verify_all()
{
    ListDef *lst;

    lst = list_list;

    while (lst) {
	plist_verify(lst);
	lst = lst->next;
    }
}

void plist_check_weight_state()
{
    /* don't weight current pos in new active list 
       on song switch                              */
    if (!user_stream && !queue_in() && !inside_queue &&
	(state == PLAY || (state == EXTERNAL && external_player)))
	ignore_stream_weight = TRUE;
}

void /* user */ plist_use_comm()
{
    ListDef *tmpl;
    char *name;

    name = dbuf_get(args, 1);

    if (!name)
	print_cmnd_help(USE_LIST);
    else if (!list_list)
	show_user(DEBUG_MSG " No lists currently defined");
    else {
	tmpl = plist_get(name);

	if (tmpl) {
	    if (active_list != tmpl)
		plist_check_weight_state();

	    active_list = edit_list = tmpl;
	    plist_update_env();
	    show_user(STATUS_MSG " Active & edit lists set to \"%s\"", name);
	} else
	    show_user(DEBUG_MSG " \"%s\" list not found", name);
    }
}

void /* user */ plist_prev_comm()
{
    ListDef *prev, *temp;

    if (!active_list) {
	if (list_list)
	    active_list = list_list;
	else {
	    show_user(DEBUG_MSG " No lists currently defined");
	    return;
	}
    }

    prev = NULL;
    temp = active_list->next;

    while (temp != active_list) {

	if (temp == NULL)
	    temp = list_list;
	else {
	    prev = temp;
	    temp = temp->next;
	}

    }

    if (prev && active_list != prev) {
	active_list = prev;
	plist_update_env();
	if (key_press)
	    update_prompt(TRUE);
    }
}

void /* user */ plist_next_comm()
{
    ListDef *new_list;

    if (!active_list) {
	if (list_list)
	    active_list = list_list;
	else {
	    show_user(DEBUG_MSG " No lists currently defined");
	    return;
	}
    }

    if (active_list->next)
	new_list = active_list->next;
    else if (list_list)
	new_list = list_list;
    else {
	show_user(ERROR_MSG " Bad list pointer");
	return;
    }

    if (active_list != new_list)
	plist_check_weight_state();

    if (active_list != new_list) {
	active_list = new_list;
	plist_update_env();
	if (key_press)
	    update_prompt(TRUE);
    }
}

void /* user */ plist_edit_comm()
{
    ListDef *tmpl;
    char *name;

    name = dbuf_get(args, 1);

    if (!name)
	print_cmnd_help(EDIT_LIST);
    else if (!list_list)
	show_user(DEBUG_MSG " No lists currently defined");
    else {
	tmpl = plist_get(name);

	if (tmpl) {
	    edit_list = tmpl;
	    plist_update_env();
	    show_user(STATUS_MSG " Edit list set to \"%s\"", name);
	} else
	    show_user(DEBUG_MSG " \"%s\" list not found", name);
    }
}

void /* user */ plist_add_comm()
{
    ListDef *lst;
    char *name;

    name = dbuf_get(args, 1);

    if (!name)
	print_cmnd_help(ADD_LIST);
    else {
	if (!strcmp(name, history->dcb->name))
	    show_user(ERROR_MSG " Cannot use name of special list : \"%s\"", 
		      name);
	else if (plist_add(name) && (lst = plist_get(name))) {
	    edit_list = lst;
	    if (!active_list)
		active_list = lst;
	    plist_update_env();
	    show_user(STATUS_MSG " \"%s\" list created successfully", name);
	} else
	    show_user(DEBUG_MSG " \"%s\" list already exists", name);
    }
}

void /* user */ plist_reset_comm()
{
    ListDef *lst;
    char *name;

    name = dbuf_get(args, 1);

    if (!name) {
	print_cmnd_help(RESET_LIST);
	return;
    }

    if (strcmp(name, "all") == 0) {
	wght_clear_all();
	show_user(STATUS_MSG " All lists were reset");
    } else if (!(lst = plist_get(name)))
	show_user(DEBUG_MSG " \"%s\" list not found", name);
    else {
	wght_clear(lst);
	show_user(STATUS_MSG " \"%s\" was reset", name);
    }
}

void /* user */ plist_rename_comm()
{
    ListDef *lst;
    char *name, *repl;

    name = dbuf_get(args, 1);
    repl = dbuf_get(args, 2);

    if (!name || !repl)
	print_cmnd_help(RENAME_LIST);
    else if (!(lst = plist_get(name)))
	show_user(DEBUG_MSG " \"%s\" list not found", name);
    else {
	free(lst->dcb->name);
	lst->dcb->name = strdup(repl);
	if (lst == active_list || lst == edit_list)
	    plist_update_env();
	show_user(STATUS_MSG " \"%s\" list renamed to \"%s\"", name, repl);
    }
}

void /* user */ plist_remove_comm()
{
    ListDef *lst;
    char *name;
    flag_t curr, edit;

    name = dbuf_get(args, 1);
    curr = edit = FALSE;

    if (!name)
	print_cmnd_help(REMOVE_LIST);
    else if (strcmp(name, "all") == 0) {
	plist_clear();
	plist_update_env();
	show_user(STATUS_MSG " Removed all lists");
    } else if (!(lst = plist_get(name)))
	show_user(DEBUG_MSG " \"%s\" list not found", name);
    else {
	if (lst == active_list)
	    curr = TRUE;
	if (lst == edit_list)
	    edit = TRUE;

	plist_remove(name);
	show_user(STATUS_MSG " Removed \"%s\" list", name);

	if (curr) {
	    active_list = list_list;
	    if (active_list)
		show_user(STATUS_MSG " Active list is now \"%s\"",
			  active_list->dcb->name);
	    else
		show_user(STATUS_MSG
			  " No lists available: Active list undefined");
	}

	if (edit) {
	    edit_list = list_list;
	    if (edit_list)
		show_user(STATUS_MSG " Edit list is now \"%s\"",
			  edit_list->dcb->name);
	    else
		show_user(STATUS_MSG
			  " No lists available: Edit list undefined");
	}

	if (curr || edit)
	    plist_update_env();
    }
}

void plist_print(ListDef * lst, int pos, int verbose)
{
    char *pat, *loc, *hedr;
    PlayerDef *plyr;
    DCharBuf *dcb;
    DCBNode *buf;
    int i, len, lpos;

    pat = dbuf_get(args, pos);
    loc = NULL;
    plyr = NULL;
    dcb = lst->dcb;
    buf = lst->dcb->buf;
    lpos = lst->dcb->pos;
    len = lst->dcb->len;

    if (!len) {
	show_user(DEBUG_MSG " \"%s\" list is empty", lst->dcb->name);
	return;
    }

    for (i = 0; i < len; i++) {
	str_get_id(lst, i, tmpstr_pl);

	if (pat && grep_against_args(pos, tmpstr_pl) == FALSE)
	    continue;

	if (i == lpos)
	    hedr = PNTR_MSG;
	else
	    hedr = LIST_MSG;

	outb_add("%s %d> %s", hedr, i, tmpstr_pl);

	if (verbose) {
	    loc = str_get(dcb, i, STR_FILE);
	    plyr = lgc_check_format(loc);
	    outb_add("%s\tWeight   : %d", hedr, buf[i].info);
	    if (plyr)
		outb_add("%s\tPlayer   : %s", hedr, plyr->name);
	    else
		outb_add("%s\t  No player defined for this file type!", hedr);
	    outb_add("%s\tLocation : %s", hedr, loc);
	}
    }

    outb_flush();
}

void /* user */ plist_print_comm()
{
    char *opt;
    int ver, pos;

    opt = dbuf_get(args, 1);
    ver = FALSE;
    pos = 1;

    if (opt && strcmp(opt, "-l") == 0) {
	ver = TRUE;
	pos++;
    }

    if (edit_list)
	plist_print(edit_list, pos, ver);
    else
	show_user(DEBUG_MSG " Edit list undefined");
}

void /* user */ plist_status_comm()
{
    ListDef *lst;
    DCharBuf *dcb;
    char *name, *hedr, unit;
    int i, cnt, skp, len, wght, plsts;
    float prc, prc2, fbytec;

    name = dbuf_get(args, 1);
    lst = plist_get_base();
    len = get_strlen(name);
    plsts = 0;

    if (!len) {
	if (history->dcb->len) {
	    outb_add(MISC_MSG " %s", history->dcb->name);
	    outb_add("\tEntry(s)  : %d/[%d]", history->dcb->len,
		     history->dcb->size);
	    outb_add("\tPosition  : %d <%s>", history->dcb->pos,
		     file_from_path(str_get(history->dcb,
					    history->dcb->pos,
					    STR_FILE)));
	} else
	    outb_add(DEBUG_MSG " Queue is empty");
    }

    if (!lst)
	outb_add(DEBUG_MSG " No lists defined");
    else {
	while (lst) {
	    dcb = lst->dcb;

	    if (len && strcmp(dcb->name, name)) {
		lst = lst->next;
		continue;
	    }

	    plsts++;

	    if (active_list == lst)
		hedr = PNTR_MSG;
	    else
		hedr = MISC_MSG;

	    if (edit_list == lst)
		outb_add("%s *%s", hedr, dcb->name);
	    else
		outb_add("%s %s", hedr, dcb->name);

	    outb_add("\tPlay mode : %s", PLAYMODES[lst->playmode]);
	    outb_add("\tList view : %s", SEARCHPRF[lst->listview]);

	    fbytec = format_bytec(dbuf_get_bytec(dcb), &unit);
	    outb_add("\tMemory    : %.2f %cbytes", fbytec, unit);

	    if (lst->dcb->len) {
		outb_add("\tEntry(s)  : %d", dcb->len);
		outb_add("\tPosition  : %d <%s>", dcb->pos,
			 file_from_path(str_get(dcb, dcb->pos, STR_FILE)));

		cnt = skp = 0;
		prc = prc2 = 0.0;

		if (lst->playmode == LEARN_MODE) {
		    wght = lgc_list_weight(lst);

		    for (i = 0; i < dcb->len; i++)
			if (dcb->buf[i].info >= wght)
			    cnt++;

		    if (dcb->len) {
			prc = ((float) cnt / (float) dcb->len) * 100.0;
			prc2 = 100.0 - prc;
		    }

		    outb_add("\tStatus    : %.1f%% liked, %.1f%% dis-liked",
			     prc, prc2);
		    outb_add("\tWeights   : play.%d skip.%d global.%d",
			     lst->play_weight, lst->skip_weight, wght);
		} else {
		    for (i = 0; i < dcb->len; i++)
			if (dcb->buf[i].info > 0)
			    cnt++;
			else if (dcb->buf[i].info < 0)
			    skp++;

		    if (dcb->len) {
			prc = ((float) cnt / (float) dcb->len) * 100.0;
			prc2 = ((float) skp / (float) dcb->len) * 100.0;
		    }

		    outb_add("\tStatus    : %.1f%% played, %.1f%% skipped",
			     prc, prc2);
		}
	    } else
		outb_add("\tList is empty");

	    lst = lst->next;
	}

	if (!plsts && len)
	    outb_add(DEBUG_MSG " \"%s\" list not found", name);
    }

    outb_flush();
}

/************************************
 *       Environment updating       *
 ************************************/

void plist_update_env()
{
    ListDef *curr;
    char tmp[32];

    if (queue_in() || inside_queue)
	curr = history;
    else
	curr = active_list;

    if (curr) {
	mysetenv(CURR_ACTIVE_SOURCE, curr->dcb->name);
	mysetenv(CURR_ACTIVE_MODE, PLAYMODES[curr->playmode]);
	mysetenv(CURR_ACTIVE_VIEW, SEARCHPRF[curr->listview]);
	safe_sprintf(tmp, 32, "%d", curr->dcb->len);
	mysetenv(CURR_ACTIVE_LEN, tmp);
	safe_sprintf(tmp, 32, "%d", curr->dcb->pos);
	mysetenv(CURR_ACTIVE_INDEX, tmp);
    } else {
	mysetenv(CURR_ACTIVE_SOURCE, NULL);
	mysetenv(CURR_ACTIVE_MODE, NULL);
	mysetenv(CURR_ACTIVE_VIEW, NULL);
	mysetenv(CURR_ACTIVE_LEN, NULL);
	mysetenv(CURR_ACTIVE_INDEX, NULL);
    }

    if (edit_list) {
	mysetenv(CURR_EDIT_SOURCE, edit_list->dcb->name);
	mysetenv(CURR_EDIT_MODE, PLAYMODES[edit_list->playmode]);
	mysetenv(CURR_EDIT_VIEW, SEARCHPRF[edit_list->listview]);
	safe_sprintf(tmp, 32, "%d", edit_list->dcb->len);
	mysetenv(CURR_EDIT_LEN, tmp);
	safe_sprintf(tmp, 32, "%d", edit_list->dcb->pos);
	mysetenv(CURR_EDIT_INDEX, tmp);
    } else {
	mysetenv(CURR_EDIT_SOURCE, NULL);
	mysetenv(CURR_EDIT_MODE, NULL);
	mysetenv(CURR_EDIT_VIEW, NULL);
	mysetenv(CURR_EDIT_LEN, NULL);
	mysetenv(CURR_EDIT_INDEX, NULL);
    }
}
