/**********************************************
 *  list.h                                                          
 *                                            
 *  Copyright 2001 Eric Kratzer.               
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 *
 *                                            
 **********************************************/

#ifndef LIST_H
#define LIST_H

#define CURR_STATE         "IMP3_STATE"
#define CURR_STREAM        "IMP3_STREAM"
#define CURR_INFO          "IMP3_INFO"
#define CURR_PROP          "IMP3_PROP"

#define CURR_ACTIVE_SOURCE "IMP3_ACTIVE_SOURCE"
#define CURR_ACTIVE_LEN    "IMP3_ACTIVE_LEN"
#define CURR_ACTIVE_INDEX  "IMP3_ACTIVE_INDEX"
#define CURR_ACTIVE_MODE   "IMP3_ACTIVE_MODE"
#define CURR_ACTIVE_VIEW   "IMP3_ACTIVE_VIEW"

#define CURR_EDIT_SOURCE   "IMP3_EDIT_SOURCE"
#define CURR_EDIT_LEN      "IMP3_EDIT_LEN"
#define CURR_EDIT_INDEX    "IMP3_EDIT_INDEX"
#define CURR_EDIT_MODE     "IMP3_EDIT_MODE"
#define CURR_EDIT_VIEW     "IMP3_EDIT_VIEW"

enum {
    ONCE_MODE,
    REPEAT_MODE,
    TRANDOM_MODE,
    SRANDOM_MODE,
    LEARN_MODE,
    QUEUE_MODE
};

enum {
    TAG_FILE_PREF,
    TAG_PATH_PREF,
    PATH_FILE_PREF,
    FILE_ONLY_PREF
};

typedef enum {
    INIT,
    IDLE,
    PLAY,
    PAUSED,
    EXTERNAL
} ProcessState;

typedef struct ListDef_s {
    int               playmode;
    int               listview;
    int               global_weight;
    int               play_weight;
    int               skip_weight;
    DCharBuf         *dcb;
    struct ListDef_s *next;
} ListDef;

/* list.c */
void         plist_init            ();
void         plist_clear           ();
unsigned int plist_get_bytec       ();
void         plist_update_env      ();
void         plist_active_env      (ListDef *);
ListDef     *plist_get_base        ();
ListDef     *plist_get             (char *);
int          plist_add             (char *);
int          plist_remove          ();
void         plist_verify          (ListDef *);
void         plist_verify_all      ();
void         plist_rename_comm     ();	/* user */
void         plist_reset_comm      ();	/* user */
void         plist_use_comm        ();	/* user */
void         plist_prev_comm       ();	/* user */
void         plist_next_comm       ();	/* user */
void         plist_edit_comm       ();	/* user */
void         plist_add_comm        ();  /* user */
void         plist_remove_comm     ();	/* user */
void         plist_print_comm      ();  /* user */
void         plist_status_comm     ();  /* user */

/* listmanip.c */
void         listm_init            ();
void         listm_clear           ();
unsigned int listm_get_bytec       ();
int          plist_compare         (const void *, const void *);
void         plist_sort_comm       ();	/* user */
void         plist_add_file_comm   ();	/* user */
void         plist_play_index_comm ();	/* user */
void         plist_move_comm       ();	/* user */
void         plist_delete_comm     ();	/* user */
void         plist_copy_comm       ();	/* user */
void         plist_add_dir_comm    ();	/* user */
void         plist_set_mode_comm   ();	/* user */
void         plist_set_view_comm   ();	/* user */
void         plist_set_weights_comm();	/* user */

/* listfile.c */
void         listf_init            ();
void         listf_clear           ();
unsigned int listf_get_bytec       ();
void         plist_save_all_comm   ();	/* user */
void         plist_load_all_comm   ();	/* user */
void         plist_save_comm       ();  /* user */
void         plist_load_comm       ();	/* user */

/* listlogic.c */
void         lgc_init              ();
void         lgc_clear             ();
unsigned int lgc_get_bytec         ();
void         lgc_set_state         (ProcessState);
PlayerDef   *lgc_check_format      (char *);
int          lgc_play_stream       (char *);
void         lgc_report_name       (char *);
void         lgc_play_list         (ListDef *, int);
void         lgc_determine_next    ();
int          lgc_list_weight       (ListDef *);
void         lgc_clear_time        ();
void         lgc_clear_duration    ();
void         lgc_clear_info        ();
void         lgc_play_comm         ();  /* user */
void         lgc_resume_comm       ();  /* user */
void         lgc_stop_comm         ();  /* user */
void         lgc_pause_comm        ();  /* user */
void         lgc_next_comm         ();  /* user */
void         lgc_last_comm         ();  /* user */
void         toggle_auto_rprt_comm ();  /* user */

#endif
