/***********************************************
 *  listfile.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.          
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* list.c */
extern ListDef *active_list, *edit_list, *history;
extern char *SEARCHPRF[];

/* command.c */
extern DCharBuf *args;

/* listlogic.c */
extern char *PLAYMODES[];

/************************************
 *        Global Variables          *
 ************************************/

/************************************
 *        Local variables           *
 ************************************/

static char tmpstr_lf[COMM_SIZE], tmpstr2_lf[COMM_SIZE];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int listf_get_bytec()
{
    unsigned int bytec;

    bytec = 0;

    return bytec;
}

void listf_init()
{
    memset(tmpstr_lf,  0, COMM_SIZE);
    memset(tmpstr2_lf, 0, COMM_SIZE);
}

void listf_clear()
{
}

/************************************
 *       Playlist file manip        *
 ************************************/

void plist_write_header(FILE * fd)
{
    time_t tim;
    char *time_ascii;

    tim = time(NULL);
    time_ascii = ctime(&tim);
    remove_newline(time_ascii);

    fprintf(fd, "# Playlist created by %s %s on %s\n\n", NAME, VERSION, 
	    time_ascii);
}

void plist_write_stream(FILE * fd, int ndx, DCharBuf * lst)
{
    char *tag, *file;
    int len;

    tag = str_get(lst, ndx, STR_TAG);
    file = str_get(lst, ndx, STR_FILE);

    if (tag) {
	len = (file - tag) - TAG_CODE_LEN + 1;
	if (len < (COMM_SIZE - TAG_CODE_LEN - 1)) {
	    strncpy(tmpstr_lf, tag, len);
	    tmpstr_lf[len - 1] = ZCHAR;
	} else
	    *tmpstr_lf = ZCHAR;
    } else
	*tmpstr_lf = ZCHAR;

    fprintf(fd, "%s %d %s\n", TAG_HEADER, lst->buf[ndx].info, tmpstr_lf);
    fprintf(fd, "%s\n", file);
}

void plist_write_info(FILE * fd, ListDef * lst)
{
    char view[32];

    if (!fd || !lst)
	return;

    switch (lst->listview) {
    case TAG_FILE_PREF:
	safe_strcpy(view, "viewTF", 32);
	break;
    case TAG_PATH_PREF:
	safe_strcpy(view, "viewTPF", 32);
	break;
    case PATH_FILE_PREF:
	safe_strcpy(view, "viewPF", 32);
	break;
    case FILE_ONLY_PREF:
	safe_strcpy(view, "viewF", 32);
	break;
    default:
	safe_strcpy(view, "viewTF", 32);
    }

    if (is_any_space(lst->dcb->name))
	fprintf(fd, "%s \"%s\" %s %s %d %d %d\n", LIST_INFO_TAG,
		lst->dcb->name, PLAYMODES[lst->playmode], view,
		lst->dcb->pos, lst->play_weight, lst->skip_weight);
    else
	fprintf(fd, "%s %s %s %s %d %d %d\n", LIST_INFO_TAG,
		lst->dcb->name, PLAYMODES[lst->playmode], view,
		lst->dcb->pos, lst->play_weight, lst->skip_weight);
}

void plist_save(char *file, ListDef * lst)
{
    FILE *fd;
    DCharBuf *dcb;
    int i, len;

    dcb = lst->dcb;
    len = dcb->len;

    if (!len)
	show_user(DEBUG_MSG " \"%s\" list is empty", dcb->name);
    else if (!(fd = myfopen(file, "w+")))
	show_user(ERROR_MSG " Could not open \"%s\" : %s", file, 
		  strerror(errno));
    else {
	plist_write_header(fd);

	for (i = 0; i < len; i++)
	    plist_write_stream(fd, i, dcb);

	fclose(fd);
	show_user(STATUS_MSG" Saved %d entry(s) from \"%s\" list to \"%s\"",
		  len, dcb->name, file);
    }
}

void /* user */ plist_save_comm()
{
    char *file;

    file = dbuf_get(args, 1);

    if (!file)
	print_cmnd_help(SAVE_LIST);
    else if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else
	plist_save(file, edit_list);
}

void /* user */ plist_save_all_comm()
{
    ListDef *lst;
    DCharBuf *dcb;
    int i, len, entries, nlists;
    char *file;
    FILE *fd;

    file = dbuf_get(args, 1);
    entries = nlists = 0;

    if (!file)
	print_cmnd_help(SAVE_ALL_LISTS);
    else if (!plist_get_base())
	show_user(DEBUG_MSG " No lists defined");
    else if (!(fd = myfopen(file, "w+")))
	show_user(ERROR_MSG " Failed file open of \"%s\" : %s", file, 
		  strerror(errno));
    else {
	plist_write_header(fd);
	lst = plist_get_base();

	while (lst) {
	    plist_write_info(fd, lst);
	    dcb = lst->dcb;
	    len = dcb->len;
	    nlists++;

	    for (i = 0; i < len; i++) {
		plist_write_stream(fd, i, dcb);
		entries++;
	    }

	    lst = lst->next;
	}

	if (history && history->dcb && history->dcb->len) {
	    plist_write_info(fd, history);
	    dcb = history->dcb;
	    len = dcb->len;
	    nlists ++;

	    for(i = 0; i < len; i++) {
		plist_write_stream(fd, i, dcb);
		entries++;
	    }
	}

	fclose(fd);
	show_user(STATUS_MSG " Saved %d entry(s) from %d list(s) to \"%s\"",
		  entries, nlists, file_from_path(file));
    }
}

int plist_get_mode(char *mode)
{
    if (strcmp(mode, PLAYMODES[ONCE_MODE]) == 0)
	return ONCE_MODE;
    else if (strcmp(mode, PLAYMODES[REPEAT_MODE]) == 0)
	return REPEAT_MODE;
    else if (strcmp(mode, PLAYMODES[TRANDOM_MODE]) == 0)
	return TRANDOM_MODE;
    else if (strcmp(mode, PLAYMODES[SRANDOM_MODE]) == 0)
	return SRANDOM_MODE;
    else if (strcmp(mode, PLAYMODES[LEARN_MODE]) == 0)
	return LEARN_MODE;
    else if (strcmp(mode, PLAYMODES[QUEUE_MODE]) == 0)
	return QUEUE_MODE;
    else
	return ONCE_MODE;
}

int plist_get_view(char *mode)
{
    if (strcmp(mode, "viewTF") == 0)
	return TAG_FILE_PREF;
    else if (strcmp(mode, "viewTPF") == 0)
	return TAG_PATH_PREF;
    else if (strcmp(mode, "viewPF") == 0)
	return PATH_FILE_PREF;
    else if (strcmp(mode, "viewF") == 0)
	return FILE_ONLY_PREF;
    else
	return TAG_FILE_PREF;;
}

void /* user */ plist_load_all_comm()
{
    ListDef *lst;
    FILE *fd;
    int entries, nlists, alists, pnfo, pos, plw, skw;
    char *cm, *cmem, *file, *info, *name, *mode, *view;

    file = dbuf_get(args, 1);
    entries = nlists = alists = 0;
    lst = NULL;

    if (!file)
	print_cmnd_help(LOAD_ALL_LISTS);
    else if (!(fd = myfopen(file, "r")))
	show_user(ERROR_MSG " Failed file open of \"%s\" : %s", file, 
		  strerror(errno));
    else {
	while (!feof(fd)) {
	    get_fline(fd, tmpstr_lf, COMM_SIZE);

	    if (*tmpstr_lf) {
		info = grep_str(LIST_INFO_TAG, tmpstr_lf);

		if (info) {
		    cm = cmem = NULL;
		    cm = tokenize(info, &cmem, ' ');
		    if (cm) name = cm;
		    else name = NULL;
		    cm = tokenize(NULL, &cmem, ' ');
		    if (cm) mode = cm;
		    else mode = NULL;
		    cm = tokenize(NULL, &cmem, ' ');
		    if (cm) view = cm;
		    else view = NULL;
		    cm = tokenize(NULL, &cmem, ' ');
		    if (cm) pos = atoi(cm);
		    else pos = 0;
		    cm = tokenize(NULL, &cmem, ' ');
		    if (cm) plw = atoi(cm);
		    else plw = 0;
		    cm = tokenize(NULL, &cmem, ' ');
		    if (cm) skw = atoi(cm);
		    else skw = 0;

		    if (name) {
			if (!strcmp(name, history->dcb->name)) 
			    lst = history;
			else {
			    lst = plist_get(name);
			    if (!lst) {
				plist_add(name);
				lst = plist_get(name);
				if (!active_list)
				    active_list = lst;
				if (!edit_list)
				    edit_list = lst;
				alists++;
			    }
			}
			nlists++;
		    }

		    if (lst) {
			lst->playmode = plist_get_mode(mode);
			lst->listview = plist_get_view(view);
			lst->dcb->pos = pos;
			lst->play_weight = plw;
			lst->skip_weight = skw;
		    }

		    continue;
		}

		info = grep_str(TAG_HEADER, tmpstr_lf);

		if (info) {
		    pnfo = 0;
		    memset(tmpstr2_lf, 0, COMM_SIZE);
		    sscanf(tmpstr_lf, TAG_HEADER " %d %1024c", &pnfo, 
			   tmpstr2_lf);
		    if (!feof(fd)) {
			get_fline(fd, tmpstr_lf, COMM_SIZE);
			if (*tmpstr2_lf)
			    strcat(tmpstr2_lf, TAG_CODE);
			strcat(tmpstr2_lf, tmpstr_lf);
			if (lst) {
			    str_add(lst->dcb, tmpstr2_lf, (short)pnfo, 
				    TRUE, FALSE);
			    entries++;
			}
		    }
		} else if (lst && *tmpstr_lf != '#') {
		    str_add(lst->dcb, tmpstr_lf, 0, FALSE, FALSE);
		    entries++;
		}
	    }
	}

	if (!nlists)
	    show_user(DEBUG_MSG " No lists defined in \"%s\"", file);
	else {
	    plist_verify_all();
	    plist_update_env();
	    if (alists)
		show_user(STATUS_MSG " Created %d new list(s)", alists);
	    show_user(STATUS_MSG " Added %d entry(s) to %d list(s)", entries,
		      nlists);
	}

	fclose(fd);
    }
}


void plist_load(char *file, ListDef * lst)
{
    DCharBuf *dcb;
    char *tag;
    int len, info;
    FILE *fd;

    fd = myfopen(file, "r");
    dcb = lst->dcb;
    len = dcb->len;
    info = 0;

    if (!fd)
	show_user(ERROR_MSG " Failed file open of \"%s\" : %s", file, 
		  strerror(errno));
    else {
	while (!feof(fd)) {
	    get_fline(fd, tmpstr_lf, COMM_SIZE);

	    if (*tmpstr_lf) {
		tag = grep_str(TAG_HEADER, tmpstr_lf);
		if (tag) {
		    memset(tmpstr2_lf, 0, COMM_SIZE);
		    sscanf(tmpstr_lf,
			   TAG_HEADER " %d %1024c", &info, tmpstr2_lf);
		    if (!feof(fd)) {
			get_fline(fd, tmpstr_lf, COMM_SIZE);
			if (*tmpstr2_lf)
			    strcat(tmpstr2_lf, TAG_CODE);
			strcat(tmpstr2_lf, tmpstr_lf);
			str_add(dcb, tmpstr2_lf, (short) info, TRUE, FALSE);
		    }
		} else if (*tmpstr_lf != '#')
		    str_add(dcb, tmpstr_lf, 0, FALSE, FALSE);
	    }
	}

	fclose(fd);
	show_user(STATUS_MSG " Added %d entry(s) to \"%s\" list",
		  dcb->len - len, dcb->name);
    }
}

void /* user */ plist_load_comm()
{
    char *file;

    file = dbuf_get(args, 1);

    if (!file)
	print_cmnd_help(LOAD_LIST);
    else if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else {
	plist_load(file, edit_list);
	plist_verify(edit_list);
	plist_update_env();
    }
}
