/***********************************************
 *  listlogic.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.          
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* decoder.c */
extern XA_TimecodeInfo tinfo, dinfo;

/* command.c */
extern DCharBuf *args;
extern flag_t running_builtin;

/* queue.c */
extern ListDef *history;
extern flag_t inside_queue;

/* list.c */
extern ListDef *active_list;
extern flag_t ignore_stream_weight;

/* player.c */
extern PlayerDef *player_list;
extern flag_t external_player, external_paused;

/* keys.c */
extern DJKeyDef djkeys[];
extern flag_t key_press;

/* shell.c */
extern char *c_path;

/* info.c */
extern char *OnOff[];

/* main.c */
extern flag_t show_time_stamp, piped_io_flag;

/* tags.c */
extern flag_t read_file_tags;

/************************************
 *        Global Variables          *
 ************************************/

char *PLAYMODES[] = { 
    "Once",
    "Repeat",
    "TrueRandom",
    "SmartRandom",
    "Learn",
    "Queued"
};

char *STATES[] = { 
    "Init", 
    "Idle", 
    "Play", 
    "Paused", 
    "External" 
};

char *current, *current_info, *last_stream, *strinfo, *strprop;
flag_t user_stream, auto_report, tried_gettag;
ProcessState state;

/************************************
 *        Local variables           *
 ************************************/

static char tmpstr_ll[COMM_SIZE];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int lgc_get_bytec()
{
    unsigned int bytec;

    bytec = 0;
    bytec += PATH_MAX*2 + COMM_SIZE + INFO_SIZE*2;

    return bytec;
}

void lgc_init()
{
    struct timeval tv;

    state = IDLE;
    user_stream = auto_report = tried_gettag = FALSE;
    current = malloc(PATH_MAX);
    last_stream = malloc(PATH_MAX);
    current_info = malloc(COMM_SIZE);
    strinfo = malloc(INFO_SIZE);
    strprop = malloc(INFO_SIZE);

    memset(current,      0, PATH_MAX);
    memset(last_stream,  0, PATH_MAX);
    memset(current_info, 0, COMM_SIZE);
    memset(strinfo,      0, INFO_SIZE);
    memset(strprop,      0, INFO_SIZE);
    memset(tmpstr_ll,    0, COMM_SIZE);

    gettimeofday(&tv, NULL);
    srandom((unsigned int) tv.tv_usec);
}

void lgc_clear()
{
    free(current);
    free(current_info);
    free(last_stream);
    free(strinfo);
    free(strprop);
}

/************************************
 *        Playlist logic            *
 ************************************/

void /* user */ toggle_auto_rprt_comm()
{
    auto_report = !auto_report;
    show_user(STATUS_MSG " Auto report %s", OnOff[auto_report]);
}

void lgc_set_state(ProcessState s)
{
    if (external_player && s == IDLE)
	state = EXTERNAL;
    else
	state = s;

    mysetenv(CURR_STATE, STATES[state]);
}

void lgc_clear_time()
{
    int i;

    for (i = 0; i < MAX_DJ_KEYS; i++)
	djkeys[i].set = FALSE;

    memset(&tinfo, 0, sizeof(XA_TimecodeInfo));
}

void lgc_clear_duration()
{
    memset(&dinfo, 0, sizeof(XA_TimecodeInfo));
}

void lgc_clear_info()
{
    lgc_clear_time();
    tried_gettag = FALSE;

    if (*current != ZCHAR)
	safe_strcpy(last_stream, current, PATH_MAX);

    memset(current, 0, PATH_MAX);
    memset(current_info, 0, COMM_SIZE);
    memset(strinfo, 0, INFO_SIZE);
    memset(strprop, 0, INFO_SIZE);

    mysetenv(CURR_INFO, NULL);
    mysetenv(CURR_PROP, NULL);
    mysetenv(CURR_STREAM, NULL);
}

int lgc_active()
{
    switch (state) {
    case PLAY:
    case PAUSED:
	return TRUE;
    case EXTERNAL:
	if (external_player)
	    return TRUE;
    default:
	break;
    }

    return FALSE;
}

int lgc_paused()
{
    switch (state) {
    case PAUSED:
	return TRUE;
    case EXTERNAL:
	if (external_paused)
	    return TRUE;
    default:
	break;
    }

    return FALSE;
}

int lgc_idle()
{
    switch (state) {
    case IDLE:
	return TRUE;
    case EXTERNAL:
	if (!external_player)
	    return TRUE;
    default:
	break;
    }

    return FALSE;
}

int lgc_need_key_update()
{
    if (key_press && (auto_report || !show_time_stamp || state == EXTERNAL))
	return TRUE;
    else
	return FALSE;
}

int lgc_need_tty_update()
{
    if (have_tty() && (auto_report || !show_time_stamp || state == EXTERNAL))
	return TRUE;
    else
	return FALSE;
}

void lgc_report_name(char *stream)
{
    char *what;

    if (!stream)
	return;

    safe_strcpy(current, stream, PATH_MAX);
    mysetenv(CURR_STREAM, current);

    if (!auto_report)
	return;

    if (read_file_tags && !tried_gettag) {
	tag_make(current, current_info);
	tried_gettag = TRUE;
    }

    if (*current_info == ZCHAR)
	what = file_from_path(current);
    else
	what = skip_space(current_info);
	
    if ((key_press || !running_builtin) && have_tty() && !piped_io_flag)
	out_str_n("\n");

    show_user(REPORT_MSG " %s", what);
}

int lgc_verify(char *stream)
{
    struct stat sstat;

    if (!stream)
	return FALSE;

    if (is_url(stream))
	return TRUE;

    if (stat(stream, &sstat)) {
	conb_add(ERROR_MSG " Error reading file \"%s\" : %s", stream, 
		 strerror(errno));
	return FALSE;
    } else if (sstat.st_mode & S_IFDIR) {
	conb_add(ERROR_MSG " \"%s\" is a directory", stream);
	return FALSE;
    }

    return TRUE;
}

PlayerDef *lgc_check_format(char *name)
{
    char tmptype[INFO_SIZE];
    PlayerDef *curr;
    int len, slen;
    char *cmem, *cm;

    curr = player_list;
    len = get_strlen(name);
    cmem = cm = NULL;

    if (!len)
	return NULL;
	
    while (curr) {
	safe_strcpy(tmptype, curr->sup_types, INFO_SIZE);
	cm = tokenize(tmptype, &cmem, ',');

	while (cm) {
	    slen = get_strlen(cm);
	    
	    if (len >= slen && strcasecmp(cm, &name[len - slen]) == 0)
		return curr;

	    cm = tokenize(NULL, &cmem, ',');
	}
	curr = curr->next;
    }

    return NULL;
}

int lgc_play_stream(char *stream)
{
    PlayerDef *plyr;

    if (!lgc_verify(stream))
	return FALSE;

    if (!(plyr = lgc_check_format(stream))) {
	show_user(ERROR_MSG " No suitable player defined for: \"%s\"", stream);
	return FALSE;
    }

    if (external_player) {
	if (external_paused)
	    ext_resume();
	ext_stop();
    }

    lgc_clear_info();
    lgc_report_name(stream);
    event_post(SSTREAM_EVENT);

    if (plyr == player_list)
	dec_play(stream);
    else
	ext_play(plyr, stream);

    return TRUE;
}

void lgc_play_queue(int pos)
{
    if (!history->dcb->len)
	return;

    inside_queue = TRUE;
    user_stream = FALSE;

    lgc_play_stream(str_get(history->dcb, pos, STR_FILE));
    plist_update_env();
}

void lgc_play_list(ListDef * lst, int queue)
{
    DCharBuf *dcb;

    if (!lst)
	return;

    dcb = lst->dcb;
    inside_queue = user_stream = FALSE;

    if (lgc_play_stream(str_get(dcb, dcb->pos, STR_FILE)) && queue)
	queue_add(dbuf_get(dcb, dcb->pos));
    plist_update_env();
}

void /* user */ lgc_play_comm()
{
    char *stream;

    stream = dbuf_get(args, 1);

    if (stream) {
	if (*stream != '/' && !is_url(stream)) {
	    safe_sprintf(tmpstr_ll, COMM_SIZE, "%s/%s", c_path, stream);
	    stream = tmpstr_ll;
	}

	if (lgc_play_stream(stream)) {
	    queue_add(stream);
	    user_stream = TRUE;
	}
    } else if (lgc_paused())
	lgc_resume_comm();
    else if (lgc_idle()) {
	if (queue_in() || inside_queue)
	    lgc_play_queue(history->dcb->pos);
	else if (active_list && active_list->dcb->len) {
	    if (history->dcb->len)
		lgc_play_list(active_list, FALSE);
	    else
		lgc_play_list(active_list, TRUE);
	}

	if (!show_time_stamp && key_press && auto_report)
	    update_prompt(FALSE);
    }
}

void /* user */ lgc_resume_comm()
{
    if (state == EXTERNAL && external_player)
	ext_resume();
    else
	dec_resume();
}

void /* user */ lgc_stop_comm()
{
    if (state == EXTERNAL && external_player)
	ext_stop();
    else
	dec_stop();

    ignore_stream_weight = FALSE;
    lgc_clear_time();

    if (key_press)
	update_prompt(TRUE);
}

void /* user */ lgc_pause_comm()
{
    if (state == EXTERNAL && external_player)
	ext_pause();
    else
	dec_pause();
}

int lgc_list_weight(ListDef * lst)
{
    if (lst && lst->dcb->len)
	return (lst->global_weight / lst->dcb->len);
    else
	return 0;
}

int lgc_final_learned(DCBNode * buf, int len, int weight)
{
    int i;

    for (i = 0; i < len; i++)
	if (buf[i].info >= (short) weight)
	    return i;

    return -1;
}

int lgc_final_srandom(DCBNode * buf, int len)
{
    int i;

    for (i = 0; i < len; i++)
	if (buf[i].info == 0)
	    return i;

    return -1;
}

int lgc_random(int len)
{
    return (int) (((float) len) * ((float) random() / (float) RAND_MAX));
}

int lgc_learned(ListDef * lst)
{
    DCharBuf *dcb;
    DCBNode *buf;
    int len, weight, ndx, loops, tries;

    dcb = lst->dcb;
    buf = dcb->buf;
    len = dcb->len;
    ndx = lgc_random(len);
    loops = tries = 0;
    weight = lgc_list_weight(lst);

    while (buf[ndx].info < (short) weight) {
	ndx = lgc_random(len);

	if (++tries > len) {
	    ndx = lgc_final_learned(buf, len, weight);
	    tries = 0;
	    loops++;

	    if (ndx == -1) {
		if (loops > 1)
		    wght_clear(lst);
		else
		    wght_clear_good(lst, weight);

		ndx = lgc_random(len);
	    }
	}
    }

    return ndx;
}

int lgc_srandom(ListDef * lst)
{
    DCharBuf *dcb;
    DCBNode *buf;
    int len, ndx, loops, tries;

    dcb = lst->dcb;
    buf = dcb->buf;
    len = dcb->len;
    ndx = lgc_random(len);
    loops = tries = 0;

    while (buf[ndx].info) {
	ndx = lgc_random(len);

	if (++tries > len) {
	    ndx = lgc_final_srandom(buf, len);
	    tries = 0;
	    loops++;

	    if (ndx == -1) {
		if (loops > 1)
		    wght_clear(lst);
		else
		    wght_clear_good(lst, 1);

		event_post(LIST_EVENT);
		ndx = lgc_random(len);
	    }
	}
    }

    return ndx;
}

int lgc_next()
{
    int len, next, pmode;

    len = active_list->dcb->len;
    next = active_list->dcb->pos;
    pmode = active_list->playmode;

    switch (pmode) {
    case LEARN_MODE:
	next = lgc_learned(active_list);
	break;
    case SRANDOM_MODE:
	next = lgc_srandom(active_list);
	break;
    case TRANDOM_MODE:
	next = lgc_random(len);
	break;
    default:
	if (next < (len - 1))
	    next++;
	else {
	    event_post(LIST_EVENT);
	    if (pmode == REPEAT_MODE)
		next = 0;
	    else
		next = -1;
	}
	break;
    }

    return next;
}

void /* user */ lgc_next_comm()
{
    int next, ique;

    next = 0;
    ique = FALSE;

    if (lgc_active()) {
	if (queue_in()) {
	    lgc_play_queue(++history->dcb->pos);

	    if (lgc_need_key_update())
		update_prompt(TRUE);

	    return;
	}

	if (inside_queue) {
	    ique = TRUE;
	    inside_queue = FALSE;
	    event_post(QUEUE_EVENT);
	}

	if (active_list && active_list->dcb->len) {
	    if (!user_stream && !ique)
		wght_remove(active_list);

	    if (-1 == (next = lgc_next()))
		return;

	    active_list->dcb->pos = next;
	    lgc_play_list(active_list, TRUE);

	    if (lgc_need_key_update())
		update_prompt(TRUE);
	}

    } else if (lgc_idle() && queue_in()) {
	inside_queue = TRUE;
	history->dcb->pos++;

	if (key_press)
	    update_prompt(TRUE);
    }
}

void /* user */ lgc_last_comm()
{
    if (!history->dcb->len || !history->dcb->pos)
	return;

    if (lgc_active()) {
	if (!user_stream && !inside_queue)
	    wght_remove(active_list);
	lgc_play_queue(--history->dcb->pos);

	if (lgc_need_key_update())
	    update_prompt(TRUE);

    } else if (lgc_idle()) {
	inside_queue = TRUE;
	history->dcb->pos--;

	if (key_press)
	    update_prompt(TRUE);
    }
}

void lgc_determine_next()
{
    DCharBuf *dcb;

    dcb = NULL;

    if (queue_in())
	lgc_play_queue(++history->dcb->pos);
    else {
	if (active_list && active_list->dcb->len)
	    dcb = active_list->dcb;
	else {
	    lgc_set_state(IDLE);
	    goto prompt_return;
	}

	if (!user_stream && !inside_queue)
	    wght_add(active_list);

	if (inside_queue) {
	    inside_queue = FALSE;
	    event_post(QUEUE_EVENT);
	}

	switch (active_list->playmode) {
	case LEARN_MODE:
	    dcb->pos = lgc_learned(active_list);
	    break;
	case SRANDOM_MODE:
	    dcb->pos = lgc_srandom(active_list);
	    break;
	case TRANDOM_MODE:
	    dcb->pos = lgc_random(dcb->len);
	    break;
	case REPEAT_MODE:
	    if (dcb->pos < (dcb->len - 1))
		dcb->pos++;
	    else if (dcb->pos == (dcb->len - 1)) {
		event_post(LIST_EVENT);
		dcb->pos = 0;
	    }
	    break;
	default:
	    lgc_set_state(IDLE);
	    goto prompt_return;
	}

	lgc_play_list(active_list, TRUE);
    }

 prompt_return:

    if (lgc_need_tty_update())
	update_prompt(TRUE);
}
