/***********************************************
 *  listmanip.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.          
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* command.c */
extern CommandDef com[];
extern DCharBuf *args;

/* list.c */
extern ListDef *active_list, *edit_list;
extern char *SEARCHPRF[];

/* listlogic.c */
extern char *PLAYMODES[];

/* shell.c */
extern char *c_path;

/* tags.c */
extern flag_t read_file_tags;

/************************************
 *        Global Variables          *
 ************************************/

flag_t sort_view;

/************************************
 *        Local variables           *
 ************************************/

static char tmpstr_lm[COMM_SIZE], tmpstr2_lm[COMM_SIZE];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int listm_get_bytec()
{
    unsigned int bytec;

    bytec = 0;

    return bytec;
}

void listm_init()
{
    sort_view = TAG_FILE_PREF;
    memset(tmpstr_lm,  0, COMM_SIZE);
    memset(tmpstr2_lm, 0, COMM_SIZE);
}

void listm_clear()
{
}

/************************************
 *         Playlist manip           *
 ************************************/

void plist_clr(ListDef * lst)
{
    dbuf_clear(lst->dcb);
}

int plist_compare(const void *vel1, const void *vel2)
{
    str_get_sort_id(((DCBNode *) vel1)->ptr, tmpstr_lm);
    str_get_sort_id(((DCBNode *) vel2)->ptr, tmpstr2_lm);

    return strcmp(tmpstr_lm, tmpstr2_lm);
}

void /* user */ plist_sort_comm()
{
    if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else if (edit_list->dcb->len <= 1)
	show_user(DEBUG_MSG " \"%s\" list size too small",
		  edit_list->dcb->name);
    else {
	sort_view = edit_list->listview;
	qsort((void *) edit_list->dcb->buf, edit_list->dcb->len, 
	      sizeof(DCBNode), plist_compare);

	show_user(STATUS_MSG " Sorted %d entries in \"%s\" list",
		  edit_list->dcb->len, edit_list->dcb->name);
    }
}

void /* user */ plist_move_comm()
{
    char *tok1, *tok2;
    int targ, dest;

    tok1 = dbuf_get(args, 1);
    tok2 = dbuf_get(args, 2);

    if (!(tok1 && tok2 && is_all_digits(tok1) && is_all_digits(tok2)))
	print_cmnd_help(MOVE_LIST_INDEX);
    else if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else {
	targ = atoi(tok1);
	dest = atoi(tok2);

	if (dbuf_swap(edit_list->dcb, targ, dest))
	    show_user(STATUS_MSG " Swapped \"%s\" list indices %d and %d",
		      edit_list->dcb->name, targ, dest);
	else
	    show_user(ERROR_MSG " Invalid index specified");
    }
}

void /* user */ plist_play_index_comm()
{
    char *str_index;
    int index;

    str_index = dbuf_get(args, 1);

    if (!(str_index && is_all_digits(str_index)))
	print_cmnd_help(PLAY_LIST_INDEX);
    else if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else {
	index = atoi(str_index);

	if (index >= 0 && index < edit_list->dcb->len) {
	    edit_list->dcb->pos = index;
	    lgc_play_list(edit_list, TRUE);
	} else
	    show_user(DEBUG_MSG " index %d out of range", index);
    }
}
 
/*
 * When deleting with grep strings, it is likely that
 * the entries to be deleted are in groups, so we try to take
 * advantage of this by deleting them in groups (if possible).
 *
 * tests show remarkable speedups when entries are in groups and
 * the list size is large. 
 *
 * should have negligible slowdown for small lists
 */
void plist_grep_delete(ListDef * lst, char *str)
{
    int i, rem, low;

    rem = 0;
    low = -1;

    for(i = 0; i < lst->dcb->len; i++) {
	str_get_id(lst, i, tmpstr_lm);

	if (grep_against_args(1, tmpstr_lm)) {
	    if (low == -1) low = i;
	    rem++;
	} else if (low >= 0) {
	    if (str_remove_range(lst, low, (i-1)))
		i -= (i-low);
	    low = -1;
	}
    }

    if (low >= 0) /* last list entry matched */
	str_remove_range(lst, low, (i-1));

    outb_add(STATUS_MSG" %d entry(s) removed from \"%s\" [%d remain]", rem,
	     lst->dcb->name, lst->dcb->len);
}

/* 
 * Now removes entire ranges at a time, instead of one by one.
 */
void plist_range_delete(ListDef * lst, char *str)
{
    int low, high;

    get_digit_range(str, &low, &high, 0, lst->dcb->len - 1);

    if (low < high &&
	(low >= 0 && low < lst->dcb->len) &&
	(high >= 0 && high < lst->dcb->len)) {

	str_remove_range(lst, low, high);

	outb_add(STATUS_MSG 
		 " %d entry(s) removed from \"%s\" in range %d-%d [%d remain]",
		 (high - low) + 1, lst->dcb->name, low, high, lst->dcb->len);
    } else
	outb_add(DEBUG_MSG " Bad index range");
}

void plist_index_delete(ListDef * lst, char *str)
{
    int index;

    if (!lst->dcb->len)
	outb_add(DEBUG_MSG " \"%s\" list is empty", lst->dcb->name);
    else if (is_digit_range(str))
	plist_range_delete(lst, str);
    else if (is_all_digits(str)) {
	index = atoi(str);

	if (index >= 0 && index < lst->dcb->len) {
	    str_remove(lst, index);
	    outb_add(STATUS_MSG " Removed index %d from \"%s\" [%d remain]",
		     index, lst->dcb->name, lst->dcb->len);
	} else
	    outb_add(ERROR_MSG " index %d out of range", index);

    } else if (strcmp("all", str) == 0) {
	plist_clr(lst);
	outb_add(STATUS_MSG " \"%s\" list cleared", lst->dcb->name);
    } else
	plist_grep_delete(lst, str);

    outb_flush();
}

void /* user */ plist_delete_comm()
{
    char *str;

    str = dbuf_get(args, 1);

    if (!str)
	print_cmnd_help(DELETE_LIST_INDEX);
    else if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else {
	plist_index_delete(edit_list, str);
	plist_update_env();
    }
}

void plist_grep_add(ListDef * to, ListDef * from)
{
    int i, len, add;

    add = 0;
    len = from->dcb->len;

    for (i = 0; i < len; i++) {
	str_get_id(from, i, tmpstr_lm);

	if (grep_against_args(2, tmpstr_lm)) {
	    dbuf_add_str(to->dcb, dbuf_get(from->dcb, i));
	    add++;
	}
    }

    show_user(STATUS_MSG " %d entry(s) added to \"%s\" list", add,
	      to->dcb->name);
}

void plist_range_add(char *str, ListDef * to, ListDef * from)
{
    int i, add, low, high;

    add = 0;
    get_digit_range(str, &low, &high, 0, from->dcb->len - 1);

    if (low < high &&
	(low >= 0 && low < from->dcb->len) &&
	(high >= 0 && high < from->dcb->len)) {
	show_user(STATUS_MSG
		  " Copying \"%s\" entry(s) in range \"%d-%d\" to \"%s\"",
		  from->dcb->name, low, high, to->dcb->name);

	for (i = low; i <= high; i++) {
	    dbuf_add_str(to->dcb, dbuf_get(from->dcb, i));
	    add++;
	}

	show_user(STATUS_MSG " Copied %d entry(s) to \"%s\" list", add,
		  to->dcb->name);
    } else
	show_user(DEBUG_MSG " Bad index range");
}

void /* user */ plist_copy_comm()
{
    ListDef *copyto, *copyfrom;
    char *s, *name, *smem, *str;
    int i, flen, tlen, index;

    name = dbuf_get(args, 1);
    s = dbuf_get(args, 2);
    smem = NULL;

    if (!name || !s)
	print_cmnd_help(ADD_TO_LIST);
    else {
	copyto = plist_get(name);
	copyfrom = edit_list;
	str = tokenize(s, &smem, ',');

	if (!copyfrom) {
	    show_user(DEBUG_MSG " Edit list undefined");
	    return;
	}

	if (!copyto) {
	    show_user(DEBUG_MSG " \"%s\" list not found", name);
	    return;
	}

	if (copyfrom->dcb->len == 0) {
	    show_user(DEBUG_MSG " \"%s\" list is empty", copyfrom->dcb->name);
	    return;
	}

	while (str && *str != ZCHAR) {
	    if (is_digit_range(str))
		plist_range_add(str, copyto, copyfrom);
	    else if (is_all_digits(str)) {
		index = atoi(str);

		if (index >= 0 && index < copyfrom->dcb->len) {
		    dbuf_add_str(copyto->dcb, dbuf_get(copyfrom->dcb, index));
		    show_user(STATUS_MSG
			      " Copied index %d from \"%s\" to \"%s\"", index,
			      copyfrom->dcb->name, copyto->dcb->name);
		} else
		    show_user(ERROR_MSG " index %d out of range", index);
	    } else if (strcmp("all", str) == 0) {
		flen = copyfrom->dcb->len;
		tlen = copyto->dcb->len;

		for (i = 0; i < flen; i++)
		    dbuf_add_str(copyto->dcb, dbuf_get(copyfrom->dcb, i));

		show_user(STATUS_MSG " Copied %d entry(s) to \"%s\" list",
			  copyto->dcb->len - tlen,
			  copyto->dcb->name);
	    } else
		plist_grep_add(copyto, copyfrom);
	    str = tokenize(NULL, &smem, ',');
	}

	plist_update_env();
    }
}

void /* user */ plist_add_file_comm()
{
    char *file, tmpfl[COMM_SIZE];
    struct stat sstat;
    int ndx;

    ndx = 1;
    file = dbuf_get(args, ndx++);

    if (!file) {
	print_cmnd_help(ADD_FILE_TO_LIST);
	return;
    }

    if (!edit_list) {
	show_user(DEBUG_MSG " Edit list undefined");
	return;
    }

    while (file) {
	if (is_url(file)) {
	    if (lgc_check_format(file))
		str_add(edit_list->dcb, file, 0, FALSE, read_file_tags);
	    else
		show_user(DEBUG_MSG " No suitable player defined for \"%s\"",
			  file);
	} else {
	    if (*file != '/') {
		safe_sprintf(tmpfl, COMM_SIZE, "%s/%s", c_path, file);
		file = tmpfl;
	    }

	    if (stat(file, &sstat))
		show_user(ERROR_MSG " Error reading file \"%s\" : %s", file,
			  strerror(errno));
	    else if (sstat.st_mode & S_IFDIR)
		show_user(DEBUG_MSG " \"%s\" is a directory", file);
	    else {
		if (lgc_check_format(file))
		    str_add(edit_list->dcb, file, 0, FALSE, read_file_tags);
		else
		    show_user(DEBUG_MSG 
			      " No suitable player defined for \"%s\"", file);
	    }
	}
	file = dbuf_get(args, ndx++);
    }

    plist_update_env();
}

void /* user */ plist_add_dir_comm()
{
    DCharBuf *plist, *files, *dcb;
    int i, len, rec;
    char *dir;

    if (!edit_list) {
	show_user(DEBUG_MSG " Edit list undefined");
	return;
    }

    rec = FALSE;
    dir = dbuf_get(args, 1);

    if (!dir)
	dir = c_path;

    if (strcmp(com[ADD_DIR_RECURSE].name, dbuf_get(args, 0)) == 0)
	rec = TRUE;

    plist = dbuf_init(NULL, DCHAR_GROW, 256, 256);
    files = dbuf_init(NULL, DCHAR_GROW, LIST_SIZE, LIST_SIZE);

    dbuf_add_str(plist, dir);

    if (rec)
	show_user(STATUS_MSG 
		  " Recursively adding playable files in \"%s\" ...", dir);
    else
	show_user(STATUS_MSG " Adding playable files in \"%s\" ...", dir);

    while (plist->len > 0)
	add_dir(rec, plist, files);

    chdir(c_path);

    if (files->len) {
	dcb = edit_list->dcb;
	len = files->len;

	if (len > 1) {
	    sort_view = edit_list->listview;
	    qsort((void *) files->buf, len, sizeof(DCBNode), plist_compare);
	}

	for (i = 0; i < len; i++)
	    dbuf_add_str(dcb, dbuf_get(files, i));

	plist_update_env();
	show_user(STATUS_MSG " Added %d file(s) to \"%s\" list", len, 
		  dcb->name);
    } else
	show_user(DEBUG_MSG " No playable file(s) found in \"%s\"", dir);

    dbuf_free(files);
    dbuf_free(plist);
}

void /* user */ plist_set_mode_comm()
{
    char *comm;
    
    comm = dbuf_get(args, 0);

    if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else {
	if (strcmp(comm, com[SET_MODE_ONCE].name) == 0)
	    edit_list->playmode = ONCE_MODE;
	else if (strcmp(comm, com[SET_MODE_REPEAT].name) == 0)
	    edit_list->playmode = REPEAT_MODE;
	else if (strcmp(comm, com[SET_MODE_TRANDOM].name) == 0)
	    edit_list->playmode = TRANDOM_MODE;
	else if (strcmp(comm, com[SET_MODE_SRANDOM].name) == 0)
	    edit_list->playmode = SRANDOM_MODE;
	else if (strcmp(comm, com[SET_MODE_LEARN].name) == 0)
	    edit_list->playmode = LEARN_MODE;

	show_user(STATUS_MSG " \"%s\" list mode set to %s",
		  edit_list->dcb->name, PLAYMODES[edit_list->playmode]);
    }
}

void /* user */ plist_set_view_comm()
{
    char *comm;

    comm = dbuf_get(args, 0);

    if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else {
	if (strcmp(comm, com[SET_PREF_TF].name) == 0)
	    edit_list->listview = TAG_FILE_PREF;
	else if (strcmp(comm, com[SET_PREF_TPF].name) == 0)
	    edit_list->listview = TAG_PATH_PREF;
	else if (strcmp(comm, com[SET_PREF_PF].name) == 0)
	    edit_list->listview = PATH_FILE_PREF;
	else if (strcmp(comm, com[SET_PREF_F].name) == 0)
	    edit_list->listview = FILE_ONLY_PREF;

	show_user(STATUS_MSG " \"%s\" list view set to: %s",
		  edit_list->dcb->name, SEARCHPRF[edit_list->listview]);
    }
}

void /* user */ plist_set_weights_comm()
{
    char *pw, *sw;
    int ipw, isw;

    pw = dbuf_get(args, 1);
    sw = dbuf_get(args, 2);

    if (!pw || !sw)
	print_cmnd_help(SET_LEARN_WEIGHTS);
    else if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else {
	ipw = atoi(pw);
	isw = atoi(sw);
	if (ipw < 1 || ipw > WEIGHT_MAX)
	    show_user(DEBUG_MSG
		      " Invalid weight: \"%s\", valid range is [1, %d]", 
		      pw, WEIGHT_MAX);
	else if (isw < 1 || isw > WEIGHT_MAX)
	    show_user(DEBUG_MSG
		      " Invalid weight: \"%s\", valid range is [1, %d]", 
		      sw, WEIGHT_MAX);
	else {
	    edit_list->play_weight = ipw;
	    edit_list->skip_weight = isw;
	}
    }
}
