/***********************************************
 *  main.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 *
 *
 *
 *  imp3sh is an interactive mp3 shell using the
 *  Xaudio Async library for decoding.  imp3sh was 
 *  NOT created to be a suitable replacement for 
 *  fully-featured shells like bash or tcsh, I just
 *  implemented the shell features I used most (and 
 *  not even all of those yet.)  
 *                                            
 *  [This has been pieced together in  
 *   my spare time and is an unsightly mess of 
 *   code, although very stable.]
 ***********************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *        External References       *
 ************************************/

/* decoder.c */
extern XA_TimecodeInfo tinfo;
extern flag_t decoder_open;

/* listlogic.c */
extern ProcessState state;

/* command.c */
extern CommandDef com[];
extern DCharBuf *args;
extern flag_t prompt_now, running_builtin, ignore_char_input;

/* keys.c */
extern TermCtrl t_ctrl[];

/* player.c */
extern flag_t external_player, external_paused;

/* shell.c */
extern flag_t ignore_hist_input;

/************************************
 *        Global Variables          *
 ************************************/

char *TIMESTAMP[] = { 
    "Off", 
    "On - Time played",
    "On - Time left"
};

int thisprgp;
volatile int tab_length, tty_width, tty_height;
flag_t piped_io_flag, exit_flag, show_time_stamp, disable_user_exec;
flag_t ignore_ints;
struct termios savtty, newtty;

/************************************
 *        Local Variables           *
 ************************************/

static char piped_outstr[INFO_SIZE], piped_instr[INFO_SIZE];
static char *termtype, *prog_name, tmpstr_ma[COMM_SIZE];
static int fdstdin;
static volatile int child_sigs;
static flag_t piped_input, piped_output, disable_rc_check;
static ModTable modtab[] = { /* the order is important */
    {buff_init,  buff_clear,  buff_get_bytec},
    {shell_init, shell_clear, shell_get_bytec},
    {proc_init,  proc_clear,  proc_get_bytec},
    {lgc_init,   lgc_clear,   lgc_get_bytec},
    {dec_init,   dec_clear,   dec_get_bytec},
    {queue_init, queue_clear, queue_get_bytec},
    {plist_init, plist_clear, plist_get_bytec},
    {listm_init, listm_clear, listm_get_bytec},
    {listf_init, listf_clear, listf_get_bytec},
    {event_init, event_clear, event_get_bytec},
    {expl_init,  expl_clear,  expl_get_bytec},
    {plyr_init,  plyr_clear,  plyr_get_bytec},
    {info_init,  info_clear,  info_get_bytec},
    {com_init,   com_clear,   com_get_bytec},
    {key_init_m, key_clear_m, key_get_bytec},
    {str_init,   str_clear,   str_get_bytec},
    {tag_init,   tag_clear,   tag_get_bytec},
    {util_init,  util_clear,  util_get_bytec}
};

/************************************
 *         Module routines          * 
 ************************************/

unsigned int main_get_bytec()
{
    unsigned int bytec;
    int i;

    bytec = 0;

    for(i = 0; i < NUM_MODS; i++)
	bytec += (modtab[i].bytec)();

    return bytec;
}

void main_init(int argc, char **argv)
{
    int i;

    termtype = prog_name = NULL;
    piped_input = piped_output = disable_rc_check = disable_user_exec = FALSE;
    exit_flag = piped_io_flag = show_time_stamp = ignore_ints = FALSE;
    thisprgp = tab_length = tty_width = tty_height = 0;
    fdstdin = child_sigs = 0;

    memset(&savtty, 0, sizeof(struct termios));
    memset(&newtty, 0, sizeof(struct termios));
    memset(piped_outstr, 0, INFO_SIZE);
    memset(piped_instr,  0, INFO_SIZE);
    memset(tmpstr_ma,    0, COMM_SIZE);

    thisprgp = getpid();
    setpgid(thisprgp, thisprgp);
    prog_name = *argv;
    dup2(0, fdstdin); 

    safe_sprintf(tmpstr_ma, COMM_SIZE, "%s/.imp3rc", getenv("HOME"));
    parse_args(argc, argv);

    if (piped_input || piped_output || !isatty(0) || !isatty(1))
	piped_io_flag = TRUE;
    else
	setup_terminal();

    for(i = 0; i < NUM_MODS; i++)
	(modtab[i].init)();

    if (!disable_rc_check) {
	show_user("...executing resource file:");
	source_file(tmpstr_ma);
    }

    outb_clear();
    com_info_clear();
    prompt(TRUE);
}

void main_clear()
{
    int i;

    for(i = 0; i < NUM_MODS; i++)
	(modtab[i].clear)();
    
    if (!piped_io_flag)
	tcsetattr(0, TCSAFLUSH, &savtty);
}

/************************************
 *         Output Functions         *
 ************************************/

void out_strpipe(char *str, int nl)
{
    int n, ret, len;

    ret = n = 0;
    len = get_strlen(str);

    if (!len)
	return;

    do {

	do {
	    ret = write(1, &str[n], len - n);
	} while (ret == -1 && errno == EINTR);

	n += ret;

    } while (ret > 0 && n < len);

    if (n == len && nl) 
	write(1, "\n", 1);
}

void out_str(char *str)
{
    if (str) {
	if (piped_io_flag)
	    out_strpipe(str, TRUE);
	else
	    puts(str);
    }
}

void out_str_n(char *str)
{
    if (str) {
	if (piped_io_flag)
	    out_strpipe(str, FALSE);
	else
	    fputs(str, stdout);
    }
}

void out_flush()
{
    if (!piped_io_flag)
	fflush(stdout);
}

int have_tty()
{
    return (!ignore_char_input && outb_empty()) ? TRUE : FALSE;
}

void open_inpipe()
{
    int fd;

    fd = open(piped_instr, O_RDONLY | O_NONBLOCK);

    if (fd != -1) {
	dup2(fd, fdstdin);
	piped_input = TRUE;
    } else {
	fprintf(stderr, "%s: error opening input: %s : %s\n", prog_name,
		piped_instr, strerror(errno));
	exit(1);
    }
}

void err_outpipe()
{
    fprintf(stderr, "%s: error opening output: %s : %s\n", prog_name,
	    piped_outstr, strerror(errno));
    exit(1);
}

void open_outpipe()
{
    int fd, fd2;

    fd = open(piped_outstr, O_RDONLY | O_NONBLOCK);

    if (fd == -1)
	err_outpipe();
    else {
	fd2 = open(piped_outstr, O_WRONLY);

	if (fd2 == -1) {
	    close(fd);
	    err_outpipe();
	} else {
	    piped_output = TRUE;
	    dup2(fd2, 1);
	    dup2(fd2, 2);
	    close(fd);
	}
    }
}

/* main loop routines */

void monitor_setup(fd_set * read_set, int *max_fd)
{
    int decfd;

    FD_ZERO(read_set);
    *max_fd = 0;

    if (!ignore_char_input) {
	FD_SET(0, read_set);
	*max_fd = 1;
    }

    if (decoder_open) {
	decfd = dec_get_read_pipe();
	FD_SET(decfd, read_set);
	*max_fd = decfd + 1;
    }

    proc_set_pipes(read_set, max_fd);
}

void monitor_check(fd_set * read_set)
{
    if (!ignore_char_input && FD_ISSET(0, read_set) && !com_monitor()
	&& piped_io_flag) {
	close(0);
	if (piped_input)
	    open_inpipe();
	else
	    ignore_char_input = TRUE;
    }

    if (decoder_open && FD_ISSET(dec_get_read_pipe(), read_set))
	dec_monitor();

    proc_check_pipes(read_set);
}

void monitor()
{
    struct timeval tv;
    fd_set read_set;
    int ret, max_fd;

    while (!exit_flag) {
	clear_child_sigs();

	do {
	    monitor_setup(&read_set, &max_fd);

	    tv.tv_sec = 0;
	    tv.tv_usec = SELECT_TIMEOUT;

	    ret = select(max_fd, &read_set, NULL, NULL, &tv);
	}
	while (ret == -1 && errno == EINTR);

	if (ret > 0)
	    monitor_check(&read_set);
	event_check();
    }
}

/* signal handlers */

void int_sig_handler()
{
    if (!piped_io_flag) {
	tcsetpgrp(0, thisprgp);
	tcsetattr(0, TCSAFLUSH, &newtty);
    }
    ignore_char_input = FALSE;
}

void winch_sig_handler(int a)
{
    if (!piped_io_flag && tgetent(NULL, termtype) == 1) {
	tty_width = tgetnum("co");
	tty_height = tgetnum("li");

	if (tty_width == ERR || tty_height == ERR) {
	    tty_width = 80;
	    tty_height = 20;
	}
	if (outb_empty())
	    outb_clear();
    }
}

/* external player timer */
void alarm_sig_handler(int a)
{
    if (state == EXTERNAL) {
	alarm(1);

	if (!external_paused) {
	    if (external_player) {
		if (tinfo.m == 59 && tinfo.s == 59) {
		    tinfo.m = tinfo.s = 0;
		    tinfo.h++;
		} else if (tinfo.s == 59) {
		    tinfo.s = 0;
		    tinfo.m++;
		} else
		    tinfo.s++;
	    }

	    if (show_time_stamp && !ignore_char_input && !running_builtin)
		update_prompt(TRUE);
	}
    }
}

void clear_child_sigs()
{
    int sigs;

    sigs = child_sigs;

    if (sigs) {
	proc_protect();

	handle_child_sigs();
	child_sigs -= sigs;

	proc_unprotect();
    }
}

void count_child_sigs(int a)
{
    if (!exit_flag)
	child_sigs++;
}

void handle_child_sigs(int a)
{
    char procname[32];
    ProcessEntry *prev, *pptr;
    int stat, jobt;
    pid_t pid;

    if (ignore_ints)
	return;

    pid = waitpid(-1, &stat, WNOHANG | WUNTRACED);

    while (pid > 0) {
	prev = NULL;
	pptr = proc_get(pid, &prev);
	jobt = -1;

	if (pptr)
	    safe_strcpy(procname, pptr->name, 32);
	else
	    *procname = ZCHAR;

	if (WIFSTOPPED(stat)) {
	    jobt = proc_stop_fg(pptr);

	    if (jobt == -1)
		conb_add(DEBUG_MSG " Unregistered child stopped [pid=%d]", 
			 pid);
	} else {
	    jobt = proc_clean(pptr, prev);
	    stat = WEXITSTATUS(stat);

	    switch (jobt) {
	    case J_DECEXT:
		ext_handle_death(pid);
		break;
	    case -1:
		conb_add(DEBUG_MSG " Unregistered child died [pid=%d]", pid);
		break;
	    }

	    switch (stat) {
	    case CHILD_EXEC_FAILED:
		conb_add(ERROR_MSG " Command not found: \"%s\"", procname);
		break;
	    case DECODER_EXEC_FAILED:
		conb_add(ERROR_MSG " External player not found: \"%s\"",
			 procname);
		break;
	    }
	}

	if (!ignore_hist_input && have_tty())
	    switch (jobt) {
	    case J_BGRUN:
	    case -1:
		update_prompt(FALSE);
		break;
	    case J_STOPPED:
	    case J_FGRUN:
		conb_flush();
		prompt(TRUE);
		break;
	    }

	pid = waitpid(-1, &stat, WNOHANG | WUNTRACED);
    }
}


void source_file(char *file)
{
    char *strptr;
    FILE *fd;
    flag_t setflag;

    setflag = FALSE;
    fd = myfopen(file, "r");

    if (!ignore_hist_input) {
	ignore_hist_input = TRUE;
	setflag = TRUE;
    }

    if (fd == NULL)
	show_user(DEBUG_MSG " Open failed for: \"%s\": %s", file,
		  strerror(errno));
    else {
	get_fline(fd, tmpstr_ma, (COMM_SIZE - 1));
	while (!feof(fd)) {
	    strptr = skip_space(tmpstr_ma);
	    if (*strptr && *strptr != '#')
	        com_handle(strptr);
	    get_fline(fd, tmpstr_ma, (COMM_SIZE - 1));
	}

	fclose(fd);
    }

    /* Don't unset flag unless we
       are top-level call         */
    if (setflag)
	ignore_hist_input = FALSE;
}

void /* user */ source_comm()
{
    char *file;

    file = dbuf_get(args, 1);

    if (file)
	source_file(file);
    else
	print_cmnd_help(SOURCE_FILE);
}

void /* user */ set_time_stamp_comm()
{
    char *comm; 

    comm = dbuf_get(args, 0);

    if (strcmp(comm, com[SET_STAMP_OFF].name) == 0)
	show_time_stamp = TIME_OFF;
    else if (strcmp(comm, com[SET_STAMP_TIMP].name) == 0)
	show_time_stamp = TIME_NORM;
    else if (strcmp(comm, com[SET_STAMP_TIML].name) == 0)
	show_time_stamp = TIME_LEFT;

    show_user(STATUS_MSG " Timestamp %s", TIMESTAMP[show_time_stamp]);
}

void setup_terminal()
{
    char *cttype, *ttype;
    int i;

    tcsetpgrp(0, thisprgp);
    cttype = getenv("COLORTERM");
    ttype = getenv("TERM");

    if (!cttype && !ttype) {
	ttype = "VT100";
	fprintf(stderr, "%s: defaulting to a \"%s\" terminal\n", prog_name, 
		ttype);
    }

    if (cttype && tgetent(NULL, cttype) == 1)
	termtype = cttype;
    else if (tgetent(NULL, ttype) == 1)
	termtype = ttype;
    else {
	if (cttype)
	    fprintf(stderr, "%s: terminal \"%s\": tgetent() failed : %s\n",
		    prog_name, cttype, strerror(errno));
	if (ttype)
	    fprintf(stderr, "%s: terminal \"%s\": tgetent() failed : %s\n",
		    prog_name, ttype, strerror(errno));
	exit(1);
    }

    for (i = 0; i < (NUM_TERM_CTRL - NUM_TERM_USER); i++) {
	t_ctrl[i].string = tgetstr(t_ctrl[i].name, 0);
	if (!t_ctrl[i].string)
	    fprintf(stderr,
		    "%s: terminal \"%s\": key \"%s\" not supported!\n",
		    prog_name, termtype, t_ctrl[i].desc);	    
    }

    tty_width = tgetnum("co");
    tty_height = tgetnum("li");
    tab_length = tgetnum("it");

    if (tab_length == ERR || tab_length == 0) {
	fprintf(stderr,
		"%s: terminal \"%s\": reporting bad <tab> character length!\n",
		prog_name, termtype);
	tab_length = 8;
    }

    if (tty_width == ERR || tty_height == ERR) {
	fprintf(stderr, "%s: terminal \"%s\": screen size not available!\n",
		prog_name, termtype);
	fprintf(stderr, "%s: defaulting to 80 columns x 20 lines\n",
		prog_name);
	tty_width = 80;
	tty_height = 20;
    }

    if (tcgetattr(0, &savtty) == -1) {
	fprintf(stderr, "%s: terminal \"%s\": tcgetattr() failed : %s\n",
		prog_name, termtype, strerror(errno));
	exit(1);
    }

    cfmakeraw(&newtty);
    newtty.c_iflag &= ~IGNCR;
    newtty.c_iflag |= (ICRNL | IMAXBEL);
    newtty.c_lflag |= ISIG;
    newtty.c_oflag |= (OPOST | ONLCR);
    newtty.c_cc[VMIN] = 1;
    newtty.c_cc[VTIME] = 0;

    if (tcsetattr(0, TCSAFLUSH, &newtty) == -1) {
	fprintf(stderr, "%s: terminal \"%s\": tcsetattr() failed : %s\n",
		prog_name, termtype, strerror(errno));
	exit(1);
    }
}

void /* user */ exit_comm()
{
    event_post(EXIT_EVENT);
    prompt_now = FALSE;
    ignore_ints = TRUE;

    if (decoder_open)
	dec_close();
    else
	exit_flag = TRUE;
}

void print_usage()
{
    fprintf(stderr,
	    "imp3sh %s\n" 
	    "Copyright (C) %s Eric Kratzer\n"
	    "Includes Xaudio software Copyright (C) 1996-2000\n"
	    "         Xaudio Corporation All Rights Reserved.\n\n"
	    "usage: %s [OPTIONS...]\n\n"
	    "OPTIONS are:\n"
	    "-r <resource_file>    - Set resource file to use\n"
	    "-i <input_fifo>       - Set named-pipe to use for stdin\n"
	    "-o <output_fifo>      - Set named-pipe to use for stdout\n"
	    "-norc                 - Overrides \"-r\" option\n"
	    "-noexec               - Disables exec'ing of external commands\n",
	    VERSION, DATE, prog_name);
    exit(1);
}

void parse_args(int argc, char **argv)
{
    int i, ret, argsleft;

    argsleft = argc - 1;

    if (argc <= 1)
	return;

    for (i = 1; i < argc; i++) {
	if (strcmp(argv[i], "-i") == 0) {
	    if (--argsleft) {
		ret = mkfifo(argv[++i], 0777);

		argsleft--;
		if (ret == 0 || (ret == -1 && errno == EEXIST)) {
		    strcpy(piped_instr, argv[i]);
		    open_inpipe();
		} else {
		    fprintf(stderr,
			    "%s: error creating named pipe \"%s\" : %s\n",
			    prog_name, argv[i], strerror(errno));
		    exit(1);
		}
	    } else
		print_usage();
	} else if (strcmp(argv[i], "-o") == 0) {
	    if (--argsleft) {
		ret = mkfifo(argv[++i], 0777);

		argsleft--;
		if (ret == 0 || (ret == -1 && errno == EEXIST)) {
		    strcpy(piped_outstr, argv[i]);
		    open_outpipe();
		} else {
		    fprintf(stderr,
			    "%s: error creating named pipe \"%s\" : %s\n",
			    prog_name, argv[i], strerror(errno));
		    exit(1);
		}
	    } else
		print_usage();
	} else if (strcmp(argv[i], "-r") == 0) {
	    if (--argsleft) {
		argsleft--;
		strcpy(tmpstr_ma, argv[++i]);
	    } else
		print_usage();
	} else if (strcmp(argv[i], "-noexec") == 0) {
	    argsleft--;
	    disable_user_exec = TRUE;
	} else if (strcmp(argv[i], "-norc") == 0) {
	    argsleft--;
	    disable_rc_check = TRUE;
	} else
	    print_usage();
    }
}

int main(int argc, char **argv)
{
    main_init(argc, argv);
    monitor();
    main_clear();
    return 0;
}
