/**********************************************
 *  proc.h                                                          
 *                                            
 *  Copyright 2001 Eric Kratzer.               
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 *
 *                                            
 **********************************************/

#ifndef PROC_H
#define PROC_H

#define CHILD_EXEC_FAILED   211
#define DECODER_EXEC_FAILED 212

enum {
    J_FGRUN,
    J_BGRUN,
    J_STOPPED,
    J_DECINT,
    J_DECEXT
};

typedef struct ProcessEntry_s {
    pid_t                  pid;
    int                    redir;
    int                    mode;
    int                    pipestdout[2];
    int                    pipestderr[2];
    char                   name[32];
    struct termios         tty;
    struct DCharBuf_s     *buff;
    struct ProcessEntry_s *next;
} ProcessEntry;

void          proc_init       ();
void          proc_clear      ();
unsigned int  proc_get_bytec  ();
void          proc_protect    ();
void          proc_unprotect  ();
ProcessEntry *proc_get        (pid_t, ProcessEntry **);
int           proc_stop_fg    (ProcessEntry *);
void          proc_set_term   (ProcessEntry *);
int           proc_clean      (ProcessEntry *, ProcessEntry *);
void          proc_add        (pid_t, int, int, int[2], int[2], char *);
void          proc_set_xg     (ProcessEntry *, int);
void          proc_send_2xg   (int mode);
void          proc_check_pipes(fd_set *);
void          proc_clean_pipes(ProcessEntry *);
void          proc_set_pipes  (fd_set *, int *);
void          proc_2fg_comm   ();	/* user */
void          proc_2bg_comm   ();	/* user */
void          proc_print_comm ();	/* user */
void          proc_flush_comm ();	/* user */

#endif
