/***********************************************
 *  queue.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.          
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* list.c */
extern ListDef *edit_list;

/* command.c */
extern DCharBuf *args;

/* shell.c */
extern char *c_path;

/* tags.c */
extern flag_t read_file_tags;

/* listmanip.c */
extern flag_t  sort_view;

/************************************
 *        Global Variables          *
 ************************************/

ListDef *history;
flag_t inside_queue;

/************************************
 *        Local variables           *
 ************************************/

static char tmpstr_q[COMM_SIZE];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int queue_get_bytec()
{
    unsigned int bytec;

    bytec = 0;
    bytec += sizeof(ListDef);
    bytec += dbuf_get_bytec(history->dcb);
  
    return bytec;
}

void queue_init()
{
    inside_queue = FALSE;
    history = malloc(sizeof(ListDef));
    history->playmode = QUEUE_MODE;
    history->listview = TAG_FILE_PREF;
    history->dcb = dbuf_init("queue", DCHAR_SHIFT, HIST_SIZE, 0);
    memset(tmpstr_q,  0, COMM_SIZE);
}

void queue_clear()
{
    dbuf_free(history->dcb);
    free(history);
}

/************************************
 *     Stream queue functions       *
 ************************************/

int queue_in()
{
    return (history->dcb->len && 
	    (history->dcb->pos < (history->dcb->len - 1))) ? TRUE : FALSE;
}

void queue_add(char *info)
{
    char *tagptr;
    flag_t tagpresent;

    if (!info)
	return;

    tagpresent = FALSE;
    tagptr = grep_str(TAG_CODE, info);
    if (tagptr) tagpresent = TRUE;

    str_add(history->dcb, info, 0, tagpresent, read_file_tags);
    history->dcb->pos = history->dcb->len - 1;
}

void /* user */ queue_resize_comm()
{
    char *str;
    int new_size;

    str = dbuf_get(args, 1);

    if (!str)
	print_cmnd_help(RESIZE_QUEUE);
    else {
	new_size = atoi(str);
	if (new_size <= 0) 
	    show_user(DEBUG_MSG " Invalid queue size: %d", new_size);
	else {
	    if (!dbuf_resize(history->dcb, new_size))
		show_user(ERROR_MSG " Failed to resize queue");
	    else
		show_user(STATUS_MSG " Maximum queue size set to: %d",
			  history->dcb->size);
	}
    }
}

void /* user */ queue_print_comm()
{
    char *pat, *hedr;
    int i, len, pos, hlen;

    hlen = history->dcb->len;

    if (!hlen)
	outb_add(DEBUG_MSG " Queue is empty");
    else {
	pat = dbuf_get(args, 1);
	len = get_strlen(pat);
	pos = history->dcb->pos;

	for (i = 0; i < hlen; i++) {
	    str_get_id(history, i, tmpstr_q);

	    if (len && grep_against_args(1, tmpstr_q) == FALSE)
		continue;

	    if (i == pos)
		hedr = PNTR_MSG;
	    else
		hedr = LIST_MSG;

	    outb_add("%s %d> %s", hedr, i, tmpstr_q);
	}
    }

    outb_flush();
}

void queue_add_dir(char *dir)
{
    DCharBuf *files, *dirlist, *dcb;
    int i, len;

    if (!dir)
	return;

    dirlist = dbuf_init(NULL, DCHAR_GROW, 256, 256);
    files = dbuf_init(NULL, DCHAR_GROW, 256, 256);

    show_user(STATUS_MSG " Recursively queueing files in \"%s\" ...", dir);
    dbuf_add_str(dirlist, dir);

    while (dirlist->len)
	add_dir(TRUE, dirlist, files);

    chdir(c_path);

    if (files->len) {
	dcb = history->dcb;
	len = files->len;

	if (len > 1) {
	    sort_view = TAG_FILE_PREF;
	    qsort((void *) files->buf, len, sizeof(DCBNode), plist_compare);
	}

	for (i = 0; i < len; i++)
	    dbuf_add_str(dcb, dbuf_get(files, i));

	show_user(STATUS_MSG " Queued %d file(s) from \"%s\"", len, dir);
    } else
	show_user(DEBUG_MSG " No playable files found in \"%s\"", dir);

    dbuf_free(dirlist);
    dbuf_free(files);
}

void /* user */ queue_files_comm()
{
    struct stat sstat;
    char *file;
    int ndx;

    ndx = 1;
    file = dbuf_get(args, ndx++);

    if (!file) {
	print_cmnd_help(QUEUE_FILE);
	return;
    }

    while (file) {
	if (is_url(file))
	    str_add(history->dcb, file, 0, FALSE, FALSE);
	else {
	    if (*file != '/') {
		safe_sprintf(tmpstr_q, COMM_SIZE, "%s/%s", c_path, file);
		file = tmpstr_q;
	    }

	    if (stat(file, &sstat))
		show_user(ERROR_MSG " Error reading file \"%s\" : %s", file,
			  strerror(errno));
	    else if (sstat.st_mode & S_IFDIR)
		queue_add_dir(file);
	    else {
		if (lgc_check_format(file))
		    str_add(history->dcb, file, 0, FALSE, read_file_tags);
		else
		    show_user(ERROR_MSG
			      " No suitable player defined for: \"%s\"", file);
	    }
	}
	file = dbuf_get(args, ndx++);
    }

    plist_update_env();
}

void queue_grep_delete(char *pat)
{
    int i, rem;

    rem = 0;

    for (i = 0; i < history->dcb->len; i++) {
	str_get_id(history, i, tmpstr_q);
	if (grep_against_args(1, tmpstr_q)) {
	    dbuf_remove(history->dcb, i);
	    i--; rem++;
	}
    }

    outb_add(STATUS_MSG" %d entry(s) removed from Queue [%d remain]", rem,
	     history->dcb->len);
}

void queue_range_delete(char *str)
{
    int low, high;

    get_digit_range(str, &low, &high, 0, history->dcb->len - 1);

    if (low < high &&
	(low >= 0 && low < history->dcb->len) &&
	(high >= 0 && high < history->dcb->len)) {

	dbuf_remove_range(history->dcb, low, high);

	outb_add(STATUS_MSG 
		 " %d entry(s) removed from Queue in range %d-%d [%d remain]",
		 (high - low) + 1, low, high, history->dcb->len);
    } else
	outb_add(DEBUG_MSG " Bad index range");
}

void queue_index_delete(char *str)
{
    int index;

    if (!history->dcb->len)
	outb_add(DEBUG_MSG " Queue is empty");
    else if (is_digit_range(str))
	queue_range_delete(str);
    else if (is_all_digits(str)) {
	index = atoi(str);

	if (index >= 0 && index < history->dcb->len) {
	    str_remove(history, index);
	    outb_add(STATUS_MSG" Removed index %d from Queue [%d remain]",
		     index, history->dcb->len);
	} else
	    outb_add(ERROR_MSG " index %d out of range", index);

    } else if (strcmp("all", str) == 0) {
	dbuf_clear(history->dcb);
	inside_queue = FALSE;
	outb_add(STATUS_MSG " Queue cleared");
    } else
	queue_grep_delete(str);

    outb_flush();
}

void /* user */ queue_delete_comm()
{
    char *str;

    str = dbuf_get(args, 1);

    if (!str)
	print_cmnd_help(DELETE_QUEUE_INDEX);
    else {
	queue_index_delete(str);
	plist_update_env();
    }
}

void queue_list_grep(ListDef * from)
{
    int i, add, len;

    add = 0;
    len = from->dcb->len;

    for (i = 0; i < len; i++) {
	str_get_id(from, i, tmpstr_q);
	if (grep_against_args(1, tmpstr_q)) {
	    add++;
	    dbuf_add_str(history->dcb, dbuf_get(from->dcb, i));
	}
    }

    show_user(STATUS_MSG " %d entry(s) added to queue", add);
}

void queue_list_range(char *str, ListDef * from)
{
    int i, low, high;

    get_digit_range(str, &low, &high, 0, from->dcb->len - 1);

    if (low < high &&
	(low >= 0 && low < from->dcb->len) &&
	(high >= 0 && high < from->dcb->len)) {
	show_user(STATUS_MSG" Queuing \"%s\" list entry(s) in range %d-%d:",
		  from->dcb->name, low, high);

	for (i = low; i <= high; i++)
	    dbuf_add_str(history->dcb, dbuf_get(from->dcb, i));

	show_user(STATUS_MSG " %d entry(s) added to queue", (high - low) + 1);
    } else
	show_user(DEBUG_MSG " Bad index range");
}

void /* user */ queue_index_comm()
{
    char *s, *smem, *str;
    int ndx;

    s = dbuf_get(args, 1);

    if (!s)
	print_cmnd_help(QUEUE_INDEX);
    else if (!edit_list)
	show_user(DEBUG_MSG " Edit list undefined");
    else {
	smem = NULL;
	str = tokenize(s, &smem, ',');

	while (str && *str != ZCHAR) {
	    if (is_digit_range(str))
		queue_list_range(str, edit_list);
	    else if (is_all_digits(str)) {
		ndx = atoi(str);

		if (ndx >= 0 && ndx < edit_list->dcb->len) {
		    dbuf_add_str(history->dcb, dbuf_get(edit_list->dcb, ndx));
		    show_user(STATUS_MSG 
			      " \"%s\" list index %d added to queue",
			      edit_list->dcb->name, ndx);
		} else
		    show_user(DEBUG_MSG " index %d out of range", ndx);
	    } else
		queue_list_grep(edit_list);

	    str = tokenize(NULL, &smem, ',');
	}

	if (queue_in())
	    plist_update_env();
    }
}
