/***********************************************
 *  shell.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.               
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* command.c */
extern CommandDef com[];
extern DCharBuf *args;

extern char **environ;

/************************************
 *        Global Variables          *
 ************************************/

DCharBuf *chist, *aliases;

char *c_path, *host;

flag_t ignore_hist_input;

/************************************
 *        Local variables           *
 ************************************/

static char tmpstr_sh[COMM_SIZE];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int shell_get_bytec()
{
    unsigned int bytec;

    bytec = 0;
    bytec += dbuf_get_bytec(chist);
    bytec += dbuf_get_bytec(aliases);
    bytec += PATH_MAX + INFO_SIZE;

    return bytec;
}

void shell_init()
{
    char *tmp;

    ignore_hist_input = FALSE;
    chist = dbuf_init(NULL, DCHAR_SHIFT, CHIST_SIZE, 0);
    aliases = dbuf_init(NULL, DCHAR_GROW, 128, 128);

    c_path = malloc(PATH_MAX);
    host = malloc(INFO_SIZE);

    memset(c_path, 0, PATH_MAX);
    memset(host,   0, INFO_SIZE);

    gethostname(host, INFO_SIZE);
    tmp = strchr(host, '.');
    if (*host && tmp)
	*tmp = '\0';

    getcwd(c_path, PATH_MAX);
    setenv("OLDPWD", c_path, TRUE);
    setenv("IMP3_VERSION", VERSION, TRUE);
    chist_read();
}

void shell_clear()
{
    chist_write();
    free(c_path);
    free(host);
    dbuf_free(chist);
    dbuf_free(aliases);
}

/************************************
 *   Command History functions      *
 ************************************/

void /* user */ chist_show_comm()
{
    char *name;
    int i, len;

    if (!chist->len)
	outb_add(DEBUG_MSG " Command history is empty");
    else {
	name = dbuf_get(args, 1);
	len = get_strlen(name);

	for (i = 0; i < chist->len; i++) {
	    if (len && grep_against_args(1, dbuf_get(chist, i)) == FALSE)
		continue;
	    outb_add("%d %s", i + 1, dbuf_get(chist, i));
	}
    }

    outb_flush();
}

void chist_read()
{
    int len;
    FILE *fd;

    safe_sprintf(tmpstr_sh, COMM_SIZE, "%s/.imp3sh_hist", getenv("HOME"));
    fd = myfopen(tmpstr_sh, "r");

    if (fd) {
	len = get_fline(fd, tmpstr_sh, COMM_SIZE);
	while (!feof(fd)) {
	    if (*tmpstr_sh)
		dbuf_add(chist, tmpstr_sh, ++len, 0);
	    len = get_fline(fd, tmpstr_sh, COMM_SIZE);
	}
	fclose(fd);
    }
}

void chist_write()
{
    FILE *fd;
    int i;

    if (!chist->len)
	return;

    safe_sprintf(tmpstr_sh, COMM_SIZE, "%s/.imp3sh_hist", getenv("HOME"));
    fd = myfopen(tmpstr_sh, "w+");

    if (fd) {
	for (i = 0; i < chist->len; i++)
	    fprintf(fd, "%s\n", dbuf_get(chist, i));
	fclose(fd);
    }
}

void chist_add(char *command)
{
    if (!ignore_hist_input && !is_all_space(command))
        dbuf_add_str(chist, command);
}

/************************************
 *         Aliasing functions       *
 ************************************/

int alias_check(char *str)
{
    char *tmp;
    int i, ndx, len, len2;

    len = get_strlen(str);
    ndx = -1;

    if (len)
	for (i = 0; i < aliases->len; i++) {
	    tmp = dbuf_get(aliases, i);
	    len2 = get_strlen(tmp);

	    if (tmp &&
		len2 > len &&
		strncmp(str, tmp, len) == 0 &&
		tmp[len] == ALIAS_SEPERATOR) {
		ndx = i;
		break;
	    }
	}

    return ndx;
}

char *alias_get(int ndx)
{
    char *tm;

    if (ndx >= 0 && ndx < aliases->len) {
	tm = dbuf_get(aliases, ndx);

	if (tm) {
	    tm = strchr(tm, ALIAS_SEPERATOR);
	    return (tm + 1);
	}
    }

    return NULL;
}

void /* user */ alias_remove_comm()
{
    char *name;
    int ndx;

    name = dbuf_get(args, 1);

    if (!name)
	print_cmnd_help(REMOVE_ALIAS);
    else {
	ndx = alias_check(name);

	if (ndx == -1)
	    show_user(DEBUG_MSG " Alias for \"%s\" not found", name);
	else {
	    dbuf_remove(aliases, ndx);
	    expl_comp_list_remove(name);
	}
    }
}

void alias_print()
{
    char *name, *repl;
    int i;


    if (!aliases->len)
	outb_add(DEBUG_MSG " No aliases defined");
    else
	for (i = 0; i < aliases->len; i++) {
	    strcpy(tmpstr_sh, dbuf_get(aliases, i));
	    repl = strchr(tmpstr_sh, ALIAS_SEPERATOR);
	    if (repl) {
		*repl = ZCHAR; repl++; name = tmpstr_sh;
		outb_add("%-10s : %s", name, repl);
	    }
	}

    outb_flush();
}

void /* user */ alias_add_comm()
{
    char *name, *repl, *cm;
    int i, ndx, cdx;

    name = dbuf_get(args, 1);
    repl = dbuf_get(args, 2);

    if (name && repl) {
	ndx = alias_check(name);
	cdx = com_find_match(name);

	if (cdx != NO_BUILTIN_COMM) {
	    show_user(ERROR_MSG " Built-in command \"%s\" cannot be aliased",
		      name);
	    return;
	}

	if (strcmp(repl, com[ADD_ALIAS].name) == 0 ||
	    strcmp(repl, com[ADD_EVENT].name) == 0 || 
	    strcmp(repl, com[MOD_KEY].name) == 0) {
	    show_user(ERROR_MSG " Special command \"%s\" cannot have an alias",
		      repl);
	    return;
	}

	if (ndx > -1)
	    show_user(DEBUG_MSG " Alias already exists for \"%s\"", name);
	else {
	    expl_comp_list_insert(name, repl, args->len);
	    i = 2;
	    cm = dbuf_get(args, i++);

	    safe_sprintf(tmpstr_sh, COMM_SIZE, "%s%c", name, ALIAS_SEPERATOR);

	    while (cm) {
		strcat(tmpstr_sh, cm);
		cm = dbuf_get(args, i++);
		if (cm)
		    strcat(tmpstr_sh, " ");
	    }
	    dbuf_add_str(aliases, tmpstr_sh);
	}
    } else if (name || repl)
	print_cmnd_help(ADD_ALIAS);
    else
	alias_print();		/* user just wants to see */
}

/************************************
 * ENVIRONMENT control functions    *
 ************************************/

void /* user */ env_set_comm()
{
    char *name, *val;

    name = dbuf_get(args, 1);
    val = dbuf_get(args, 2);

    if (!name)
	print_cmnd_help(SET_ENVIRON);
    else {
	if (val)
	    setenv(name, val, TRUE);
	else
	    setenv(name, "\0", TRUE);
    }
}

void /* user */ env_show_comm()
{
    char *name, **tmp;
    int len;

    name = dbuf_get(args, 1);
    tmp = environ;
    len = get_strlen(name);

    while (*tmp) {
	if (len && grep_against_args(1, *tmp) == FALSE) {
	    tmp++;
	    continue;
	}

	outb_add("%s", *tmp);
	tmp++;
    }

    outb_flush();
}

void /* user */ dir_change_comm()
{
    char *dir;

    dir = dbuf_get(args, 1);
    if (dir && strcmp(dir, "-") == 0)
	dir = getenv("OLDPWD");
    else if (!dir)
	dir = getenv("HOME");
    
    if (chdir(dir))
	show_user(ERROR_MSG " cd : \"%s\" : %s", dir, strerror(errno));
    else {
	mysetenv("OLDPWD", c_path);
	getcwd(c_path, PATH_MAX);
	mysetenv("PWD", c_path);
    }
}
