/***********************************************
 *  stream.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.          
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *       External References        *
 ************************************/

/* list.c */
extern ListDef *edit_list;
extern flag_t ignore_stream_weight;

/* command.c */
extern DCharBuf *args;

/* listmanip.c */
extern flag_t sort_view;

/************************************
 *        Global Variables          *
 ************************************/

/************************************
 *        Local variables           *
 ************************************/

static char tmpstr_st[COMM_SIZE], tmpstr2_st[COMM_SIZE];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int str_get_bytec()
{
    unsigned int bytec;

    bytec = 0;

    return bytec;
}

void str_init()
{
    memset(tmpstr_st,  0, COMM_SIZE);
    memset(tmpstr2_st, 0, COMM_SIZE);
}

void str_clear()
{
}

/************************************
 *       Stream weight manip        *
 ************************************/

void wght_clear_all()
{
    ListDef *lst;

    lst = plist_get_base();

    while (lst) {
	wght_clear(lst);
	lst = lst->next;
    }
}

void wght_clear(ListDef * lst)
{
    DCBNode *buf;
    int i, len;

    if (lst) {
	buf = lst->dcb->buf;
	len = lst->dcb->len;

	for (i = 0; i < len; i++)
	    buf[i].info = 0;

	lst->global_weight = 0;
    }
}

void wght_clear_good(ListDef * lst, int weight)
{
    DCBNode *buf;
    int i, len;

    if (lst) {
	buf = lst->dcb->buf;
	len = lst->dcb->len;

	for (i = 0; i < len; i++)
	    if (buf[i].info >= (short) weight) {
		lst->global_weight -= buf[i].info;
		buf[i].info = 0;
	    }
    }
}

void wght_set(ListDef *lst, int pos, int val)
{
    short nfo;

    nfo = dbuf_get_info(lst->dcb, pos);
    dbuf_set_info(lst->dcb, pos, (short) val);
    lst->global_weight += val - nfo;
}

void wght_inc(ListDef *lst, int val)
{
    DCharBuf *dcb;
    int nfo, pos, inc;
    
    dcb = lst->dcb;
    pos = dcb->pos;
    nfo = dbuf_get_info(dcb, pos);
    inc = nfo + val;

    if (inc > WEIGHT_MAX) {
	dbuf_set_info(dcb, pos, (short) WEIGHT_MAX);
	inc = WEIGHT_MAX - nfo;
    } else if (inc < WEIGHT_MIN) {
	dbuf_set_info(dcb, pos, (short) WEIGHT_MIN);
	inc = WEIGHT_MIN - nfo;
    } else {
	dbuf_set_info(dcb, pos, (short) inc);
	inc = val;
    }

    lst->global_weight += inc;
}

void wght_add(ListDef * lst)
{
    if (ignore_stream_weight)
	ignore_stream_weight = FALSE;
    else if (lst)
	wght_inc(lst, lst->play_weight);
}

void wght_remove(ListDef * lst)
{
    if (ignore_stream_weight)
	ignore_stream_weight = FALSE;
    else if (lst)
	wght_inc(lst, -(lst->skip_weight));
}

void /* user */ wght_set_index_comm()
{
    char *tok, *wght;
    int i, ind, val, low, high;

    if (!edit_list) {
	show_user(DEBUG_MSG " Edit list undefined");
	return;
    } else if (!edit_list->dcb->len) {
	show_user(DEBUG_MSG " \"%s\" list is empty", edit_list->dcb->name);
	return;
    }
    
    ind = 0;
    tok = dbuf_get(args, 2);
    wght = dbuf_get(args, 1);

    if (!tok || !wght) {
	print_cmnd_help(SET_INDEX_WEIGHT);
	return;
    }

    val = atoi(wght);
    if (val < WEIGHT_MIN || val > WEIGHT_MAX) {
	show_user(DEBUG_MSG 
		  " Invalid weight: %d. Valid range is [%d, %d]", 
		  WEIGHT_MIN, WEIGHT_MAX, val);
	return;
    }

    if (is_digit_range(tok)) {
	get_digit_range(tok, &low, &high, 0, edit_list->dcb->len - 1);
	for(i=low; i<=high; i++)
	    wght_set(edit_list, i, (short)val);
	show_user(STATUS_MSG " Entry(s) %d-%d of \"%s\" weighted to %d", 
		  low, high, edit_list->dcb->name, val);

    } else if (is_all_digits(tok)) {
	ind = atoi(tok);
	if (ind < 0 || ind > edit_list->dcb->len)
	    show_user(DEBUG_MSG " List index out of range");
	else
	    wght_set(edit_list, ind, (short)val);

    } else if (0 == strcmp("all", tok)) {
	for(i=0; i<edit_list->dcb->len; i++)
	    wght_set(edit_list, i, (short)val);
	show_user(STATUS_MSG " All entry(s) in \"%s\" weighted to %d",
		  edit_list->dcb->name, val);

    } else {
	for(i=0; i<edit_list->dcb->len; i++) {
	    str_get_id(edit_list, i, tmpstr_st);
	    if (grep_against_args(2, tmpstr_st) == FALSE)
		continue;
	    wght_set(edit_list, i, (short)val);
	    ind++;
	}
	show_user(STATUS_MSG " Weighted %d entry(s) in \"%s\" list", ind,
		  edit_list->dcb->name);
    }
}

/************************************
 *          Stream manip            *
 ************************************/

int str_add(DCharBuf * dcb, char *file, short info, int tag, int check)
{
    int len;

    if (check && !tag && tag_make(file, tmpstr_st)) {

	len = safe_sprintf(tmpstr2_st, COMM_SIZE, "%s%s%s", 
			   skip_space(tmpstr_st), TAG_CODE, file);

	return dbuf_add(dcb, tmpstr2_st, ++len, info);
    } else
	return dbuf_add_stri(dcb, file, info);
}

int str_remove_range(ListDef * lst, int low, int high)
{
    int i, ret, nfo;

    if (!lst)
	return FALSE;

    nfo = 0;

    for(i = low; i <= high; i++)
	nfo += dbuf_get_info(lst->dcb, i);

    ret = dbuf_remove_range(lst->dcb, low, high);
    if (ret)
	lst->global_weight -= nfo;

    return ret;
}

int str_remove(ListDef * lst, int ndx)
{
    int ret;
    short nfo;

    if (!lst)
	return FALSE;
    
    nfo = dbuf_get_info(lst->dcb, ndx);
    ret = dbuf_remove(lst->dcb, ndx);
    if (ret) 
	lst->global_weight -= nfo;

    return ret;
}

char *str_get(DCharBuf * dcb, int index, int what)
{
    char *tag, *fn;

    tag = dbuf_get(dcb, index);

    if (!tag)
	return NULL;

    fn = grep_str(TAG_CODE, tag);

    if (!fn) {
	fn = tag;
	tag = NULL;
    }

    return (what == STR_FILE) ? fn : tag;
}

void str_get_info(DCharBuf * dcb, int index, char **file, char **tag)
{
    char *t, *f;

    t = dbuf_get(dcb, index);

    if (!t) {
	*file = *tag = NULL;
	return;
    }

    f = grep_str(TAG_CODE, t);

    if (!f) {
	*file = t;
	*tag = NULL;
    } else {
	*file = f;
	*tag = t;
    }
}

void str_get_id(ListDef * lst, int ndx, char *out)
{
    char *file, *tag;
    int pos;

    *out = ZCHAR;
    str_get_info(lst->dcb, ndx, &file, &tag);

    if (!file)
	return;

    switch (lst->listview) {
    case TAG_FILE_PREF:
    case TAG_PATH_PREF:
	if (tag) {
	    pos = (file - tag) - TAG_CODE_LEN;

	    if (pos < COMM_SIZE) {
		strncpy(out, tag, pos + 1);
		out[pos] = ZCHAR;
	    }
	} else {
	    if (lst->listview == TAG_FILE_PREF)
		strcpy(out, file_from_path(file));
	    else
		strcpy(out, file);
	}
	break;

    case FILE_ONLY_PREF:
	strcpy(out, file_from_path(file));
	break;

    case PATH_FILE_PREF:
    default:
	strcpy(out, file);
	break;
    }
}

void str_get_sort_id(char *in, char *out)
{
    char *file, *tag;
    int pos;

    if (!in)
	return;

    tag = NULL;
    *out = ZCHAR;
    file = grep_str(TAG_CODE, in);

    if (!file)
	file = in;
    else
	tag = in;

    switch (sort_view) {
    case TAG_FILE_PREF:
    case TAG_PATH_PREF:
	if (tag) {
	    pos = (file - tag) - TAG_CODE_LEN;
	    
	    if (pos < COMM_SIZE) {
		strncpy(out, in, pos + 1);
		out[pos] = ZCHAR;
	    }
	} else {
	    if (sort_view == TAG_FILE_PREF)
		strcpy(out, file_from_path(in));
	    else
		strcpy(out, in);
	}
	break;

    case FILE_ONLY_PREF:
	strcpy(out, file_from_path(file));
	break;

    case PATH_FILE_PREF:
    default:
	strcpy(out, file);
	break;
    }
}
