/***********************************************
 *  util.c                                 
 *                                            
 *  Copyright 2001 Eric Kratzer.              
 *  Includes Xaudio software Copyright (C)
 *           1996-2000 Xaudio Corporation
 *           All Rights Reserved.              
 *
 *  This file is part of imp3sh.
 *                                            
 *  imp3sh is free software; you can distribute
 *  it and/or modify it under the terms of the
 *  GNU General Public License as published by
 *  the Free Software Foundation; either version
 *  2 of the License, or (at your option) any
 *  later version.
 *
 *  imp3sh is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of 
 *  MERCHAN'\t'ILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE. See the GNU General Public License
 *  for more details.
 *
 *  You should have recieved a copy of the GNU
 *  General Public License along with imp3sh; if
 *  not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place, Suite 330, Boston MA
 *  02111-1307 USA
 ************************************************/

/************************************
 *          Include Files           *
 ************************************/

#include "imp3sh.h"
#include "all.h"

/************************************
 *        External References       *
 ************************************/

/* command.c */
extern DCharBuf *args;

/* tags.c */
extern flag_t read_file_tags;

/************************************
 *        Global Variables          *
 ************************************/

/************************************
 *         Local variables          *
 ************************************/

static char tmpstr_ut[PATH_MAX], search_dir[PATH_MAX];

/************************************
 *         Module routines          * 
 ************************************/

unsigned int util_get_bytec()
{
    unsigned int bytec;

    bytec = 0;

    return bytec;
}

void util_init()
{
    memset(tmpstr_ut,  0, PATH_MAX);
    memset(search_dir, 0, PATH_MAX);
}

void util_clear()
{
}

/************************************
 *        Utility functions         *
 ************************************/

int is_upper_c(char c)
{
    return (c >= 'A' && c <= 'Z') ? TRUE : FALSE;
}

int is_all_digits(char *str)
{
    if (str) {
	while (*str != ZCHAR) {
	    if (!isdigit(*str))
		return FALSE;
	    str++;
	}
	return TRUE;
    }

    return FALSE;
}

int is_digit_range(char *s)
{
    char *dash, *str2, str[INFO_SIZE];

    safe_strcpy(str, s, INFO_SIZE);
    dash = strchr(str, '-');

    if (dash) {
	str2 = (dash + 1);
	*dash = ZCHAR;

	if ((*str == ZCHAR && *str2 == ZCHAR) ||
	    (*str != ZCHAR && !is_all_digits(str)) ||
	    (*str2 != ZCHAR && !is_all_digits(str2)))
	    return FALSE;
	else
	    return TRUE;
    }

    return FALSE;
}

void get_digit_range(char *str, int *low, int *high, int lrep, int hrep)
{
    char *dash, *str2;

    dash = strchr(str, '-');

    if (dash) {
	str2 = (dash + 1);
	*dash = ZCHAR;

	if (*str != ZCHAR) {
	    *low = atoi(str);
	    if (*low < lrep)
		*low = lrep;
	    else if (*low > hrep)
		*low = hrep;
	} else
	    *low = lrep;

	if (*str2 != ZCHAR) {
	    *high = atoi(str2);
	    if (*high < lrep)
		*high = lrep;
	    else if (*high > hrep)
		*high = hrep;
	} else
	    *high = hrep;
    } else
	*low = *high = -1;
}

void remove_end_space(char *str)
{
    char *t;

    t = NULL;

    if (str) {
	while (*str != ZCHAR) {
	    if (*str != '\t' && *str != ' ')
		t = str;
	    str++;
	}

	if (t) *(t + 1) = ZCHAR;
    }
}

int get_strlen(char *str)
{
    char *w;

    if (str) {
	w = str;
	while (*w++ != ZCHAR);
	return (w - str - 1);
    }

    return 0;
}

int remove_newline(char *str)
{
    int len;

    len = get_strlen(str);

    if (len && str[len - 1] == END_COMM)
        str[--len] = ZCHAR;

    return len;
}

int get_fline(FILE * fd, char *str, int size)
{
    *str = 0;
    fgets(str, size, fd);
    return remove_newline(str);
}

void make_printable(char *str)
{
    if (str)
	while (*str != ZCHAR) {
	    if (*str < 0x20 || *str > 0x7e)
		*str = ' ';
	    str++;
	}
}

int is_all_space(char *str)
{
    if (str)
	while (*str != ZCHAR)
	    if (*str++ > 0x20)
		return FALSE;

    return TRUE;
}

int is_any_space(char *str)
{
    if (str)
	while (*str != ZCHAR)
	    if (*str++ == 0x20)
		return TRUE;

    return FALSE;
}

char *skip_space(char *str)
{
    if (str)
	while (*str == ' ' || *str == '\t')
	    str++;

    return str;
}

char any_upper_case(char *str)
{
    if (str)
	while (*str != ZCHAR)
	    if (is_upper_c(*str++))
		return TRUE;

    return FALSE;
}

char *grep_str(char *pat, char *str)
{
    char *ret, not;
    int plen, (*compare_func)(const char *, const char *, size_t);

    ret = NULL;
    not = FALSE;
    compare_func = strncasecmp;

    if (!str || !pat)
	return NULL;

    if (*pat == '!') {
	not = TRUE; pat++;
    }

    plen = get_strlen(pat);

    if (any_upper_case(pat))
	compare_func = strncmp;

    while (*str != ZCHAR)
	if (compare_func(pat, str++, plen) == 0) {
	    ret = (--str + plen);
	    break;
	}

    if (not) {
	if (ret) ret = NULL;
	else ret = "ok"; /* not writeable */
    }

    return ret;
}

int grep_against_args(int start, char *str)
{
    char *pat;

    pat = dbuf_get(args, start++);

    if (!pat)
	return FALSE;

    while (pat) {
	if (grep_str(pat, str) == NULL)
	    return FALSE;
	pat = dbuf_get(args, start++);
    }

    return TRUE;
}

void unescape_string (char *match) /* original from Caleb Crome */
{
  char c, *str, *mat;

  *tmpstr_ut = 0;
  str = tmpstr_ut;
  mat = match;

  while ((c = *mat++))
      if (c != '\\')
	  *str++ = c; 

  *str++ = 0;
  safe_strcpy(match, tmpstr_ut, (int)(str-tmpstr_ut));
}

/* Add '\' characters before any characters that would be interpreted by the
   shell. For example, a space doesn't do much good for a shell when it's in
   a file name.  Neither does $%&;|'(){} and who knows what else.
 */
/* Danger: "match" must point to enough memory to hold the newly escaped
           string. Maybe pass in a max length for "match" ?
*/
void escape_string(char *match) /* original from Caleb Crome */
{
    char c, b, *mat, *str, *bad;

    *tmpstr_ut = 0;
    str = tmpstr_ut;
    mat = match;

    while ((c = *mat++)) {
	bad = BAD_SHELL_CHARACTERS;
	while((b = *bad++))
	    if (c == b) {
		*str++ = '\\';
		break;
	    }
	*str++ = c;
    }

    *str++ = 0;
    safe_strcpy(match, tmpstr_ut, (int)(str-tmpstr_ut));
}

char *arg_tokenize(char *t, char **mem, char delim, char *paren)
{
    char *ret, *end;

    ret = end = NULL;
    *paren = FALSE;

    if (t)
	*mem = skip_space(t);
    if (!*mem || !**mem)
	return NULL;

    ret = *mem;
    if (*ret == '\"') {
	*paren = TRUE;
	end = strchr(++ret, '\"');
    } else {
	end = ret;
	while (TRUE) {
	    end = strchr(end, delim);
	    if (!end || (*(end-1) != '\\'))
		break;
	    end++;
	}
    }
      
    if (end) {
	*end = ZCHAR;
	*mem = skip_space((end + 1));
    } else
	*mem = NULL;

    return ret;
}

char *tokenize(char *t, char **mem, char delim)
{
    char *ret, *end;

    ret = end = NULL;

    if (t)
	*mem = skip_space(t);
    if (!*mem || !**mem)
	return NULL;

    ret = *mem;
    if (*ret == '\"')
	end = strchr(++ret, '\"');
    else {
	/* ignore escaped characters as delimeters (Caleb Crome) */
	end = ret;
	while (TRUE) {
	    end = strchr(end, delim);
	    if (!end || (*(end-1) != '\\'))
		break;
	    end++;
	}
    }
      
    if (end) {
	*end = ZCHAR;
	*mem = skip_space((end + 1));
    } else
	*mem = NULL;

    return ret;
}

int tokenize_to_buf(DCharBuf * dcb, char *str, char sep)
{
    char *cm, *cmem;
    int toks;

    if (!dcb || !str)
	return 0;

    toks = 0;
    cmem = NULL;
    cm = tokenize(str, &cmem, sep);
    
    while (cm) {
	toks ++;
	dbuf_add_str(dcb, cm);
	cm = tokenize(NULL, &cmem, sep);
    }

    return toks;
}

int get_cloc(char c, char *str)
{
    char *tmp;

    tmp = strchr(str, c);

    return (tmp) ? tmp - str : -1;
}

/* undefined if called with '\t' or ' ' */
int get_cpos(char c, char *str)
{
    flag_t aft, bef, fnd;

    if (!str)
        return AMP_NONE;

    bef = aft = fnd = FALSE;

    while(*str != ZCHAR) {
        if (!fnd && *str == c)
	    fnd = TRUE;
        else if (*str != ' ' && *str != '\t') {
	    if (fnd)
	        aft = TRUE;
	    else
	        bef = TRUE;
	}
	str++;
    }

    if (!fnd)
        return AMP_NONE;
    else if (aft && bef)
        return AMP_MID;
    else if (aft)
        return AMP_FRONT;
    else if (bef)
        return AMP_END;
    else
        return AMP_ALL;
}

char *safe_strcpy(char *dest, char *src, int len)
{
    dest[len - 1] = ZCHAR;
    return strncpy(dest, src, len - 1);
}

int safe_sprintf(char *dest, size_t size, const char *frmt, ...)
{
    int len;
    va_list ap;

    dest[size-1] = 0;
    va_start(ap, frmt);
    len = vsnprintf(dest, size-1, frmt, ap);
    va_end(ap);

    return len;
}

char *file_from_path(char *name)
{
    char *ret;

    if (name) {
	ret = strrchr(name, '/');

	if (ret)
	    return (ret + 1);
    }

    return name;
}

struct timeval add_timevals(struct timeval a, struct timeval b)
{
    struct timeval c;

    c.tv_sec = a.tv_sec + b.tv_sec;
    c.tv_usec = a.tv_usec + b.tv_usec;

    if (c.tv_usec >= 1000000) {
	c.tv_sec ++;
	c.tv_usec -= 1000000;
    }

    return c;
}

struct timeval sub_timevals(struct timeval a, struct timeval b) /* a - b */
{
    struct timeval c;

    c.tv_sec = a.tv_sec - b.tv_sec;
    c.tv_usec = a.tv_usec - b.tv_usec;

    if (c.tv_usec < 0) {
	c.tv_sec --;
	c.tv_usec = 1000000 + c.tv_usec;
    }

    return c;
}

void build_msec_time_string(struct timeval tv, char *out, int maxlen)
{
    int hrs, min, sec, msec;

    hrs = tv.tv_sec / 3600; tv.tv_sec -= hrs * 3600;
    min = tv.tv_sec / 60; tv.tv_sec -= min * 60;
    sec = tv.tv_sec;

    msec = 0;
    if (!hrs && !min && tv.tv_usec)
	msec = tv.tv_usec / 1000;
    
    if (msec)
	safe_sprintf(out, maxlen, "%02d:%02d:%02d.%03d", hrs, min, sec, msec);
    else
	safe_sprintf(out, maxlen, "%02d:%02d:%02d", hrs, min, sec);
}

void build_time_string(time_t len, char *out, int maxlen)
{
    int hrs, min, sec;

    hrs = len / 3600; len -= hrs * 3600;
    min = len / 60; len -= min * 60;
    sec = len;

    safe_sprintf(out, maxlen, "%02d:%02d:%02d", hrs, min, sec);
}

FILE *myfopen(char *name, const char *perm)
{
    struct stat sstat;
    FILE *fd;

    if (!stat(name, &sstat) && (sstat.st_mode & S_IFDIR)) {
	errno = EISDIR;
	return NULL;
    }

    fd = fopen(name, perm);

    return fd;
}

void mysetenv(char *name, char *val)
{
    if (name) {
	/* unsetenv(name); */
	if (val)
	    setenv(name, (const char *) val, /* FALSE */ TRUE);
	else
	    setenv(name, "\0", /* FALSE */ TRUE);
    }
}

int is_url(char *stream)
{
    int len;

    len = get_strlen(stream);

    if ((len >= 5 && strncasecmp(stream, "http:", 5) == 0) ||
	(len >= 4 && strncasecmp(stream, "ftp:", 4) == 0) ||
	(len >= 4 && strncasecmp(stream, "udp:", 4) == 0))
	return TRUE;
    else
	return FALSE;
}

float format_bytec(unsigned int bytec, char *unit)
{
    if (bytec > (1024 * 1024)) {
	*unit = 'M';
	return ((float) bytec) / 1024.0 / 1024.0;
    } else if (bytec > 1024) {
	*unit = 'K';
	return ((float) bytec) / 1024.0;
    } else {
	*unit = ' ';
	return (float) bytec;
    }
}

int d_chck(mode_t mode)
{
    return ((mode & S_IFDIR) && !(mode & S_IFLNK)) ? TRUE : FALSE;
}

void add_dir(int add_d, DCharBuf * pplist, DCharBuf * lst)
{
    struct dirent *dirent;
    struct stat sstat;
    DIR *dirstr;
    char *dir, *fname;

    dir = dbuf_get(pplist, 0);

    if (!dir)
	return;

    if (chdir(dir))
	show_user(ERROR_MSG " Failed to read \"%s\" : %s", dir,
		  strerror(errno));
    else {
	dirstr = opendir(".");

	if (!dirstr) 
	    show_user(ERROR_MSG " Failed to open \"%s\" : %s", dir,
		      strerror(errno));
	else {
	    getcwd(search_dir, PATH_MAX);
	    dirent = readdir(dirstr);

	    while (dirent) {
		fname = dirent->d_name;
		if (strcmp(fname, ".") != 0 && strcmp(fname, "..") != 0) {
		    lstat(fname, &sstat);
		    if (d_chck(sstat.st_mode)) {
			if (add_d) {
			    safe_sprintf(tmpstr_ut, PATH_MAX, "%s/%s", 
					 search_dir, fname);
			    dbuf_add_str(pplist, tmpstr_ut);
			}
		    } else if (lgc_check_format(fname)) {
			safe_sprintf(tmpstr_ut, PATH_MAX, "%s/%s", 
				     search_dir, fname);
			str_add(lst, tmpstr_ut, 0, FALSE, read_file_tags);
		    }
		}
		dirent = readdir(dirstr);
	    }
	    closedir(dirstr);
	}   
    }

    dbuf_remove(pplist, 0);
}
